package UBERNUtils;
use strict;
use POSIX qw(setsid);
use DateTime::Format::MySQL;
use SOAP::Transport::HTTP;
use Date::Parse;
use MIME::Base64 qw( encode_base64 decode_base64 );


#   Trims and chomps tabs/spaces/newlines

sub trimchomp
{
    my $string = $_[0];
    chomp($string);
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    chomp($string);
    return $string;
}


#   Function that retrieves the raw input from the TCP sockets of the meshnodes<->portal
#   and that extracts sensor node id and the data line.

sub purifyRetrievedSocketData {
    my $data = $_[0];
    $data = trimchomp($data);
    # after the '|' sign, the line begins!
    my $indexOfSign =  index($data, '|');
    my $number = substr($data,0,$indexOfSign);
    $number =~ s/'//g;
    my $puredata = substr($data,$indexOfSign+1);
    $puredata = trimchomp($puredata);
    
    if( !isInteger($number) ) {
        $number = ""; 
    }
     
    return ($number, $puredata);
}

#   gets only one lines in the form "line <CURRENT_TIME: ...>"
#   and removes - http fragments (e.g. </pre> or similia)
#               - empty lines
#               - lines with only time values

sub purifyLine {
    my $data = $_[0];
    my $line = trimchomp($data);
    $line =~ s/<.*>//;
    $line =~ s/<\.*>//;
    $line =~ s/\x00//;
    $line = trimchomp($line);
          
    # check if the line without the CURRENT_TIME part would be empty!
    my $tempLine = $line;
    my @toks = split(/CURRENT_TIME/,$tempLine);
    my $lineValue = trimchomp(@toks[0]);
    if( ! ( $line eq "") && !( $lineValue eq "" )) {
       return $line;
    }
    return "";
}


#   gets data lines in the form "line <CURRENT_TIME: ...>"
#   and removes - http fragments (e.g. </pre> or similia)
#               - empty lines
#               - lines with only time values

sub purifyData {
    my @data = @_;
    chomp(@data);
    my @result;
    for(my $i=0; $i<@data; $i++){
       my $line = trimchomp(@data[$i]);
       $line =~ s/<.*>//;
       $line =~ s/<\.*>//;
       $line =~ s/\x00//;
       $line = trimchomp($line);
          
       # check if the line without the CURRENT_TIME part would be empty!
       my $tempLine = @data[$i];
       my @toks = split(/CURRENT_TIME/,$tempLine);
       my $lineValue = trimchomp(@toks[0]);
       
       if( ! ( $line eq "") && !( $lineValue eq "" )) {
          push(@result, $line);
       }
       
    }
    return @result;
}

#   Returns 1 if argument is integer, 0 if not...

sub isInteger {
    if ( $_[0] =~ /^\d+$/ ) { 
       return 1;
    } 
    return 0;
}



sub readConfiguration {
     # read the configuration
     my $conf = $_[0];
     
     open(CONFIG, $conf)  or die "Can't read file $conf :$!";   
     while (<CONFIG>) {
         chomp;                                       # no newline
         s/#.*//;                                     # no comments
         s/^\s+//;                                    # no leading white
         s/\s+$//;                                    # no trailing white
         next unless length;                          # anything left?
         my ($var, $value) = split(/\s*=\s*/, $_, 2);
         $ENV{"$var"} = $value;
     } 
}

sub generateURN {
   my $id = $_[0];
   my $urn = "urn:wisebed:node:ubern:" . $id;
   return $urn;
}

sub stripeIDfromURN {
   my $urn   = $_[0];
   # stripe away the testbed name
   $urn =~ s/.*://;
   return int($urn);
}

sub containsElement {
   my $arrayRef   = $_[0];
   my $element = $_[1];
   for(my $i=0; $i<@$arrayRef; $i++){
      if($element eq @$arrayRef[$i]) {
         return 1;
      }
   }
   return 0;
}

sub indexOfElement {
   my $arrayRef   = $_[0];
   my $element = $_[1];
   for(my $i=0; $i<@$arrayRef; $i++){
      if($element eq @$arrayRef[$i]) {
         return $i;
      }
   }
   return 0;
}


#   Fork and daemonize a perl subroutine
#   write output and stderr to files

sub daemonize {
    
    my $service = $_[0];
    my $serviceName = $_[1];
    
    # redirect STDOUT and STDERR to outfiles
    
    my $outfile = "$ENV{'TARWIS_DIRECTORY'}"."UBERN_TESTBED_IMPLEMENTATION/$serviceName.out";
    my $errfile = "$ENV{'TARWIS_DIRECTORY'}"."UBERN_TESTBED_IMPLEMENTATION/$serviceName.err";
    
    chdir '/'                   or die "Can't chdir to /: $!";
    open STDIN,  "/dev/null"    or die "Can't read /dev/null: $!";
    open STDOUT, ">$outfile"  or die "Can't write to $outfile: $!"; 
    open STDERR, ">$errfile"  or die "Can't write to $errfile: $!";
    
    
    # fork process to background!
    my $myproc = Proc::Simple->new();  # Create a new process object
    $myproc->start($service);  
    
}



#   contains non-ASCII characters
sub containsBadEncoding {
    my $data = $_[0];
    my @chars=split(//,$data);
    for(my $i=0; $i<@chars; $i++){
       my $n = ord(@chars[$i]);
       if( ! ( ( $n >= 0 && $n <= 13) || ($n >= 32 && $n <= 127) ) ) {
           return 1;
       }
       
    }
    return 0;
}

#   contains non-ASCII characters
sub removeBadEncoding {
    my $data = $_[0];
    my $puredata = "";
    my @chars=split(//,$data);
    for(my $i=0; $i<@chars; $i++){
       my $n = ord(@chars[$i]);
       if( $n == 0 || $n == 9 || ($n >= 32 && $n <= 126) ) {
         $puredata .=  @chars[$i];
       }
    }
    return $puredata;
}



sub getTimeInMySQLFormat() {
   my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(time);
   my $time = sprintf("%4d-%02d-%02d %02d:%02d:%02d", $year+1900,$mon+1,$mday,$hour,$min,$sec);
   return $time;
}

#   for log files: cut off all lines of <arg1> if the file is bigger than <arg2> lines 

sub truncateFile {
    my $file = $_[0];
    my $lines = $_[1];

    open (FILE, "<$file") or die "Can't read $file $!";
    my(@lines) = <FILE>;
    close(FILE);
    
    if(@lines > $lines) {
         open (FILE, ">$file") or die "Can't read $file $!";
         my $fileLines = @lines;
         for(my $i= $fileLines - $lines; $i<@lines; $i++)
         {
             print FILE @lines[$i];
         }
         close(FILE);
    }
    
}

sub portChecker {
    my $host = $_[0];
    my $port = $_[1];
    my $sock = new IO::Socket::INET(PeerAddr=>$host,PeerPort=>$port,Proto=>'tcp');
    if($sock)
    {
       close $sock;
       return 1;
    }
    return 0;
}

sub verifyPort {
    my $host = $_[0];
    my $port = $_[1];
    my $portServed = portChecker($_[0],$_[1]);
    return $portServed;
}


sub encodeFile {
   my $filepath = $_[0];
   open INFILE, '<', $filepath;
   binmode INFILE;
   my $buf;
   my $encoded;
   while ( read( INFILE, $buf, 100000000 ) ) {
        $encoded.=encode_base64($buf);
   }
   close INFILE;
   return $encoded;
}

sub decodeFile {
   my $filepath = $_[0];
   open INFILE, '<', $filepath;
   binmode INFILE;
   my $buf;
   my $decoded;
   while ( read( INFILE, $buf, 100000000 ) ) {
        $decoded.=decode_base64($buf);
   }
   close INFILE;
   return $decoded;
}

sub decodeAndWriteToFile {
   my $image = $_[0];
   my $filename = $_[1];
   $image = decode_base64($image);
   open (FILE, ">$filename") or die "Can't read $filename !";
   print FILE $image;
   close(FILE);
}


1;