package DatabaseInterface;


use strict;
use POSIX qw(setsid);
use POSIX ();
use LWP;
use HTTP::Request;
use HTTP::Response;
use DBI;
use DBD::mysql;
use Carp::Assert;
use XML::Writer;
use IO::File;
use List::Util qw(max);
use Date::Parse;
use Date::Format;


require "NetworkResourcesReader.pm";
require "Utils.pm";

# read the configuration 
my $conffileOK = &Utils::readConfiguration("/etc/TARWIS.conf");
if(!$conffileOK) {
    print "DatabaseInterface: configuration file not found or non-readable, exiting now!\n";
    exit(1);
}



# class constructor
sub new {
    my ($class) = @_;
    my $self = {
       _dbh => undef,
       _owner => undef,
    };
    
    bless $self, $class;
 
    my $host     =  $ENV{"MYSQL_DATABASE_HOST"};
    my $database =  $ENV{"MYSQL_DATABASE_NAME"};
    my $user     =  $ENV{"MYSQL_DATABASE_USER"};
    my $pw       =  $ENV{"MYSQL_DATABASE_PASSWORD"};
    my $connectionInfo="dbi:mysql:$database;$host";
    
    $self->{_dbh} = newConnection($connectionInfo, $user, $pw);
    $self->{_dbh}->{'mysql_auto_reconnect'} = 1;
    
    return $self;
}

sub owner {
    my ( $self, $owner ) = @_;
    $self->{_owner} = $owner if defined($owner);
    return $self->{_owner};
}

sub getDatabase {
    my ($self) = shift;
    return $self->{_dbh};
}


sub newConnection {
    my $connectionInfo = $_[0];
    my $user = $_[1];
    my $pw = $_[2];
    my $db = DBI->connect($connectionInfo,$user,$pw) or die "Couldn't connect to database: " . DBI->errstr;   
    return $db;
} 



#############################################################################################################
### FUNCTIONS on TABLE sensornodes_output  ##################################################################
#############################################################################################################

#   Resets the table wsn_info:sensornodes_output



sub resetSensorNodesOutputTables {
    my $self = shift;
    my $exp_id  = $_[0];
    if($exp_id eq "") {
        $self->executeDropTruncateQuery("TRUNCATE TABLE sensornodes_output ");
    }
    else {
        my $dbh = $self->getDatabase();  
        $exp_id  = $dbh->quote( $exp_id );
        $self->executeDropTruncateQuery("DELETE FROM sensornodes_output WHERE exp_id=$exp_id");
        # $self->executeDropTruncateQuery("DELETE FROM sensornodes_output_binary WHERE exp_id=$exp_id");
    }
}

#   Function writes node_id, line and time into the table wsn_info:sensornodes_output. This table
#   lists any output that has been retrieved from the serial ports of the sensors

sub writeTimestampedSensorNodeOutputToDB {
    my $self    = shift;
    my $node_id = $_[0];
    my $exp_id  = $_[1];
    my $line    = $_[2];
    my $time    = $_[3];
    my $usecs   = $_[4]; 
 
 
    # make connection to database
    my $dbh = $self->getDatabase();    
    
    # use quote to omit escape character errors
    $node_id = $dbh->quote( $node_id );
    $exp_id  = $dbh->quote( $exp_id );
    $line    = $dbh->quote( $line );
    $time    = $dbh->quote( $time );
    $usecs   = $dbh->quote( $usecs );
    
    # use NULL as row-parameter as it is auto-incremented!
    my $query = "INSERT DELAYED INTO sensornodes_output VALUES($node_id,$exp_id,$line,$time,$usecs,NULL)";
    
    eval{
        # prepare and execute query
        my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
        $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
        $sth->finish();
        return 1;
    };
    if ($@) {
        return 0;
    };
    return 0;
}

# faster function for inserting several lines in a row

sub writeTimestampedSensorNodeOutputToDB_arrays {
    my $self    = shift;
    my @node_id = @{$_[0]};
    my @exp_id  = @{$_[1]};
    my @line    = @{$_[2]};
    my @time    = @{$_[3]};
    my @usecs   = @{$_[4]}; 
 
    if(@node_id != @exp_id) {
        print STDOUT "DatabaseInterface::writeTimestampedSensorNodeOutputToDB_arrays #node_id != #exp_id\n";
        print STDERR "DatabaseInterface::writeTimestampedSensorNodeOutputToDB_arrays #node_id != #exp_id\n";
        return;
    }
 
    # make connection to database
    my $dbh = $self->getDatabase();    
    
    my $query = "INSERT IGNORE INTO sensornodes_output VALUES"; 
    
    my $counter = 0;
    
    for(my $i=0; $i<@node_id; $i++) {
    	my $exp_id = @exp_id[$i];
    	if(!($exp_id =~ m/NOEXPERIMENT/)) {
	        # use quote to omit escape character errors
	        my $node_id = $dbh->quote( @node_id[$i] );
	        $exp_id  = $dbh->quote( $exp_id );
	        my $line    = $dbh->quote( @line[$i] );
	        my $time    = $dbh->quote( @time[$i] );
	        my $usecs   = $dbh->quote( @usecs[$i] );
	        if($counter>0) {
	           $query.=",\n";
            }
	        $query .= "($node_id,$exp_id,$line,$time,$usecs,NULL)";	
	        $counter++;
    	}
    }
    
    $query .= ";";
    
    # print "writeTimestampedSensorNodeOutputToDB_arrays: query: $query\n";
    
    if($counter > 0) {
	    eval{
	        my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
	        $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
	        $sth->finish();
	        return 1;
	    };
	    if ($@) {
	    	print "DatabaseInterface: something went wrong with MySQL and the query: $query\n";
	        return 0;
	    };
    }
    return 0;
}

#   return the timestamped output of a node
sub getOutputOfNode {
    my $self        = shift;
    my $id          = $_[0];
    my $exp_id      = $_[1];
    my @times       = @{$_[2]};
    my @usecs       = @{$_[3]};
    my @lines       = @{$_[4]};
    
    my $dbh = $self->getDatabase();    
    
    $id         = $dbh->quote( $id );
    $exp_id     = $dbh->quote( $exp_id );
    
    my $query = "SELECT * FROM sensornodes_output WHERE node_urn = $id AND exp_id = $exp_id  ORDER BY rowindex";
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
    
    # assign fields to variables
    my $ID;
    my $line;
    my $time;
    my $usecs;
    my $rowindex;
    
    $sth->bind_columns(\$ID, \$line, \$time, \$usecs, \$rowindex);
    while($sth->fetch()) {
       push(@times,$time);
       push(@usecs,$usecs);
       push(@lines,$line);
    }
    $sth->finish();
    
    return(\@times, \@usecs, \@lines);
}

sub getAllTextOutputOfNodes_orderedByTime {
    my $self          = shift;
    my $exp_id        = $_[0];
    my $dbh           = $self->getDatabase();    
    $exp_id           = $dbh->quote( $exp_id );
    my $query         = "SELECT node_urn, line, time, usecs from sensornodes_output WHERE exp_id = $exp_id ORDER BY time, usecs";
    my @allRows       = $self->executeSelectQuery($query);
    return @allRows;
}

sub getAllNodeTypes {
    my $self          = shift;
    my $dbh           = $self->getDatabase();    
    my $query         = "SELECT DISTINCT value FROM sensornodes natural join properties WHERE properties.description='nodeType'"; 
    my @allRows       = $self->executeSelectQuery($query);
    my $numberOfRows  = @allRows;
    my @ids = ();
    for(my $i=0; $i<$numberOfRows; $i++){
       push(@ids, @allRows[$i]->[0]);
    }
    return @ids;
}



#   get the most recent time of a registered output line
sub getLatestTimeInOutput {
    my $self     = shift;
    my $exp_id   = $_[0];
    my $dbh = $self->getDatabase();    
    $exp_id      = $dbh->quote( $exp_id );
    my @rows     = $self->executeSelectQuery("SELECT MAX(time) FROM sensornodes_output WHERE exp_id = $exp_id");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    return @row[0];   
}


#   get the most recent usecs of the most recent registered output line

sub getLatestUsecsInOutput {
    my $self     = shift;
    my $exp_id   = $_[0];
    my $dbh = $self->getDatabase();
    my @rows     = $self->executeSelectQuery("SELECT MAX(usecs) FROM sensornodes_output WHERE time = (SELECT MAX(time) AND exp_id = $exp_id FROM sensornodes_output)");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    return @row[0];   
}

#   get the time since the last output 

sub timeSinceLatestTimeOfOutput {
    my $self     = shift;
    my $dbh      = $self->getDatabase();
    my @rows     = $self->executeSelectQuery("SELECT MAX(sensornodes_output.time) from sensornodes_output");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    my $lasttime = &Utils::parseDate(@row[0]);   
    my $now = time();
    return ($now-$lasttime) 
}


#   Function writes node_id, line and time into the table wsn_info:sensornodes_output. This table
#   lists any output that has been retrieved from the serial ports of the sensors

sub writeTimestampedSensorNodeOutputToDB_binary {
    my $self    = shift;
    my $node_id = $_[0];
    my $exp_id  = $_[1];
    my $line    = $_[2];
    my $time    = $_[3];
    my $usecs   = $_[4]; 
 
 
    # make connection to database
    my $dbh = $self->getDatabase();    
    
    # use quote to omit escape character errors
    $node_id    = $dbh->quote( $node_id );
    $exp_id     = $dbh->quote( $exp_id );
    $line       = $dbh->quote( $line );
    $time       = $dbh->quote( $time );
    $usecs      = $dbh->quote( $usecs );
    
    # use NULL as row-parameter as it is auto-incremented!
    my $query = "INSERT DELAYED INTO sensornodes_output_binary VALUES($node_id,$exp_id, $line,$time,$usecs,NULL)";
    
    # print "query=$query\n"; 
    
    eval{
        # prepare and execute query
        my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
        $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
        $sth->finish();
        return 1;
    };
    if ($@) {
        return 0;
    };
    
    
    return 0;

}

sub deleteLastNRowsOfNodeOutput {
    my $self     = shift;
    my $node_urn = $_[0];
    my $exp_id   = $_[1];
    my $N        = $_[2];
    my $dbh = $self->getDatabase(); 
    
    $node_urn = $dbh->quote( $node_urn );
    $exp_id     = $dbh->quote( $exp_id );
    
    # do not quote the $N!!!
    my $query = "DELETE FROM sensornodes_output WHERE node_urn=$node_urn AND exp_id = $exp_id ORDER BY time,usecs DESC LIMIT $N";
    # print $query . "\n";
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
   
    $sth->finish();
}

#   get the output the node with the URN supplied as first argument
#   that is not older than N seconds (N supplied as second argument)

sub getLastNSecsOutputOfNode  {
    my $self     = shift;
    my $node_urn = $_[0];
    my $exp_id   = $_[1];
    my $secs     = $_[2];
    my @times    = @{$_[3]};
    my @usecs    = @{$_[4]};
    my @lines    = @{$_[5]};
    
    my $dbh = $self->getDatabase();    
    
    $node_urn = $dbh->quote( $node_urn );
    $exp_id     = $dbh->quote( $exp_id );
    
    my $query = "SELECT * FROM sensornodes_output WHERE node_urn=$node_urn AND exp_id = $exp_id AND time > (SUBTIME((SELECT MAX(time) FROM sensornodes_output WHERE node_urn=$node_urn), SEC_TO_TIME($secs))) ORDER BY rowindex";
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
    
    # assign fields to variables
    my $ID;
    my $line;
    my $time;
    my $usecs;
    my $rowindex;
    
    $sth->bind_columns(\$ID, \$line, \$time, \$usecs, \$rowindex);
    while($sth->fetch()) {
       push(@times,$time);
       push(@usecs,$usecs);
       push(@lines,$line);
    }
    $sth->finish();
    return(\@times, \@usecs, \@lines);
}


#############################################################################################################
### FUNCTIONS on TABLE outputBuffer        ##################################################################
#############################################################################################################

sub resetSensorNodesOutputBufferTable(){
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE outputBuffer");
}


sub writeTimestampedSensorNodeOutputToOutputBuffer {
    my $self    = shift;
    my $node_id = $_[0];
    my $line    = $_[1];
    my $time    = $_[2];
    my $usecs   = $_[3]; 
 
    my $dbh = $self->getDatabase();    
    
    $node_id = $dbh->quote( $node_id );
    $line    = $dbh->quote( $line );
    $time    = $dbh->quote( $time );
    $usecs   = $dbh->quote( $usecs );
    
    my $query = "INSERT IGNORE INTO outputBuffer VALUES($node_id,$line,$time,$usecs,NULL)";
    
    eval{
        my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
        $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
        $sth->finish();
        return 1;
    };
    if ($@) {
        return 0;
    };
    return 0;
}


#   return the timestamped output of a node

sub getBufferedOutput {
    my $self                 = shift;
    my @urns                 = ();
    my @times                = ();
    my @usecs                = ();
    my @lines                = ();
    my @rowIndex             = ();
    
    my $dbh = $self->getDatabase();    
    
    my $query = "SELECT * FROM outputBuffer ORDER BY rowindex;";
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
    
    # assign fields to variables
    my $ID;
    my $line;
    my $time;
    my $usecs;
    my $rowindex;
    
    $sth->bind_columns(\$ID, \$line, \$time, \$usecs, \$rowindex);
    while($sth->fetch()) {
       push(@urns,$ID);
       push(@times,$time);
       push(@usecs,$usecs);
       push(@lines,$line);
       push(@rowIndex, $rowindex);
    }
    $sth->finish();
    
    $rowindex = max(@rowIndex);
    
    # delete the lines
    my $rowindex = $dbh->quote($rowindex);
    my $deletequery = "DELETE FROM outputBuffer WHERE rowindex<=$rowindex";
    $sth = $dbh->prepare($deletequery);
    $sth->execute() or die "Couldn't execute statement (query: $deletequery): " . $sth->errstr;
    $sth->finish();
    
    # print "getBufferedOutput::returning " . @urns . " rows\n";
    # for(my $i=0; $i<@urns; $i++) {
    #	print "index: " . @rowIndex[$i] . " urn: " . @urns[$i] . " " . @lines[$i] . "\n";
    # }
    return(\@urns, \@times, \@usecs, \@lines);
}

#############################################################################################################
### FUNCTIONS on TABLE experimentcontroller_output ##########################################################
#############################################################################################################


#   writes a line into the status bar of the experiment controller

sub writeExperimentControllerStatusLine {
    my $self        = shift;
    my $exp_id      = $_[0];
    my $line        = $_[1];
    
    # make connection to database
    my $dbh = $self->getDatabase();    
    # print "got DB handle\n";
    
    # use quote to omit escape character errors
    $line = $dbh->quote( $line );
    $exp_id = $dbh->quote( $exp_id );
    
    # print "getting time\n";
    my $time = &Utils::getTimeInMySQLFormat();
    $time = $dbh->quote( $time );
    
    # print "got time $time\n";
    my $query = "INSERT IGNORE INTO experimentcontroller_output VALUES($exp_id,$line,$time,0,NULL);";
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
}


sub writeExperimentControllerStatusLine_arrays {
    my $self = shift;
    my @exp_ids     = @{$_[0]};
    my @lines       = @{$_[1]};
    my @times       = @{$_[2]}; 
    
    
    my $dbh = $self->getDatabase();   
    my $counter = 0;
    my $query = "INSERT IGNORE INTO experimentcontroller_output VALUES"; 
    
    for(my $i=0; $i<@exp_ids; $i++) {
        my $exp_id    = $dbh->quote(@exp_ids[$i]);
        my $line      = $dbh->quote(@lines[$i]);
        my $time      = $dbh->quote(@times[$i] );
        if ($counter > 0) {
            $query .= ",\n";
        }
        $query .= "($exp_id, $line, $time, 0, NULL)";
        $counter++;
    }
    
    # print "writeExperimentControllerStatusLine_arrays $query\n";
    
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";  
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
}




















#   Resets the table wsn_info:experimentcontroller_output
sub resetExperimentControllerOutputTable {
    my $self = shift;
    my $exp_id  = $_[0];
    if($exp_id eq "") {
        $self->executeDropTruncateQuery("TRUNCATE TABLE experimentcontroller_output");
    }
    else {
        my $dbh = $self->getDatabase();  
        $exp_id  = $dbh->quote( $exp_id );
        $self->executeDropTruncateQuery("DELETE FROM experimentcontroller_output WHERE exp_id=$exp_id");
        $self->executeDropTruncateQuery("DELETE FROM experimentcontroller_output WHERE exp_id=$exp_id");
    }
}


sub getLastRowIndexOfExperimentController {
    my $self        = shift;
    my @rows = $self->executeSelectQuery("SELECT MAX(rowindex) FROM experimentcontroller_output");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    return @row[0];   
}

    
    
#############################################################################################################
### FUNCTIONS on TABLE experiment_info     ##################################################################
#############################################################################################################


sub resetRunningExperiments {
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE running_experiments");
}

sub deleteRunningExperiment {
    my $self = shift;
    my $dbh = $self->getDatabase(); 
    my $exp_id = $dbh->quote($_[0]);
    $self->executeDropTruncateQuery("DELETE FROM running_experiments WHERE exp_id=$exp_id");
}

sub deleteExpiredRunningExperiments {
	my $self = shift;
	$self->executeDropTruncateQuery("DELETE FROM running_experiments WHERE exp_end_time < DATE_ADD(UTC_TIMESTAMP(), INTERVAL 60 SECOND)");
}


#   get all the scheduled experiment id's of the experiment_info table

sub getExperimentIDsInDatabase {
    my $self     = shift;
    my @rows     = $self->executeSelectQuery("SELECT experiment_info.exp_id FROM experiment_info");
    my $numberOfRows = @rows;
    my @ids = ();
    for(my $i=0; $i<$numberOfRows; $i++){
       push(@ids, @rows[$i]->[0]);
    }
    return @ids;
}

sub setExperimentEndTime {
    my $self     = shift;
    my $dbh      = $self->getDatabase(); 
    my $exp_id   = $dbh->quote($_[0]);
    my $time     = $dbh->quote($_[1]);
    my $query = "UPDATE experiment_info SET experiment_info.exp_end_time=$time WHERE experiment_info.exp_id=$exp_id";
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
    $sth->finish();
}

sub experimentEnded {
    my $self     = shift;
    my $dbh      = $self->getDatabase(); 
    my $exp_id   = $dbh->quote($_[0]);
    my @rows     = $self->executeSelectQuery("SELECT experiment_info.exp_id FROM experiment_info WHERE experiment_info.exp_done=1 AND experiment_info.exp_id=$exp_id");
    if(@rows>0){
        return 1;
    }   
    else {
        return 0;
    }
}

sub getUndoneExperimentIDsInDatabase {
    my $self     = shift;
    my @rows     = $self->executeSelectQuery("SELECT experiment_info.exp_id FROM experiment_info WHERE experiment_info.exp_done=0");
    my $numberOfRows = @rows;
    my @ids = ();
    for(my $i=0; $i<$numberOfRows; $i++){
       push(@ids, @rows[$i]->[0]);
    }
    return @ids;
}


sub getEndpointURLofRunningExperiment {
    my $self     = shift;
    my $dbh      = $self->getDatabase(); 
    my $exp_id   = $dbh->quote($_[0]);
    my @rows     = $self->executeSelectQuery("SELECT wsn_instance_url FROM running_experiments WHERE exp_id=$exp_id");
    if(@rows>0) {
    my @row = @{$rows[0]};
        return @row[0]; 
    }
    else {
    	return "NOSUCHEXPERIMENT";
    }
}


#   get all the scheduled experiment id's of the experiment_info table

sub getEmailOfExperimentWithID {
    my $self     = shift;
    my $ID       = $_[0];
    my $dbh = $self->getDatabase(); 
    $ID = $dbh->quote( $ID );
    my @rows     = $self->executeSelectQuery("SELECT experiment_info.exp_user_email FROM experiment_info WHERE experiment_info.exp_id=$ID");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    return @row[0]; 
}

sub getUserUniqueIDOfExperimentWithID {
    my $self     = shift;
    my $dbh = $self->getDatabase(); 
    my $ID = $dbh->quote( $_[0] );
    my @rows     = $self->executeSelectQuery("SELECT experiment_info.exp_user_unique_id FROM experiment_info WHERE experiment_info.exp_id=$ID");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    return @row[0]; 
}

sub getDescriptionOfExperimentWithID(){
    my $self     = shift;
    my $dbh = $self->getDatabase(); 
    my $ID = $dbh->quote( $_[0] );
    my @rows     = $self->executeSelectQuery("SELECT experiment_info.exp_description FROM experiment_info WHERE experiment_info.exp_id=$ID");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    return @row[0]; 
}

#   get all the scheduled experiment id's of the experiment_info table

sub getNameOfExperimentWithID {
    my $self     = shift;
    my $ID       = $_[0];
    my $dbh = $self->getDatabase(); 
    $ID = $dbh->quote( $ID );
    my @rows     = $self->executeSelectQuery("SELECT experiment_info.exp_name FROM experiment_info WHERE experiment_info.exp_id=$ID");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    return @row[0]; 
}



#   get the start of the experiment with the supplied id

sub getStartOfExperimentWithID {
    my $self     = shift;
    my $dbh      = $self->getDatabase(); 
    my $ID       = $dbh->quote( $_[0] );
    my $query    = "SELECT experiment_info.exp_start_time FROM experiment_info WHERE experiment_info.exp_id=$ID";
    # print "getStartOfExperimentWithID: $query\n";
    my @rows     = $self->executeSelectQuery($query);
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    # print "returning: " . @row[0] . "\n"; 
    return @row[0]; 
}

sub getRunsOfExperimentWithID {
    my $self     = shift;
    my $dbh      = $self->getDatabase(); 
    my $ID       = $dbh->quote( $_[0] );
    my $query    = "SELECT experiment_info.exp_runs FROM experiment_info WHERE experiment_info.exp_id=$ID";
    my @rows     = $self->executeSelectQuery($query);
    assert(scalar @rows == 1);
    my @row = @{$rows[0]};
    return @row[0]; 
}


#   get the end of the experiment with the supplied id

sub getEndOfExperimentWithID {
    my $self     = shift;
    my $dbh      = $self->getDatabase(); 
    my $ID       = $dbh->quote($_[0]);
    my $query    = "SELECT experiment_info.exp_end_time FROM experiment_info where experiment_info.exp_id=$ID";
    # print "getEndOfExperimentWithID: $query\n";
    my @rows     = $self->executeSelectQuery($query);
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    # print "returning: " . @row[0] . "\n";     
    return @row[0]; 
}

#   get scheduled experiment which is not already running or return "NOEXPERIMENT"
sub checkAndGetScheduledNonRunningExperiment {
    my $self     = shift;
    my $query    = "SELECT exp_id FROM experiment_info WHERE exp_done=0 AND exp_start_time < UTC_TIMESTAMP() AND exp_end_time > UTC_TIMESTAMP() AND exp_id NOT IN (SELECT exp_id from running_experiments)";
    my @rows     = $self->executeSelectQuery($query);
    if(@rows>0){
    	my @row = @{$rows[0]};
        return @row[0];	
    }
    else {
        return "NOEXPERIMENT";	
    }
}

#   isExperimentRunningWithNode
sub isExperimentRunningWithNode {
    my $self     = shift;
    my $dbh = $self->getDatabase();
    my $node     = $dbh->quote($_[0]);
    my $query    = "SELECT exp_id FROM experiment_info NATURAL JOIN experiment_network  WHERE exp_start_time <= UTC_TIMESTAMP() AND exp_end_time >= UTC_TIMESTAMP() AND node_urn=$node;";
    my @rows     = $self->executeSelectQuery($query);
    if(@rows>0){
        return 1;
    }
    else {
        return 0;
    }
}

sub hasExperimentExpiredWithNode_WithinSeconds {
    my $self     = shift;
    my $dbh      = $self->getDatabase();
    my $node     = $dbh->quote($_[0]);
    my $secs     = $_[1];
    my $query    = "SELECT exp_id, node_urn, exp_start_time, exp_end_time FROM experiment_info NATURAL JOIN experiment_network WHERE node_urn=$node AND UTC_TIMESTAMP() > exp_end_time AND UTC_TIMESTAMP() < DATE_ADD(exp_end_time,INTERVAL $secs SECOND);";
    # print "query=$query\n";
    my @rows     = $self->executeSelectQuery($query);
    if(@rows>0){
        return 1;
    }
    else {
        return 0;
    }
}

#   isExperimentRunningWithNode
sub isExperimentStartingWithNode_WithinSeconds {
    my $self     = shift;
    my $dbh      = $self->getDatabase();
    my $node     = $dbh->quote($_[0]);
    my $secs     = $_[1];
    my $query    = "SELECT exp_id, node_urn, exp_start_time, exp_end_time FROM experiment_info NATURAL JOIN experiment_network WHERE node_urn=$node AND exp_start_time >= UTC_TIMESTAMP() AND exp_start_time <= DATE_ADD(UTC_TIMESTAMP(),INTERVAL $secs SECOND);";
    my @rows     = $self->executeSelectQuery($query);
    if(@rows>0){
        return 1;
    }
    else {
        return 0;
    }
}



# SUCCESSOR OF THE ABOVE "checkAndGetScheduledExperiment"

sub getRunningExperiments {
    my $self     = shift;
    my @rows     = $self->executeSelectQuery("SELECT running_experiments.exp_id FROM running_experiments");
    my $numberOfRows = @rows;
    my @ids = ();
    for(my $i=0; $i<$numberOfRows; $i++){
       push(@ids, @rows[$i]->[0]);
    }
    return @ids;
}

sub insertRunningExperiment {
    my $self           = shift;
    
    my $dbh = $self->getDatabase();
    my $exp_id             = $dbh->quote($_[0]);
    my $wsn_instance_url   = $dbh->quote($_[1]);
    my $exp_start_time     = $dbh->quote($_[2]);
    my $exp_end_time       = $dbh->quote($_[3]);

    my $query = "REPLACE INTO running_experiments VALUES($exp_id, $wsn_instance_url, $exp_start_time, $exp_end_time);";
    my $sth = $dbh->prepare($query) or die "insertRunningExperiment: couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute() or die DBI->errstr;;
    $sth->finish();
    
}


sub getCurrentExperiments {
    my $self   = shift;
    my @expIDs = $self->getUndoneExperimentIDsInDatabase();
    my @currentExperiments = ();
   
    foreach my $expID (@expIDs) {
        my $starttime   = &Utils::parseDate($self->getStartOfExperimentWithID($expID));
        my $endtime     = &Utils::parseDate($self->getEndOfExperimentWithID($expID));
        my $now         = &Utils::get_UTC_time();
        if( $starttime <= $now && $endtime > $now) {
            push(@currentExperiments, $expID);
        }
    }
    return @currentExperiments;
}

sub getCurrentExperimentOfNode {
	my $self     = shift;
    my $dbh = $self->getDatabase();
    my $node_urn = $dbh->quote($_[0]);
    my $i;
    my $query  = "SELECT exp_id FROM experiment_network NATURAL JOIN experiment_info WHERE node_urn=$node_urn AND UTC_TIMESTAMP() > exp_start_time AND exp_end_time > UTC_TIMESTAMP();";
    my @rows   = $self->executeSelectQuery($query);
    if(@rows>0){
        my @row = @{$rows[0]};
        return @row[0]; 
    }
    else {
        return "NOEXPERIMENT";
    }
}

sub getEndOfExperimentWithIDasLinuxTime {
    my $self   = shift;
    my $endtime = &Utils::parseDate($self->getEndOfExperimentWithID($_[0]));
    return $endtime;    
}

sub getStartOfExperimentWithIDasLinuxTime {
    my $self   = shift;
    my $starttime = &Utils::parseDate($self->getStartOfExperimentWithID($_[0]));
    return $starttime;    
}



sub zipSaveExperimentOutputToDatabaseAndSetDone {

    my $self = shift;
    my $exp_id = $_[0];
    my $done   = $_[1];
    my $run    = $_[2];
    
    my $unzipped_file = &NetworkResourcesReader::generateWiseMLFileForExperiment($exp_id, $run);
    
    print "unzipped datafile generated for expid: $exp_id run: $run\n";
    
    
    my $filesize = 0;
    
    if($done==1){
    	
    	my @filenames = ();
    	for(my $k=0; $k<=$run; $k++){
    		my $file = &NetworkResourcesReader::getExperimentWiseMLFileName($exp_id, $k);
    		if( -e $file ) {
    		  print "adding file: $file to be zipped\n";
    		  push(@filenames, $file);
    		}
    		
    	}
    	
    	my $results_zipped_filename = &Utils::correctPath($ENV{'TARWIS_DIRECTORY'}."/results_" . $exp_id . ".zip");
    	&Utils::flushFile($results_zipped_filename);
        &Utils::createZipfile($results_zipped_filename, \@filenames);
        $self->saveExperimentResultsSetDone($exp_id,$results_zipped_filename);
        
        for(my $k=0; $k<=@filenames; $k++){
            if(-e @filenames[$k]){
                unlink(@filenames[$k]);
            }
        }
        
        if(-e $results_zipped_filename){
            $filesize = -s $results_zipped_filename;
            unlink($results_zipped_filename);
        }
    }
    
    return $filesize;
}


sub saveExperimentResultsSetDone {
    
    my $self                    = shift;
    my $dbh                     = $self->getDatabase();
    my $exp_id                  = $dbh->quote($_[0]);
    my $results_zipped_filename = $_[1];
        
    my @rows = $self->executeSelectQuery("SELECT exp_id, exp_name, exp_description, exp_start_time, exp_end_time, exp_user_unique_id, exp_user_email, exp_visibility, exp_runs, exp_result FROM experiment_info WHERE exp_id=$exp_id");
    my @row = @{$rows[0]};
    
    # need to get the entire expermient row to replace it!
    my $exp_id              = @row[0];
    my $exp_name            = @row[1];
    my $exp_description     = @row[2];
    my $exp_start_time      = @row[3];
    my $exp_end_time        = @row[4];
    my $exp_user_unique_id  = @row[5];
    my $exp_user_email      = @row[6];
    my $exp_visibility      = @row[7];
    my $exp_runs            = @row[8];
    my $exp_result          = @row[9];
    
    # set done!
    my $exp_done            = 1;
    
    # zip it!
    open(my $fh, $results_zipped_filename ) or die $!;
    my $var;   
    read( $fh, $var, -s $fh );
    my $query = "REPLACE INTO experiment_info VALUES(?,?,?,?,?,?,?,?,?,?,?);";
    my $sth = $dbh->prepare($query) or die "saveExperimentResults: couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute($exp_id, $exp_name , $exp_description , $exp_start_time , $exp_end_time , $exp_user_unique_id, $exp_user_email, $var, $exp_visibility, $exp_done, $exp_runs) or die DBI->errstr;;
    $sth->finish();
    close($fh);
}

sub insertExperiment {  
    my $self = shift;
    my $dbh = $self->getDatabase();
    
    print "insertExperiment experiment_id: $_[0]\n";
    
    my $exp_id             = $dbh->quote($_[0]);
    my $exp_name           = $dbh->quote($_[1]);
    my $exp_description    = $dbh->quote($_[2]);
    my $exp_start_time     = $dbh->quote($_[3]);
    my $exp_end_time       = $dbh->quote($_[4]);
    my $exp_user_unique_id = $dbh->quote($_[5]);
    my $exp_user_email     = $dbh->quote($_[6]);
    my $exp_visibility     = $dbh->quote($_[7]);
    my $runs               = $dbh->quote($_[8]);

    my $query = "REPLACE INTO experiment_info VALUES($exp_id,$exp_name,$exp_description,$exp_start_time,$exp_end_time,$exp_user_unique_id,$exp_user_email,NULL,$exp_visibility,0, $runs);";
    my $sth = $dbh->prepare($query) or die "insertExperiment: couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute() or die DBI->errstr;;
    $sth->finish();
}


#############################################################################################################
### FUNCTIONS on TABLE experiment_network          ##########################################################
#############################################################################################################


# CREATE TABLE experiment_network ( 
#     exp_id VARCHAR(100) NOT NULL, 
#     node_urn VARCHAR(50), 
#     image_id INT, 
#     PRIMARY KEY(exp_id, node_urn) 
# );

sub insertNodeToExperiment {
     my $self = shift;
     my $dbh = $self->getDatabase();
     my $exp_id   = $dbh->quote($_[0]);
     my $node_urn = $dbh->quote($_[1]);
     my $image_id = $dbh->quote($_[2]);
     
     my $query = "REPLACE INTO experiment_network VALUES($exp_id,$node_urn,$image_id);";
     my $sth = $dbh->prepare($query) or die "insertNodesToExperiment: couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
     $sth->execute() or die DBI->errstr;;
     $sth->finish();
    
}

#   get the pairs <node_urn, image_id> of the experiment with the id supplied as argument

sub getNodeURNImageIDPairs {
    my $self   = shift;
    my $dbh = $self->getDatabase(); 
    my $ID = $dbh->quote( $_[0] );
    my @rows = $self->executeSelectQuery("SELECT node_urn, image_id FROM experiment_info NATURAL JOIN experiment_network WHERE exp_id=$ID");
    
    my $numberOfRows = @rows;
    my @node_urns = ();
    my @images = ();
    
    for(my $i=0; $i<$numberOfRows; $i++){
          push(@node_urns,    @rows[$i]->[0]);
          push(@images, @rows[$i]->[1]);
    }
    
    my ($ref1, $ref2) = &Utils::bubbleSortURNsImages(\@node_urns, \@images);
    @node_urns = @{$ref1};
    @images    = @{$ref2};
    
    return (\@node_urns, \@images); 
}

sub getMaxExperimentID {
    my $self = shift;    
    my $dbh = $self->getDatabase();    
    my $query = "SELECT MAX(CONVERT(exp_id,unsigned)) FROM experiment_info;";
    my @rows = $self->executeSelectQuery($query);
    my @row = @{$rows[0]};
    if(@row[0] eq "NULL"){
    	return 0;
    }    
    else {
    	return @row[0];
    }
}

sub getNodesOfExperiment {
    my $self = shift;
    my $exp_id = $_[0];
    
    my $dbh = $self->getDatabase();    
    $exp_id = $dbh->quote( $exp_id );
    
    my $query = "SELECT * FROM experiment_network WHERE exp_id=$exp_id";
    my @rows = $self->executeSelectQuery($query);
    return @rows;   
}

sub getNodesUrnsOfExperiment {
    my $self = shift;
    my $exp_id = $_[0];
    
    my @nodes = $self->getNodesOfExperiment($exp_id);
    my @urns = ();
    for(my $i=0; $i<@nodes; $i++){
       my @row = @{$nodes[$i]};
       my $node_urn = @row[1]; 
       push(@urns, $node_urn);
    }
    @urns = &Utils::sortURNs(@urns);
    
    return @urns;   
}


#############################################################################################################
### FUNCTIONS on TABLE images                      ##########################################################
#############################################################################################################


#   get the sensor node code image with <id> supplied as argument and return the 
#   filepath!

sub getImageVersionOfImageWithID {
    my $self   = shift;
    my $dbh = $self->getDatabase(); 
    my $ID = $dbh->quote( $_[0] );
    my @rows = $self->executeSelectQuery("SELECT image_version FROM images WHERE image_id=$ID");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    
    return @row[0]; 
    
}


sub getImageNameOfImageWithID {
    my $self   = shift;
    my $dbh = $self->getDatabase(); 
    my $ID = $dbh->quote( $_[0] );
    my @rows = $self->executeSelectQuery("SELECT image_name FROM images WHERE image_id=$ID");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    
    return @row[0]; 
    
}

sub getImagePlatformOfImageWithID {
    my $self   = shift;
    my $dbh = $self->getDatabase(); 
    my $ID = $dbh->quote( $_[0] );
    my @rows = $self->executeSelectQuery("SELECT image_platform FROM images WHERE image_id=$ID");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    
    return @row[0]; 
    
}

sub getImageWithID {
    my $self   = shift;
    my $ID = $_[0];
    my $DIRECTORY = $ENV{'TARWIS_DIRECTORY'};
    my $FILEPATH = $DIRECTORY . "/image_$ID.ihex";
    $FILEPATH = &Utils::correctPath($FILEPATH);
    &Utils::flushFile($FILEPATH);
    
    my $dbh = $self->getDatabase();  
    $ID = $dbh->quote($ID);
    my $query = "SELECT image_file FROM images WHERE image_id=$ID";
      
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    # print "getImageWithID query: $query\n";
    $sth->execute() or die "Couldn't execute statement: " . $sth->errstr . "\n" . "query=$query\n";
 
    
    # save the image 
    open OUTPUT, ">$FILEPATH";
    my $ref = $sth->fetchrow_hashref;
    my $newdata = $$ref{'image_file'};
    print OUTPUT $newdata;
    # print $newdata;
    close OUTPUT;
    $sth->finish;
    
    return $FILEPATH;
}


#   loads the "default image" into the database

sub registerDefaultImage {
	print "registering DEFAULT image...\n";
    my $self            = shift;
    my $path            = $_[0];
    my $image_platform  = $_[1];
    
    my $name = "DefaultImage";
    
    my $image_name        = "DefaultImage";
    my $image_version     = "1.0";
    my $image_description = "default image to be loaded whenever no experiment is scheduled...";
    my $user              = "defaultExperimentUser";
    
    print "inserting the Default Image (cf. /etc/TARWIS.conf)\n";
    print "image_name: $image_name\n";
    
    my $dbh = $self->getDatabase();
    my $var;
   
    my $filesize = -s "$path";
    print "filesize: $filesize\n";
    
    open(my $fh, $path ) or die $!;
    read( $fh, $var, -s $fh );
    my $query = "INSERT IGNORE INTO images VALUES(?,?,?,?,?,?,?,?,?,?,?,?);";
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute("NULL", $image_name , $image_version , $image_platform , $image_description , $name, $filesize, $var, &Utils::getTimeInMySQLFormat(), '0', '1', $user) or die DBI->errstr;;
    $sth->finish();   
    close($fh);
}

sub updateDefaultImage {
	print "updating DEFAULT image...\n";
    my $self            = shift;
    my $path            = $_[0];
    
    my $dbh  = $self->getDatabase();
    my $var;          
    my $filesize = -s "$path";
    print "new file size: $filesize\n";
    open(my $fh, $path ) or die $!;
    read( $fh, $var, -s $fh );
    $filesize = $dbh->quote($filesize);         $var = $dbh->quote($var);         
    my $query = "UPDATE images SET image_file_size=$filesize, image_file=$var, image_visible=1 WHERE image_id='1'";
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute() or die DBI->errstr;;
    $sth->finish();
    close($fh);
}


sub registerOtherDefaultImage {
    my $self            = shift;
    my $path            = $_[0];
    my $type            = $_[1];
    
    print "registering Default image for node type: $type  ...\n";
    
    my $name = "DefaultImage $type";
    
    my $image_name        = $name;
    my $image_version     = "1.0";
    my $image_description = "default image to be loaded for the node type $type whenever no experiment is scheduled...";
    my $user              = "defaultExperimentUser";
    
    print "image_name: $image_name\n";
    
    my $dbh = $self->getDatabase();
    my $var;
   
    my $filesize = -s "$path";
    print "filesize: $filesize\n";
    
    open(my $fh, $path ) or die $!;
    read( $fh, $var, -s $fh );
    my $query = "INSERT IGNORE INTO images VALUES(?,?,?,?,?,?,?,?,?,?,?,?);";
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute("NULL", $image_name , $image_version , $type , $image_description , $name, $filesize, $var, &Utils::getTimeInMySQLFormat(), '0', '1', $user) or die DBI->errstr;;
    $sth->finish();   
    close($fh);
}

sub hasOtherDefaultImage {
    my $self = shift;
    my $type = $_[0];
    my $name = "'DefaultImage $type'";
    my @rows = $self->executeSelectQuery("SELECT image_id from images where image_name=$name");
    if(scalar @rows == 1){
        return 1;   
    }
    else {
        return 0;
    }
}

sub getDefaultImageOfType {
	my $self = shift;
    my $type = $_[0];
    my $name = "'DefaultImage $type'";
    my $query = "SELECT image_id from images where image_name=$name";
    # print "query: $query\n";
    my @rows = $self->executeSelectQuery($query);
    if(scalar @rows == 1){
        return @rows[0]->[0];   
    }
    else {
        return 1;
    }
}

sub updateOtherDefaultImage {
    print "updating DEFAULT image...\n";
    my $self            = shift;
    my $path            = $_[0];
    my $type            = $_[1];
    my $dbh  = $self->getDatabase();
    my $var;          
    my $filesize = -s "$path";
    print "new file size: $filesize\n";
    open(my $fh, $path ) or die $!;
    read( $fh, $var, -s $fh );
    $filesize = $dbh->quote($filesize);         
    $var = $dbh->quote($var);  
    my $name = "'DefaultImage $type'";
    my $query = "UPDATE images SET image_file_size=$filesize, image_file=$var, image_visible=1 WHERE image_name=$name";
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute() or die DBI->errstr;;
    $sth->finish();
    close($fh);
}



sub hasDefaultImage {
    my $self = shift;
    my @rows = $self->executeSelectQuery("SELECT image_id from images where image_name='DefaultImage'");
    if(scalar @rows == 1 && @{$rows[0]} == 1){
        return 1;   
    }
    else {
        return 0;
    }
}



sub getIDOfDefaultImage {
    my $self = shift;
    my @rows = $self->executeSelectQuery("SELECT image_id from images where image_name=\"DefaultImage\"");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    return @row[0];   
}

sub getNodeTypeOfURN {
    my $self = shift;
    my $urn  = $_[0];
    my @rows = $self->executeSelectQuery("SELECT value FROM sensornodes NATURAL JOIN properties WHERE properties.description='nodeType' and sensornodes.node_urn='$urn'");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    return @row[0];   
}


#############################################################################################################
### FUNCTIONS on TABLE properties                  ##########################################################
#############################################################################################################

sub addProperties {
    my $self = shift;
    my $prop_name = $_[0];
    my $prop_type = $_[1];
    my $prop_unit = $_[2];
    my $prop_value = $_[3];
    
    # print "adding property: name=$prop_name type=$prop_type unit=$prop_unit default=$prop_value\n";
    
    
    # make connection to database
    my $dbh = $self->getDatabase();    
    $prop_name = $dbh->quote( $prop_name );
    $prop_type = $dbh->quote( $prop_type );
    $prop_unit = $dbh->quote( $prop_unit );
    $prop_value = $dbh->quote( $prop_value );
    
    my $query = "INSERT IGNORE INTO properties VALUES(NULL, $prop_name, $prop_type, $prop_unit, $prop_value)";
    # print $query. "\n";
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
    
}

sub propertyExists {
    my $self = shift;
    my $prop_name = $_[0];  
    my $dbh = $self->getDatabase();    
    $prop_name = $dbh->quote( $prop_name );
    my @rows = $self->executeSelectQuery("SELECT prop_id FROM properties WHERE description=$prop_name");
    return (scalar @rows == 1 && scalar @{$rows[0]} == 1);      
}


sub getPropertyID {
    my $self = shift;
    my $prop_name = $_[0];  
    # make connection to database
    my $dbh = $self->getDatabase();    
    $prop_name = $dbh->quote( $prop_name );
    my @rows = $self->executeSelectQuery("SELECT prop_id FROM properties WHERE description=$prop_name");
    assert(scalar @rows == 1);
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    return @row[0];  
}


#   method inserts new timestamped value into the table "sensornodes". For each property, node_urn and 
#   time and usecs there is one value. The value should be the most recent value!

sub updateSensorNodeProperty {
    my $self        =  shift;
    my $node_urn    =  $_[0];
    my $prop_id     =  $_[1];
    my $value       =  $_[2];
    my $time        =  $_[3];
    my $usecs       =  $_[4];
    
    # make connection to database
    my $dbh   = $self->getDatabase();    
    
    # use quote to omit escape character errors
    $node_urn = $dbh->quote( $node_urn );
    $prop_id  = $dbh->quote( $prop_id );
    $value    = $dbh->quote( $value );
    $time     = $dbh->quote( $time );
    $usecs    = $dbh->quote( $usecs );
    
    my $query = "REPLACE INTO sensornodes VALUES($node_urn, $prop_id, $value, $time, $usecs)";
    # print $query. "\n";
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
}


#   get the corresponding property number for the property supplied as argument    
                       
sub getPropertyNumber {
    my $self = shift;
    my $property = $_[0];
    my $dbh = $self->getDatabase(); 
    $property = $dbh->quote( $property );
    # retrieves an array of arrays!
    my @rows = $self->executeSelectQuery("SELECT prop_id FROM properties WHERE description=\"$property\"");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    return @row[0];   
}

#############################################################################################################
### FUNCTIONS on TABLE aliveState                  ##########################################################
#############################################################################################################


sub getAliveStateOfNode {
    my $self = shift;
    my $urn = $_[1];
    my $dbh = $self->getDatabase();   
    $urn = $dbh->quote($urn);
    my $query = "SELECT node_state FROM aliveState WHERE node_urn=$urn";
    my @rows = $self->executeSelectQuery($query);
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    
    
    return @row[0];   
}

sub setAliveStateOfNode {
    my $self = shift;
    my $node_urn   = $_[0];
    my $state      = $_[1]; 
    
    my $dbh = $self->getDatabase();   
    $node_urn = $dbh->quote($node_urn);
    $state = $dbh->quote( $state );   
    my $query = "REPLACE INTO aliveState VALUES($node_urn, $state)";
    # print "setAliveStateOfNode: $node_urn, $state\n";
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
    
}

sub setAllAliveStatesTo {
    my $self    = shift;
    my $state   = $_[1]; 
    
    my $dbh = $self->getDatabase();   
    $state = $dbh->quote( $state );   
    my $query = "UPDATE aliveState SET node_state=$state";
    # print "setAliveStateOfNode: $node_urn, $state\n";
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
    
}

#############################################################################################################
### FUNCTIONS on TABLE runtimeCommands             ##########################################################
#############################################################################################################


sub addUpdateCommandEntry {
    my $self = shift;
    my $node_urn    = $_[0];
    my $command     = $_[1]; 
    my $time        = $_[2];
    my $usecs       = $_[3];
    my $experiment_id = $_[4]; 
    my $done          = $_[5]; 
    
    my $dbh = $self->getDatabase();   
    $node_urn = $dbh->quote($node_urn);
    $command = $dbh->quote($command);   
    $time = $dbh->quote($time);   
    $usecs = $dbh->quote($usecs);   
    $experiment_id =  $dbh->quote($experiment_id);
    $done = $dbh->quote($done);   
    
    my $query = "REPLACE INTO runtimeCommands VALUES($node_urn, $command, $time, $usecs, $experiment_id, $done)";
    
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
}

sub getNextUndoneCommands {
    my $self = shift;
    my $dbh = $self->getDatabase();   
    
    my $query = "SELECT * FROM runtimeCommands WHERE done=false AND time <= UTC_TIMESTAMP() ORDER BY time,usecs";
    my @rows = $self->executeSelectQuery($query);
    
    # if more than one commands, give back the first
    if(scalar @rows > 0) {
        return @rows;
    }
    else {
        return (-1);
    }
}

sub resetCommandEntries {
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE runtimeCommands");
}



sub resetCommandEntriesOfNode {
    my $self        = shift;
    my $dbh         = $self->getDatabase();   
    my $node_urn    = $_[0];
    $node_urn       = $dbh->quote($node_urn);
    my $query       = "DELETE FROM runtimeCommands WHERE node_urn=$node_urn";
    $self->executeDropTruncateQuery($query);
}

sub deleteCommandEntriesOfExperiment {
    my $self          = shift;
    my $dbh           = $self->getDatabase();
    my $experiment_id = $_[0]; 
    $experiment_id    = $dbh->quote($experiment_id);
    my $query         = "DELETE FROM runtimeCommands where experiment_id=$experiment_id";
    $self->executeDropTruncateQuery($query);
}



#############################################################################################################
### FUNCTIONS on TABLE automatedCommands                    #################################################
#############################################################################################################




sub insertAutomatedCommand
{
	my $self = shift;
    my $dbh = $self->getDatabase();   
	my $exp_id               = $dbh->quote($_[0]);
	my $node_urn             = $dbh->quote($_[1]);
	my $command              = $dbh->quote($_[2]);
	my $seconds_after_flash  = $dbh->quote($_[3]);
    my $query = "INSERT IGNORE INTO automatedCommands VALUES (NULL," . $exp_id . "," . $node_urn . "," . $command . "," . $seconds_after_flash . ");";
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
    $sth->finish();
}


sub getAutomatedCommands {
    my $self = shift;
    my $dbh = $self->getDatabase();   
    my $expid = $dbh->quote($_[0]);
    
    my $query = "SELECT node_urn, command, seconds_after_flash FROM automatedCommands WHERE exp_id=$expid ORDER BY seconds_after_flash;\n";
    my @rows = $self->executeSelectQuery($query);
    return @rows;
}

sub deleteAutomatedCommands {
    my $self = shift;
    my $dbh = $self->getDatabase();   
    my $expid = $dbh->quote($_[0]);
    
    my $query = "DELETE FROM automatedCommands WHERE exp_id=$expid;\n";
    my @rows = $self->executeDropTruncateQuery($query);
    return @rows;
}



#############################################################################################################
### FUNCTIONS on TABLE asynchronousRequestTypeAndDoneStatus #################################################
#############################################################################################################


sub getNextUnfinishedRequestIDOfType {
    my $self = shift;
    my $request_type = $_[0];
    
    my $dbh = $self->getDatabase();   
    
    $request_type = $dbh->quote($request_type);
    # retrieves an array of arrays!
    
    my $query = "SELECT request_id FROM asynchronousRequestTypeAndDoneStatus WHERE request_type=$request_type AND done=false";
    my @rows = $self->executeSelectQuery($query);
    
    my @requests = ();
    for(my $i=0; $i<@rows; $i++){
       my @singlerow = @{$rows[$i]};
       push(@requests,@singlerow[0]);
    }
    
    
    
    if(@requests > 0) {
        return @requests[0];    
    }
    else {
        return -1;
    }
}


sub setAsynchronousRequestType {
    my $self = shift;
    my $request_id   = $_[0];
    my $request_type = $_[1];
    
    my $dbh = $self->getDatabase();   
    $request_id   = $dbh->quote($request_id);
    $request_type = $dbh->quote($request_type);
    
    my $query = "REPLACE INTO asynchronousRequestTypeAndDoneStatus VALUES($request_id, $request_type, 0)";
    # print "setAsynchronousRequestType: $query\n";
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    
    
    $sth->finish();
}

sub getAsynchronousRequestType {
    my $self = shift;
    my $request_id = $_[0];
    
    my $dbh = $self->getDatabase();   
    $request_id = $dbh->quote($request_id);
    # retrieves an array of arrays!
    
    my $query = "SELECT request_type FROM asynchronousRequestTypeAndDoneStatus WHERE request_id=$request_id";
    
    my @rows = $self->executeSelectQuery($query);
    if(scalar @rows == 1) {
        my @row = @{$rows[0]};
        return @row[0];   
    }
    else {
    	return "UNKNOWNREQUESTID";
    }
}

sub getAsynchronousRequestDoneStatus {
    my $self = shift;
    my $request_id = $_[0];
    my $dbh = $self->getDatabase();   
    $request_id = $dbh->quote($request_id);
    # retrieves an array of arrays!
    
    my $query = "SELECT done FROM asynchronousRequestTypeAndDoneStatus WHERE request_id=$request_id";
    
    my @rows = $self->executeSelectQuery($query);
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
       
    return @row[0];   
}

sub setAsynchronousRequestDoneStatus {
    my $self = shift;
    my $request_id   = $_[0];
    my $request_type = $_[1];
    my $done         = $_[2];
    
    my $dbh = $self->getDatabase();   
    $request_id = $dbh->quote($request_id);
    $request_type = $dbh->quote($request_type);
    $done = $dbh->quote($done);
    
    my $query = "REPLACE INTO asynchronousRequestTypeAndDoneStatus VALUES($request_id, $request_type, $done)";
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    
    
    $sth->finish();
}

sub deleteAsynchronousRequestDoneStatusWithRequestIDs {
    my $self = shift;
    my $request_id   = $_[0];
    my $regexpattern = "#$request_id#"; 
    
    my $dbh = $self->getDatabase();   
    
    
    my $query = "DELETE FROM asynchronousRequestTypeAndDoneStatus WHERE request_id regexp '$regexpattern';";
    
    # print "query=$query\n";
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n"; 
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    
    $sth->finish();
}



#############################################################################################################
### FUNCTIONS on TABLE asynchronousOperation                #################################################
#############################################################################################################



sub getNodesOfRequest {
    my $self = shift;
    my $request_id = $_[0];
    
    my $dbh = $self->getDatabase();   
    
    $request_id = $dbh->quote($request_id);
    # retrieves an array of arrays!
    
    my $query = "SELECT node_urn FROM asynchronousOperation WHERE request_id=$request_id";
    my @rows = $self->executeSelectQuery($query);
    my @nodes = ();
    for(my $i=0; $i<@rows; $i++){
       my @singlerow = @{$rows[$i]};
       push(@nodes,@singlerow[0]);
    }
    @nodes = &Utils::sortURNs(@nodes);
    
    
    
    return @nodes;   
}
                       
sub getAsynchronousStateOfNode {
    my $self = shift;
    my $request_id = $_[0];
    my $urn = $_[1];
    
    my $dbh = $self->getDatabase();   
    
    $urn = $dbh->quote($urn);
    $request_id = $dbh->quote($request_id);
    # retrieves an array of arrays!
    
    my $query = "SELECT node_state FROM asynchronousOperation WHERE node_urn=$urn and request_id=$request_id";
    # print "getAsynchronousStateOfNode: $query\n";
    my @rows = $self->executeSelectQuery($query);
    
    # print "query=" . $query . "\n";
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    assert(scalar @row == 1);
    
    
    
    return @row[0];   
}


sub isThereAnAsynchronousStateSecs {
    my $self = shift;
    my $secs = $_[0];
    
    my $dbh = $self->getDatabase();   
    
    $secs = $dbh->quote($secs);
    
    my $query = "SELECT node_state FROM asynchronousOperation WHERE node_state='0' AND UTC_TIMESTAMP() < DATE_ADD(time,INTERVAL $secs SECOND);"; 
    my @rows = $self->executeSelectQuery($query);
    if(@rows > 0) {
        return 1;
    }   
    return 0;
}


sub isInAsynchronousStateSecs {
    my $self = shift;
    my $node_urn = $_[0];
    my $secs     = $_[1];
    
    my $dbh = $self->getDatabase();   
    
    $node_urn = $dbh->quote($node_urn);
    $secs     = $dbh->quote($secs);
    
    my $query = "SELECT node_state FROM asynchronousOperation WHERE node_state='0' AND node_urn=$node_urn AND UTC_TIMESTAMP() < DATE_ADD(time,INTERVAL $secs SECOND);"; 
    my @rows = $self->executeSelectQuery($query);
    if(@rows > 0) {
    	return 1;
    }   
    return 0;
}

sub setAsynchronousStateOfNode {
    my $self = shift;
    my $request_id = $_[0];
    my $node_urn   = $_[1];
    my $state      = $_[2]; 
    
    my $dbh = $self->getDatabase();   
    $request_id = $dbh->quote($request_id);
    $node_urn = $dbh->quote($node_urn);
    $state = $dbh->quote( $state );
    my $time = &Utils::getTimeInMySQLFormat();
    $time = $dbh->quote( $time );
   
    my $query = "REPLACE INTO asynchronousOperation VALUES($request_id, $node_urn, $state, $time)";
    
    # print "query=" . $query . "\n"; 
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
}

sub setAsynchronousStateOfNode_arrays {
    my $self = shift;
    my @request_ids = @{$_[0]};
    my @node_urns   = @{$_[1]};
    my @states      = @{$_[2]}; 
    
    
    my $dbh = $self->getDatabase();   
    
    my $time = &Utils::getTimeInMySQLFormat();
    $time = $dbh->quote($time);
        
    my $counter = 0;
    
    my $query = "REPLACE INTO asynchronousOperation VALUES"; 
    
    for(my $i=0; $i<@request_ids; $i++) {
        my $request_id = $dbh->quote(@request_ids[$i]);
        my $node_urn   = $dbh->quote(@node_urns[$i]);
        my $state      = $dbh->quote(@states[$i] );
        if ($counter > 0) {
            $query .= ",\n";
        }
        $query .= "($request_id, $node_urn, $state, $time)";
        $counter++;
    }
    
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    
    # print "setAsynchronousStateOfNode_arrays: $query\n";
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
}

sub deleteAsynchronousOperationEntriesOfNode {
    my $self = shift;
    my $node_urn   = $_[1];
    
    my $dbh = $self->getDatabase();   
    $node_urn = $dbh->quote($node_urn);
    
    my $query = "DELETE FROM asynchronousOperation WHERE node_urn=$node_urn";
    
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    
    $sth->finish();
}

sub resetImagesTable {
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE images");
}

sub resetSensorValuesTable {
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE sensorvalues");
}

sub resetAsynchronousOperationTable {
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE asynchronousOperation");
    $self->executeDropTruncateQuery("TRUNCATE TABLE asynchronousRequestTypeAndDoneStatus");
}

sub resetAliveStateTable(){
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE aliveState");
}


sub resetLinkDefaultsTable() {
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE links_setup");
    
}

sub resetPropertiesTable(){
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE properties");
}

sub resetSensorNodesTable(){
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE sensornodes");
}

sub resetSensorNodesPositionsTable(){
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE sensornodes_positions");
}

#   Resets the table wsn_info:sensornode_network

sub resetSensorNodeNetworkTable(){
    my $self = shift;
    $self->executeDropTruncateQuery("TRUNCATE TABLE sensornode_network");
}

sub deleteLinksFromSensorNodeNetworkTable(){
    my $self = shift;
    my $urn = $_[0];
    my $dbh = $self->getDatabase();   
    $urn = $dbh->quote($urn);
    my $query = "DELETE IGNORE from sensornode_network where sensornode_network.node_urn = $urn";
    $self->executeDropTruncateQuery($query);
}


sub resetTables  {
    my $self = shift;
    $self->resetExperimentControllerOutputTable();
    $self->resetSensorNodesOutputTables();
    $self->resetSensorValuesTable();
    $self->resetSensorNodeNetworkTable();
    $self->resetCommandEntries();
}

sub resetOutputTables {
    my $self = shift;
    my $experiment_id = $_[0];
    $self->resetSensorNodesOutputTables($experiment_id);
    $self->resetExperimentControllerOutputTable($experiment_id);
}

sub resetNodeOutputTables {
    my $self = shift;
    my $experiment_id = $_[0];
    $self->resetSensorNodesOutputTables($experiment_id);
}

#   Executes any select query and returns the resulting values in an array consisting in arrays (the rows);

sub executeSelectQuery {
    my $self = shift;
    my $query = $_[0];
    
    # make connection to database
    my $dbh = $self->getDatabase();    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    # print "query: $query\n";
    $sth->execute() or die "Couldn't execute statement: " . $sth->errstr;
    my @allRows = ();
    while (my @data = $sth->fetchrow_array()) {
        my @row = @data;
        push(@allRows, \@row);
    }
    $sth->finish();
    
    return @allRows;
}



#   Executes a "DROP" or "TRUNCATE" query

sub executeDropTruncateQuery {
    my $self = shift;
    my $query = $_[0];
    # make connection to database
    # print "Query: $query\n";
    # print "Owner: " . $self->owner() .  "\n";
    my $dbh = $self->getDatabase();  
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement: " . $sth->errstr;
    $sth->finish(); 
    
}

#   method inserts new timestamped value into the table wsn_info:sensorvalues. This table lists any
#   sensorvalues that have been saved so far...

sub insertSensorNodeValues {
    my $self    = shift;
    my $node_id = $_[0];
    my $prop_id = $_[1];
    my $value   = $_[2];
    my $time    = $_[3];
    my $usecs   = $_[4];
    
    # make connection to database
    my $dbh = $self->getDatabase();    
    
    # use quote to omit escape character errors
    $node_id = $dbh->quote( $node_id );
    $prop_id = $dbh->quote( $prop_id );
    $value = $dbh->quote( $value );
    $time = $dbh->quote( $time );
    $usecs = $dbh->quote( $usecs );
    my $query = "INSERT IGNORE INTO sensorvalues VALUES($node_id,$prop_id,$value,$time,$usecs)";
    # print "$query\n";
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
    $sth->finish();
    
    
}


#   function loads all the setup information into the database read by "getNetwork"

sub setSetupInformation {
    my $self        =  shift;
    my $x           = $_[0];
    my $y           = $_[1];
    my $z           = $_[2];
    
    my $phi         = $_[3];
    my $theta       = $_[4];
    
    my $coordinateType  = $_[5];
    my $description     = $_[6];
    
    # make connection to database
    my $dbh = $self->getDatabase();    
    # use quote to omit escape character errors
    $x = $dbh->quote( $x );
    $y = $dbh->quote( $y );
    $z = $dbh->quote( $z );
    
    $phi = $dbh->quote( $phi );
    
    $theta = $dbh->quote( $theta );
    
    $coordinateType = $dbh->quote( $coordinateType );
    $description = $dbh->quote( $description );
    
    $self->executeDropTruncateQuery("TRUNCATE TABLE setup");
    my $query = "INSERT INTO setup VALUES($x, $y, $z, $phi, $theta, $coordinateType, $description)";
   
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
    
}

#   function loads all the setup information into the database read by "getNetwork"
sub readSetupInformationFromDB {
    my $self =  shift;
    my @rows = $self->executeSelectQuery("SELECT * FROM setup");
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    return @row;
}


sub scaleSensorNodePositions {
	my $self = shift;
    my $min_x = $_[0];
    my $max_x = $_[1];
    my $min_y = $_[2];
    my $max_y = $_[3];
    my $min_z = $_[4];
    my $max_z = $_[5];
    
    print "max_x = $max_x min_x = $min_x\n";
    print "max_y = $max_y min_y = $min_y\n";
    print "max_z = $max_z min_z = $min_z\n";
    
    my $scaleX = 1;
    my $scaleY = 1;
    my $scaleZ = 1;
    
    if($max_x - $min_x > 0) {
        $scaleX = 100.0 / ($max_x - $min_x); 
    }
    if($max_y - $min_y > 0) {
        $scaleY = 100.0 / ($max_y - $min_y); 
    }
    if($max_z - $min_z > 0) {
        $scaleZ = 100.0 / ($max_z - $min_z); 
    }
    print "scaleX = $scaleZ\n";
    print "scaleY = $scaleY\n";
    print "scaleZ = $scaleZ\n";
    
    # make connection to database
    my $dbh = $self->getDatabase();    
    
    my $query = "UPDATE sensornodes_positions SET sensornodes_positions.x = ROUND((sensornodes_positions.original_x - $min_x) * $scaleX);";    
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
    $sth->finish();
    
    my $query = "UPDATE sensornodes_positions SET sensornodes_positions.y = ROUND((sensornodes_positions.original_y - $min_y) * $scaleY);";    
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
    $sth->finish();
    
    my $query = "UPDATE sensornodes_positions SET sensornodes_positions.z = ROUND((sensornodes_positions.original_z - $min_z) * $scaleZ);";    
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
    $sth->finish();
}


#   function adds (or updates, if there is already a node) a node with the given position

sub addUpdateNodePosition {
    my $self = shift;
    my $node_urn = $_[0];
    my $x = $_[1];
    my $y = $_[2];
    my $z = $_[3];
    my $original_x = $_[4];
    my $original_y = $_[5];
    my $original_z = $_[6];
    
    # make connection to database
    my $dbh = $self->getDatabase();    
    # use quote to omit escape character errors
    $node_urn = $dbh->quote( $node_urn );
   
    $x = $dbh->quote( $x );
    $y = $dbh->quote( $y );
    $z = $dbh->quote( $z );
   
    $original_x = $dbh->quote( $original_x );
    $original_y = $dbh->quote( $original_y );
    $original_z = $dbh->quote( $original_z );
    
    
    my $query = "REPLACE INTO sensornodes_positions VALUES($node_urn, $x, $y, $z, $original_x, $original_y, $original_z)";
    # print $query. "\n";
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
    
}

sub getMaxPositionX {
    my $self = shift;    
    my $dbh = $self->getDatabase();    
    my $query = "select MAX(sensornodes_positions.x) FROM sensornodes_positions;";
    my @rows = $self->executeSelectQuery($query);
    my @row = @{$rows[0]};
    if(@row[0] eq "NULL"){
        return 0;
    }    
    else {
        return @row[0];
    }
}

sub getMaxPositionY {
    my $self = shift;    
    my $dbh = $self->getDatabase();    
    my $query = "select MAX(sensornodes_positions.y) FROM sensornodes_positions;";
    my @rows = $self->executeSelectQuery($query);
    my @row = @{$rows[0]};
    if(@row[0] eq "NULL"){
        return 0;
    }    
    else {
        return @row[0];
    }
}


sub getMaxPositionZ {
    my $self = shift;    
    my $dbh = $self->getDatabase();    
    my $query = "select MAX(sensornodes_positions.z) FROM sensornodes_positions;";
    my @rows = $self->executeSelectQuery($query);
    my @row = @{$rows[0]};
    if(@row[0] eq "NULL"){
        return 0;
    }    
    else {
        return @row[0];
    }
}


sub getMinPositionX {
    my $self = shift;    
    my $dbh = $self->getDatabase();    
    my $query = "select MIN(sensornodes_positions.x) FROM sensornodes_positions;";
    my @rows = $self->executeSelectQuery($query);
    my @row = @{$rows[0]};
    if(@row[0] eq "NULL"){
        return 0;
    }    
    else {
        return @row[0];
    }
}

sub getMinPositionY {
    my $self = shift;    
    my $dbh = $self->getDatabase();    
    my $query = "select MIN(sensornodes_positions.y) FROM sensornodes_positions;";
    my @rows = $self->executeSelectQuery($query);
    my @row = @{$rows[0]};
    if(@row[0] eq "NULL"){
        return 0;
    }    
    else {
        return @row[0];
    }
}


sub getMinPositionZ {
    my $self = shift;    
    my $dbh = $self->getDatabase();    
    my $query = "select MIN(sensornodes_positions.z) FROM sensornodes_positions;";
    my @rows = $self->executeSelectQuery($query);
    my @row = @{$rows[0]};
    if(@row[0] eq "NULL"){
        return 0;
    }    
    else {
        return @row[0];
    }
}


sub insertLinkSetting {
     my $self         = shift;
     my $dbh          = $self->getDatabase();
     my $source       = $dbh->quote($_[0]);
     my $target       = $dbh->quote($_[1]);
     my $datatype     = $dbh->quote($_[2]);
     my $unit         = $dbh->quote($_[3]);
     my $defaultvalue = $dbh->quote($_[4]);
     my $encrypted    = $dbh->quote($_[5]);
     my $virtual      = $dbh->quote($_[6]);
     
     my $query = "REPLACE INTO links_setup VALUES($source, $target , $datatype , $unit , $defaultvalue, $encrypted, $virtual);";
     my $sth = $dbh->prepare($query) or die "insertLinkSetting: couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
     $sth->execute() or die DBI->errstr;;
     $sth->finish();
    
}

sub getLinkSettings {
     my $self         = shift;
     my $dbh          = $self->getDatabase();
     my $source       = $dbh->quote($_[0]);
     my $target       = $dbh->quote($_[1]);
     
     my $query = "SELECT  datatype, unit ,defaultvalue, encrypted, virtual FROM links_setup WHERE source=$source AND target=$target;";
     my @rows = $self->executeSelectQuery($query);
     
     my $datatype;
     my $unit;
     my $defaultvalue;
     my $encrypted;
     my $virtual;
     
     if(scalar @rows > 0) {
     	 my @row = @{$rows[0]};
	     $datatype     = @row[0];
	     $unit         = @row[1];
	     $defaultvalue = @row[2];
	     $encrypted    = @row[3];
	     $virtual      = @row[4];
     } 
     return ($datatype , $unit , $defaultvalue, $encrypted, $virtual);   
}

    

sub writeNodeCommentToDB {
    my $self    = shift;
    my $node_id = $_[0];
    my $exp_id  = $_[1];
    my $line    = $_[2];
    $self->writeTimestampedSensorNodeOutputToDB($node_id, $exp_id, $line, &Utils::getTimeInMySQLFormat(), &Utils::getUsecs());
}



#   returns a sorted list of all node URNs registered

sub getAllRegisteredSensorNodes {
    my $self = shift;
    my @rows = $self->executeSelectQuery("SELECT node_urn from sensornodes_positions");
    my @nodes = ();
    for(my $i=0; $i<@rows; $i++){
       my @singlerow = @{$rows[$i]};
       push(@nodes,@singlerow[0]);
    }
    @nodes = &Utils::sortURNs(@nodes);
    return @nodes;   
}

#   write a neighbor entry to the database (which is then used to draw the links in the WebGUI)

sub writeNeighborPairToDatabase {
    my $self        = shift;
    my $node_urn    = $_[0];
    my $neighbor_node_urn = $_[1];
    my $etx         = $_[2];
    my $time        = $_[3];
    my $usecs       = $_[4];
    
    # make connection to database
    my $dbh = $self->getDatabase();    
    
    # use quote to omit escape character errors
    $node_urn           = $dbh->quote( $node_urn );
    $neighbor_node_urn  = $dbh->quote( $neighbor_node_urn );
    $etx                = $dbh->quote( $etx );
    $time               = $dbh->quote( $time );
    $usecs              = $dbh->quote( $usecs );
    
    my $query = "INSERT IGNORE INTO sensornode_network VALUES($node_urn,$neighbor_node_urn,$etx,$time, $usecs);";
    
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";;
    $sth->execute() or die "Couldn't execute statement (query: $$query): " . $sth->errstr;
    $sth->finish();
    
}



sub deleteNeighborEntriesOlderThanNSecs {
    my $self    = shift;
    my $secs    = $_[0];
    my $dbh     = $self->getDatabase();    
    $secs = $dbh->quote( $secs );
    my $query = "DELETE from sensornode_network where time < (SUBTIME((SELECT MAX(time) FROM sensornodes_output), SEC_TO_TIME($secs)))";
    # prepare and execute query
    my $sth = $dbh->prepare($query) or die "Couldn't prepare statement: " . $dbh->errstr . "\n" . "query=$query\n";
    $sth->execute() or die "Couldn't execute statement (query: $query): " . $sth->errstr;
}


sub getDefaultPropertyValueOf {
    my $self = shift;
    my $propertyName = $_[0];
    my $dbh = $self->getDatabase();    
    
    $propertyName = $dbh->quote( $propertyName );   
    my $query = "SELECT defaultvalue FROM properties WHERE description=$propertyName";
    
    my @rows = $self->executeSelectQuery($query);
    if(@rows>0){
        my @row = @{$rows[0]};
        return @row[0]; 
    }
    else {
        return 0;  
    }
}

sub getAllPropertyNames {
    my $self = shift;
    my $propertyName = $_[0];
    my $dbh = $self->getDatabase();    
    
    $propertyName = $dbh->quote( $propertyName );   
    my $query = "SELECT description FROM properties";
    my @rows = $self->executeSelectQuery($query);
    my @properties = ();
    for(my $i=0; $i<@rows; $i++){
       my @singlerow = @{$rows[$i]};
       push(@properties,@singlerow[0]);
    }
    return @properties;   
}




sub getDecimalPropertiesOfSensorNode {
    my $self = shift;
    my $urn = $_[0];
    my $dbh = $self->getDatabase();    
    
    $urn = $dbh->quote( $urn );   
    my $query = "SELECT description FROM properties WHERE type='decimal' AND prop_id IN (SELECT prop_id FROM sensornodes WHERE sensornodes.node_urn=$urn)";
    my @rows = $self->executeSelectQuery($query);
    my @properties = ();
    for(my $i=0; $i<@rows; $i++){
       my @singlerow = @{$rows[$i]};
       push(@properties,@singlerow[0]);
    }
    return @properties;   
}


sub getPropertiesOfSensorNode_withTypeAndUnit {
    my $self = shift;
    my $urn = $_[0];
    my $dbh = $self->getDatabase();    
    
    $urn = $dbh->quote( $urn );
    
    my $query = "SELECT * FROM properties WHERE prop_id in (SELECT prop_id FROM sensornodes WHERE sensornodes.node_urn=$urn)";
    my @rows = $self->executeSelectQuery($query);
    return @rows;   
}


sub getSensorNodePropertyValue {
    my $self = shift;
    my $prop_id = $_[0];
    my $node_urn = $_[1];
    
    my $dbh = $self->getDatabase();    
    $node_urn = $dbh->quote( $node_urn );
    $prop_id = $dbh->quote( $prop_id );
    
    my @rows = $self->executeSelectQuery("SELECT value FROM sensornodes WHERE sensornodes.node_urn = $node_urn AND prop_id=$prop_id"); 
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    
    return @row[0]; 
}

sub getSensorNodePositionData {
    my $self = shift;
    my $node_urn = $_[0];
    
    my $dbh = $self->getDatabase();    
    $node_urn = $dbh->quote( $node_urn );
    
    my @rows = $self->executeSelectQuery("SELECT original_x, original_y, original_z FROM sensornodes_positions WHERE node_urn = $node_urn"); 
    assert(scalar @rows == 1);
    # dereferences first array!
    my @row = @{$rows[0]};
    
    return @row; 
}

sub getExperimentResultZipfile {
    my $self  = shift;
    my $expid = $_[0];
    
    my $dbh = $self->getDatabase();    
    $expid = $dbh->quote($expid );
    
    my @rows = $self->executeSelectQuery("SELECT exp_result from experiment_info where exp_id=$expid"); 
    my @row = @{$rows[0]};
    return @row[0]; 
}


1;