package NetworkResourcesReader;

use POSIX qw(setsid);
use POSIX ();
use LWP;
use HTTP::Request;
use HTTP::Response;
use DBI;
use DBD::mysql;
use Carp::Assert;
use XML::Simple;
use Data::Dumper;
use IO::File;
use IPC::ShareLite;
use HTTP::Request;
use HTTP::Response;
use Env;
use strict; 
use SOAP::DateTime;
use Date::Parse;
use File::Slurp;
use Date::Manip qw(ParseDate UnixDate);
use strict;

require "Utils.pm";
require "DatabaseInterface.pm";

our $DEFAULT_RSSI_UNIT      = "dBm";
our $DEFAULT_RSSI_DATATYPE  = "decimal";
our $DEFAULT_RSSI_DEFAULT   = "0";

our $link_unit;
our $link_default;
our $link_datatype;
our $link_encrypted;
our $link_virtual;

sub loadSetupInformation {
	
	my $network_xml = $_[0];
    
    # read the configuration
    my $conffileOK = &Utils::readConfiguration("/etc/TARWIS.conf");
    if(!$conffileOK) {
       print "NetworkResourcesReader: configuration file not found or non-readable, exiting now!\n";
       exit(1);
    }

    my $records = $network_xml;
    # read the data and check for correct XML syntax
    my $xml = new XML::Simple;
    my $data;
    
    eval{
        $data = $xml->XMLin($records);
    };
    if ($@) {
        print "SessionManagementService::getNetwork returns invalid XML Syntax! (in NetworkResourcesReader::loadSetupInformation)\n";
        print "error: $@\n";
        print "exiting...\n";
        exit(1);
    };
    
    # go through the CONSISTENCY CHECK for the MANDATORY tags first!
    if(!Utils::checkTagExists($data, "setup")) {
       print "SessionManagementService::getNetwork does not define the \"setup\" tag ...\n";
       print "invalid wiseml structure...exiting...\n";
       exit(1);
    }
    if(!Utils::checkTagExists($data->{setup}, "origin") || !Utils::checkTagExists($data->{setup}->{origin}, "x") ||
                                                            !Utils::checkTagExists($data->{setup}->{origin}, "y") ||
                                                            !Utils::checkTagExists($data->{setup}->{origin}, "z"  ||
                                                            !Utils::checkTagExists($data->{setup}->{origin}, "phi"||
                                                            !Utils::checkTagExists($data->{setup}->{origin}, "theta")))) {
       print "SessionManagementService::getNetwork failure: the setup tag does not define the \"origin\" tag or the necessary sub-tags x,y,z,phi,theta ... \n";
       print "invalid wiseml structure...exiting...\n";
       exit(1);
    }
    if(!Utils::checkTagExists($data->{setup}, "coordinateType")) {
       print "SessionManagementService::getNetwork does not define the \"coordinateType\" tag for the \"setup\" section... \n";
       print "invalid wiseml structure...exiting...\n";
       exit(1);
    }
    if(!Utils::checkTagExists($data->{setup}, "description")) {
       print "SessionManagementService::getNetwork does not define the \"description\" tag for \"setup\" section ... \n";
       print "invalid xml structure...exiting...\n";
       exit(1);
    }
    
    # read out the data
    my $xml = new XML::Simple;
    my $data = $xml->XMLin($records, KeyAttr => { attributes => "+for" }); 
    my $origin_position_x_content = getProperty($data->{setup}->{origin}, "x");
    my $origin_position_y_content = getProperty($data->{setup}->{origin}, "y");
    my $origin_position_z_content = getProperty($data->{setup}->{origin}, "z");
    my $phi_content = getProperty($data->{setup}->{origin}, "phi");
    my $theta_content = getProperty($data->{setup}->{origin}, "theta");
    my $coordinateType_content = getProperty($data->{setup}, "coordinateType");
    my $description_content = getProperty($data->{setup}, "description");
    my $database = new DatabaseInterface();
    $database->setSetupInformation( $origin_position_x_content, 
                                             $origin_position_y_content, 
                                             $origin_position_z_content, 
                                             $phi_content,
                                             $theta_content,
                                             $coordinateType_content,
                                             $description_content);
}

#############################################################################

sub addCapabilityEntry {
    my $record = $_[0];

    my $capability_name;
    my $capability_datatype;
    my $capability_unit;
    my $capability_value;
        
    if ( Utils::checkTagExists($record,"name") ) {
        $capability_name = getProperty( $record, "name" );
    }
    else {
        print "ERROR! Each capability definition must have a name attribute...exiting\n";
        exit(1);
    }

    if ( Utils::checkTagExists( $record, "datatype" ) ) {
        $capability_datatype = getProperty( $record, "datatype" );
        # print "capability_datatype = $capability_datatype\n";
    }
    else {
        print "ERROR! Each capability definition must have a type attribute...exiting\n";
        exit(1);
    }
    if ( Utils::checkTagExists( $record, "unit" ) ) {
        $capability_unit = getProperty( $record, "unit" );
        # print "capability_unit = $capability_unit\n";
    }
    if ( Utils::checkTagExists( $record, "default" ) ) {
        $capability_value = getProperty( $record, "default" );
        # print "capability_value = $capability_value\n";
    }
    else {
        print "ERROR! Each capability definition must have a default value...exiting\n";
        exit(1);
    }
    
    my $database = new DatabaseInterface();
    $database->addProperties($capability_name, $capability_datatype, $capability_unit, $capability_value);
}

#############################################################################

sub loadNodesInformation {
	   
    my $network_xml = $_[0];
	
    # parse data from the defaults tag
    loadNodesInformation_defaults($network_xml);    
    
    # parse node data within the rest of the setup section
    loadNodesInformation_instances($network_xml);
}

#############################################################################


sub loadNodesInformation_defaults {
	
	my $network_xml = $_[0];
    # read the configuration
    my $conffileOK = &Utils::readConfiguration("/etc/TARWIS.conf");
    if(!$conffileOK) {
       print "NetworkResourcesReader: configuration file not found or non-readable, exiting now!\n";
       exit(1);
    }
    my $records = $network_xml;
    # read the data and check for correct XML syntax
    my $xml = new XML::Simple;
    my $data = $xml->XMLin($records);
   
    # go through the CONSISTENCY CHECK for the MANDATORY tags first!   
    if(!Utils::checkTagExists($data, "setup")) {
       print "SessionManagementService::getNetwork does not define the \"setup\" tag ...\n";
       print "invalid xml structure...exiting...\n";
       exit(1);
    }
    
    print "parsing the <defaults> tag ... \n";
	if(Utils::checkTagExists($data->{setup}, "defaults")) {
	    
	    if(!Utils::checkTagExists($data->{setup}->{defaults}, "node")) {
	       print "SessionManagementService::getNetwork does not define a \"node\" tag inside the \"defaults\" tag ...\n";
	       exit(1);
	    }
	    
	    if(!Utils::checkTagExists($data->{setup}->{defaults}->{node}, "position")) {
	       print "SessionManagementService::getNetwork does not define a \"position\" tag inside the \"defaults\" tag for \"node\"...\n";
	       print "invalid xml structure...exiting...\n";
	       exit(1);
	    }
	    
	    if(!Utils::checkTagExists($data->{setup}->{defaults}->{node}, "gateway")) {
	       print "SessionManagementService::getNetwork does not define a \"gateway\" tag in the defaults tag for \"node\"...\n";
	       print "invalid xml structure...exiting...\n";
	       exit(1);
	    }
	    
	    if(!Utils::checkTagExists($data->{setup}->{defaults}->{node}, "nodeType")) {
	       print "SessionManagementService::getNetwork does not define a \"nodeType\" tag in the defaults tag for \"node\"...\n";
	       print "invalid xml structure...exiting...\n";
	       exit(1);
	    }
	    
	    if(!Utils::checkTagExists($data->{setup}->{defaults}->{node}, "description")) {
	       print "SessionManagementService::getNetwork does not define a \"description\" tag in the defaults tag for \"node\"...\n";
	       print "invalid xml structure...exiting...\n";
	       exit(1);
	    }
	    
	    # links
	    
	    print "checking for \"links\" tag inside the \"defaults\" tag\n";
	    
	    if(!Utils::checkTagExists($data->{setup}->{defaults}, "link")) {
	       print "SessionManagementService::getNetwork does not define a \"link\" tag inside the \"defaults\" tag ...\n";
	       exit(1);
	    }
	    
	    if(!Utils::checkTagExists($data->{setup}->{defaults}->{link}, "virtual")) {
	       print "SessionManagementService::getNetwork does not define a \"virtual\" tag inside the \"link\" tag inside the \"defaults\" tag ...\n";
	       print "Make sure the \"link\" tag contains a \"virtual\" tag that is either set to \"true\" or \"false\" or TARWIS will abort\n";
	       exit(1);
	    }
	    
	    if(!Utils::checkTagExists($data->{setup}->{defaults}->{link}, "rssi") || !Utils::checkTagExists($data->{setup}->{defaults}->{link}->{rssi}, "unit") 
	                                                                          || !Utils::checkTagExists($data->{setup}->{defaults}->{link}->{rssi}, "default")
	                                                                          || !Utils::checkTagExists($data->{setup}->{defaults}->{link}->{rssi}, "datatype")){  
	       print "SessionManagementService::getNetwork does not define a \"rssi\" tag inside the \"link\" tag inside the \"defaults\" tag ...\n";
	       print "Make sure the tag contains defineds attributes for\n";
	       print "\"unit\", \"default\", \"datatype\"\n";
	       print "or TARWIS will abort\n";
	       exit(1);
	    }
	
	    my $default_position_x_content; 
	    my $default_position_y_content; 
	    my $default_position_z_content;    
	    my $default_position_x_unit;
	    my $default_position_y_unit;
	    my $default_position_z_unit;
	    my $default_gateway_content;
	    my $default_gateway_type;
	    my $default_type_content;
	    my $default_description_content;    
	    my $default_type_type;
	    my $default_description_type;  
	    my $record_base = $data->{setup}->{defaults};
	    $default_position_x_content = getProperty($data->{setup}->{defaults}->{node}->{position}, "x");
	    $default_position_y_content = getProperty($data->{setup}->{defaults}->{node}->{position}, "y");
	    $default_position_z_content = getProperty($data->{setup}->{defaults}->{node}->{position}, "z");
	    $default_gateway_content = getProperty($data->{setup}->{defaults}->{node}, "gateway");
	    $default_type_content = getProperty($data->{setup}->{defaults}->{node}, "nodeType");
	    $default_description_content = getProperty($data->{setup}->{defaults}->{node}, "description");
	    
	    # print "default position_x_content: $default_position_x_content\n";
	    # print "default position_y_content: $default_position_y_content\n";
	    # print "default position_z_content: $default_position_z_content\n";
	    # print "default gateway_content: $default_gateway_content\n";
	    # print "default type_content: $default_type_content\n";
	    # print "default description_content: $default_description_content\n";
	    
	    print "#####################################################################\n";
	    print "parsing the mandatory attributes for sensor nodes and saving them in the DB...\n";
	    
	    my @propertyNames = ("gateway", "nodeType", "description" );
	    my @propertyTypes = ("", "", "");
	    my @propertyUnits = ("", "", "");
	    my @propertyValues = ($default_gateway_content, $default_type_content, $default_description_content);
	     
	    
	    my $database = new DatabaseInterface();
	    
	    for(my $i=0; $i < @propertyNames; $i++){
	       $database->addProperties(@propertyNames[$i],@propertyTypes[$i], @propertyUnits[$i], @propertyValues[$i]);
	    }
	    
	    print "#####################################################################\n";
	    print "parsing any additional (non-mandatory) capabilities (if present) ...\n";
	    
	     
	    # need to fold the capability entries on attributes 
	    $xml = new XML::Simple;
	    $data = $xml->XMLin($records, KeyAttr => { attributes => "+id" });
	    # is there only one element? 
	    if(ref($data->{setup}->{defaults}->{node}->{capability}) eq 'HASH' ){
	        my $record = $data->{setup}->{defaults}->{node}->{capability};
	        addCapabilityEntry($record);    
	    }
	    # or none or more? 
	    else {
	        foreach my $record (@{$data->{setup}->{defaults}->{node}->{capability}})  {
	            addCapabilityEntry($record);
	        }
	    }
	    print "done with capabilities ...\n";
	    print "#####################################################################\n";
	    print "parsing the mandatory link settings ...\n";
	    
	
	    $link_datatype = getProperty($data->{setup}->{defaults}->{link}->{rssi}, "datatype");
	    $link_unit = $data->{setup}->{defaults}->{link}->{rssi}{unit};
	    $link_default = $data->{setup}->{defaults}->{link}->{rssi}{default};
	    $link_encrypted = $data->{setup}->{defaults}->{link}->{encrypted};
	    $link_virtual = $data->{setup}->{defaults}->{link}->{virtual};
	    $database->insertLinkSetting("default", "default", $link_datatype, $link_unit, $link_default, $link_encrypted, $link_virtual);
	    print "done with the <defaults> tag...\n";
    }
    else {
        print "no <defaults> tag present ...\n";	
    }
}

#############################################################################


sub loadNodesInformation_instances {
    
    my $network_xml = $_[0]; 
    my $database = new DatabaseInterface();
    
    print "#####################################################################\n";
    print "parsing the instantiations of the node resources....\n";
    
    my $conffileOK = &Utils::readConfiguration("/etc/TARWIS.conf");
    if(!$conffileOK) {
       print "NetworkResourcesReader: configuration file not found or non-readable, exiting now!\n";
       exit(1);
    }

    my $records = $network_xml;
    
    # read the data and check for correct XML syntax
    my $xml = new XML::Simple;
    eval{
        my $data = $xml->XMLin($records);
    };
    if ($@){
        print "SessionManagementService::getNetwork returns invalid XML Syntax!  (in NetworkResourcesReader::loadNodesInformation)\n";
        print "error: $@\n";
        print "exiting...\n";
        exit(1);
    };
    
    my $node_urn;
   
    # need to fold the capability entries on attributes 
     
    $xml = new XML::Simple;
    my $data = $xml->XMLin($records, KeyAttr => { attributes => "+id" });
       
    
    my @node_references = ();
    if(ref($data->{setup}->{node}) eq 'HASH' ){
       push(@node_references, $data->{setup}->{node});
    }
    else {
      foreach my $node (@{$data->{setup}->{node}})  {
         push(@node_references, $node);
      }
    }
    
    if(@node_references==0) {
        print "SessionManagementService::getNetwork returns a file with 0 nodes (in NetworkResourcesReader::loadNodesInformation)\n";
        print "need at least 1 node to start TARWIS...\n";
        print "exiting...\n";
        exit(1);    
    }
    
       
    # get all the instantiated nodes definitions
    
    for(my $k=0; $k<@node_references && $k< int($ENV{'MAX_DISPLAYABLE_NODES'}); $k++) {
       my $record = @node_references[$k];
       if(Utils::checkTagExists($record, "id" )){
          $node_urn = getProperty($record,"id"); 
       }
       else{
          print "ERROR! Each node instantiation entry must have a <id> tag!\n";
          exit(1);
       }
       
       my $default_position_x_content = 0;
       my $default_position_y_content = 0;
       my $default_position_z_content = 0;
       
       if( Utils::checkTagExists($data->{setup}, "defaults") && Utils::checkTagExists($data->{setup}->{defaults}, "node")
                                                             && Utils::checkTagExists($data->{setup}->{defaults}->{node}, "position")
                                                             && Utils::checkTagExists($data->{setup}->{defaults}->{node}->{position}, "x") 
                                                             && Utils::checkTagExists($data->{setup}->{defaults}->{node}->{position}, "y")
                                                             && Utils::checkTagExists($data->{setup}->{defaults}->{node}->{position}, "z") 
                                                             ){
           $default_position_x_content = getProperty($data->{setup}->{defaults}->{node}->{position}, "x");
           $default_position_y_content = getProperty($data->{setup}->{defaults}->{node}->{position}, "y");
           $default_position_z_content = getProperty($data->{setup}->{defaults}->{node}->{position}, "z");
       }
       
       # for the previously defined attributes, take the default value!
       my $position_x_content  = $default_position_x_content; 
       my $position_y_content  = $default_position_y_content; 
       my $position_z_content  = $default_position_z_content;    
       my $gateway_content;    
       my $type_content;        
       my $description_content; 
       
       if(   Utils::checkTagExists($record, "position")     && 
             Utils::checkTagExists($record->{position},"x") &&
             Utils::checkTagExists($record->{position},"y") &&
             Utils::checkTagExists($record->{position},"z") ){
          $position_x_content = getProperty($record->{position}, "x");
          $position_y_content = getProperty($record->{position}, "y");
          $position_z_content = getProperty($record->{position}, "z");             
       }
       
       $database->addUpdateNodePosition($node_urn,         $position_x_content,
                                                           $position_y_content,
                                                           $position_z_content, 
                                                           $position_x_content, 
                                                           $position_y_content, 
                                                           $position_z_content);
       
       $gateway_content     = $database->getDefaultPropertyValueOf("gateway");
       $type_content        = $database->getDefaultPropertyValueOf("nodeType");
       $description_content = $database->getDefaultPropertyValueOf("description"); 
       
       
       # getting all the "node additional information"
       
       if(Utils::checkTagExists($record, "gateway")){
          $gateway_content = getProperty($record, "gateway");      
       }
       if(Utils::checkTagExists($record, "nodeType")){
          $type_content = getProperty($record, "nodeType");
       }
       if(Utils::checkTagExists($record, "description")){
          $description_content = getProperty($record, "description");     
       }
       
       my @propertyNames = ("gateway", "nodeType", "description" ) ;
       my @propertyTypes = ("", "", "");
       my @propertyUnits = ("", "", "");
       my @propertyValues = ($gateway_content, $type_content, $description_content);
    
       # get the tag information in the instantiated nodes!
       for(my $i=0; $i < @propertyNames; $i++){
       	   if(!$database->propertyExists(@propertyNames[$i])) {
                $database->addProperties(@propertyNames[$i], "","","");
       	   }
       	   my $prop_id = $database->getPropertyID(@propertyNames[$i]);
           my $value = @propertyValues[$i];
           $database->updateSensorNodeProperty($node_urn, $prop_id, $value, &Utils::getTimeInMySQLFormat(), 0);
       }
    
       my @props = $database->getAllPropertyNames();   
       # adding the defined capabilities with the default values!
    
       for(my $i=0; $i < @props; $i++){
           if(!&Utils::containsElement(\@propertyNames, @props[$i])){
               my $capability_name = @props[$i];
               my $capability_value = $database->getDefaultPropertyValueOf($capability_name);
               my $updatedValue = getCapabilitySettingInNodeInstantiation($record, $capability_name, "default");
               if($updatedValue ne undef) {
                    # print "found updated value: $capability_name: $updatedValue\n";
                    $capability_value = $updatedValue;
               }
               my $prop_id = $database->getPropertyID($capability_name);
               # print "setting value: $capability_name: $capability_value\n";
               $database->updateSensorNodeProperty($node_urn, $prop_id, $capability_value, &Utils::getTimeInMySQLFormat(), 0);
           }
       }      
    }
    
    if( ! $ENV{'SCALE_NODES_IN_GUI'} eq "" && $ENV{'SCALE_NODES_IN_GUI'} == 1 ) { 
    
	    print "#####################################################################\n";
	    print "finding scale factors\n";
	    
	    my $max_x = $database->getMaxPositionX();
	    my $max_y = $database->getMaxPositionY();
	    my $max_z = $database->getMaxPositionZ();
	    
	    my $min_x = $database->getMinPositionX();
	    my $min_y = $database->getMinPositionY();
	    my $min_z = $database->getMinPositionZ();
	    
	    $database->scaleSensorNodePositions($min_x,$max_x, $min_y,$max_y, $min_z,$max_z);
	}
    
    print "#####################################################################\n";
    
    
    
    foreach my $record (@node_references)  {   
       if(Utils::checkTagExists($record, "id" )){
          $node_urn = getProperty($record,"id"); 
       }
       else{
          print "ERROR! Each node instantiation entry must have a <id> tag!\n";
          exit(1);
       }
       # now every node's capabilities is set according to a) default values or b) overriden by specific capability values 
       # print "checking additional capabilities of node: " . $node_urn . "\n";
       my @capability_references = ();
       if(ref($record->{capability}) eq 'HASH' ){
          push(@capability_references, $record->{capability});
       }
       else {
         foreach my $capability (@{$record->{capability}})  {
            push(@capability_references, $capability);
         }
       }
       
       foreach my $capability (@capability_references)  {
          
          my $capability_name;
          my $capability_datatype;
          my $capability_unit;
          my $capability_value;
            
          if(Utils::checkTagExists($capability, "name")){
             $capability_name = getProperty($capability, "name");
          }
          else {
             print "ERROR! Each capability definition must have a \"name\" attribute...exiting\n";
             exit(1);
          }
          if(Utils::checkTagExists($capability, "datatype")){
             $capability_datatype = getProperty($capability, "datatype");
             if($capability_datatype ne "integer" &&  $capability_datatype ne "decimal"){
                print "ERROR! Capability definition for \"datatype\" attribute must be \"integer\" or \"decimal\"\n";
                exit(1);
             }
          }
          else {
             print "ERROR! Each capability definition must have a \"datatype\" attribute...exiting\n";
             exit(1);
          }
          if(Utils::checkTagExists($capability, "unit")){
             $capability_unit = getProperty($capability, "unit");
             
             if($capability_unit eq "meters"     ||  $capability_unit eq "centimeters"  ||
                $capability_unit eq "radians"    ||  $capability_unit eq "degrees"      ||
                $capability_unit eq "seconds"    ||  $capability_unit eq "milliseconds" ||
                $capability_unit eq "dBm"        ||  $capability_unit eq "kelvin"       ||
                $capability_unit eq "percentage" ||  $capability_unit eq "lux"          ||
                $capability_unit eq "raw" ){
             }
             else {
                print "ERROR! Capability definition for \"unit\" attribute unknown\n";
                exit(1);
             }
             
          }
          if(Utils::checkTagExists($capability, "default")){
             $capability_value = getProperty($capability, "default");
          }
          else {
             print "ERROR! Each capability definition must have a \"default\" value...exiting\n";
             exit(1);
          }
          
          if(!$database->propertyExists($capability_name)){
            $database->addProperties($capability_name, $capability_datatype, $capability_unit, $capability_value);
          }
          # print "adding capability $capability_name with unit=$capability_unit value=$capability_value to node $node_urn\n";
          my $prop_id = $database->getPropertyID($capability_name);
          $database->updateSensorNodeProperty($node_urn, $prop_id, $capability_value, &Utils::getTimeInMySQLFormat(), 0);               
       }     
    }
    print "#####################################################################\n";
    
    print "adding link settings\n";
    
    my ($datatype_default , $unit_default , $defaultvalue_default, $encrypted_default, $virtual_default) = $database->getLinkSettings("default", "default");
  
    if($datatype_default eq "" ) {
    	print "no default link settings ... skip this step \n";
    	return;
    } 
  
    my @urns = $database->getAllRegisteredSensorNodes();  
    # for each node pair, generate a link tag with default values
    for(my $i=0; $i<@urns; $i++){
        for(my $j=0; $j<@urns; $j++){
            if($i != $j) {
                $database->insertLinkSetting(@urns[$i], @urns[$j], $link_datatype, $link_unit, $link_default, $link_encrypted, $link_virtual);
            }
        }
    }
    
    if(Utils::checkTagExists($data->{setup},"link")){
    
        print "link instantiations found...\n";
        my @link_references = ();
        if(ref($data->{setup}->{link}) eq 'HASH' ){
           push(@link_references, $data->{setup}->{link});
        }
        else {
          foreach my $link (@{$data->{setup}->{link}})  {
             push(@link_references, $link);
          }
        }
        
        print "links found: " . @link_references . "\n";
    
        foreach my $link (@link_references){
            
            my $source = $link->{"source"};
            my $target = $link->{"target"};
            
            my $unit      = $link->{rssi}{unit};
            my $default   = $link->{rssi}{default};
            my $datatype  = $link->{rssi}{datatype};
            my $encrypted = $link->{encrypted};
            my $virtual   = $link->{virtual};
            
            # print "source: $source\n";
            # print "target: $target\n";
            # print "unit: $unit\n";
            # print "datatype: $datatype\n";
            # print "encrypted: $encrypted\n";
            # print "virtual: $virtual\n";
            
            $unit      = ( $unit eq "" ? $link_unit : $unit );
            $default   = ( $default eq "" ? $link_default : $default );
            $datatype  = ( $datatype eq "" ? $link_datatype : $datatype );
            $encrypted = ( $encrypted eq "" ? $link_encrypted : $encrypted );
            $virtual   = ( $virtual eq "" ? $link_virtual : $virtual );
            
            
            if(! $source eq "" && ! $target eq "") {
               $database->insertLinkSetting($source, $target, $datatype, $unit, $default, $encrypted, $virtual);
            } 
            else {
               print "link instantiations must have source and target attributes!\n";
               exit(1); 
            }
        }
    
    }

    print "#####################################################################\n";
}

sub getCapabilitySettingInNodeInstantiation {
   my $records = $_[0];
   my $capability_name = $_[1];
   my $capability_setting = $_[2];
   
   #is there only one element? 
   if(ref($records->{capability}) eq 'HASH' ){
       my $foundProperty = &Utils::trimchomp(getProperty($records->{capability}, "name"));
       if( $foundProperty eq  $capability_name) {
           return &Utils::trimchomp(getProperty($records->{capability}, $capability_setting));
       }
   }
   # or none or more? 
   else {
      foreach my $record (@{$records->{capability}})  {
        my $foundProperty = &Utils::trimchomp(getProperty($record,"name"));
        if($foundProperty eq $capability_name) {
           return &Utils::trimchomp(getProperty($record, $capability_setting));
        }
      }
   }
   return undef;
}


#############################################################################
# quick check if value is of type string or scalar
sub isString {
    my $ref = shift;
    return 0 unless ref $ref;
    if ( $ref =~ /^SCALAR/ ) { return 1; } 
    else { return 0; }
}

sub getProperty {
    my $ref = $_[0];
    my $propname = $_[1];
    
    if(!Utils::checkTagExists($ref, $propname)){
       print "ERROR!" . "$propname is not defined!...exiting...\n";
       exit(0);  
    }
    
    my $value;
    eval{
        $value = &Utils::trimchomp($ref->{$propname});
        if(!isString($value)) {
           eval{
              $value = &Utils::trimchomp($ref->{$propname}->{content});
           };
        }
    };
    if ($@){
        print "ERROR! Wrong syntax in node entry!\n";
        print Dumper($ref) . "\n";
        exit(0);
    };
    $value = &Utils::removeQuotes($value);
    $value = &Utils::trimchomp($value);
    return $value;
}


###########################################################################################################################
######################################### WISEML FILE GENERATION  #########################################################
###########################################################################################################################


sub scanDump {
    my $dumpcount = 0; $dumpcount++ while $_[0] =~ m/\{/g;
    return $dumpcount;
}

sub addUTCTimeZone {
    return $_[0] . "Z";
}

sub generateSOAPDateTimeObject {
   my $myTime = $_[0];
   my $date = ParseDate($myTime);
   my $datestr = UnixDate($date, "%a %b %e %H:%M:%S %z %Y"); 
   my $soap_datetime = ConvertDate($datestr);
   return $soap_datetime;
}

sub generateOriginTag {
    my $x     = $_[0];
    my $y     = $_[1];
    my $z     = $_[2];
    my $phi   = $_[3];
    my $theta = $_[4];
    
    return "<origin>\n\t<x>$x</x>\n\t<y>$y</y>\n\t<z>$z</z>\n\t<phi>$phi</phi>\n\t<theta>$theta</theta>\n</origin>\n";
}

sub generateCoordinateTypeTag {
    my $coordinateType  = $_[0];
    return "<coordinateType>$coordinateType</coordinateType>\n";
}

sub generateDescriptionTag {
    my $description  = $_[0];
    return "<description>$description</description>\n";
}

sub generateTimeInfoTag {
    my $start  = $_[0];
    my $end    = $_[1];
    my $unit   = $_[2];
    return "<timeinfo>\n\t<start>$start</start>\n\t<end>$end</end>\n\t<unit>$unit</unit>\n</timeinfo>\n";
}


sub generateInitialSetupSection {
 
    my $exp_id = $_[0];
    my $database = new DatabaseInterface();
    my @data = $database->readSetupInformationFromDB();
    my $origin_x        = @data[0];
    my $origin_y        = @data[1];
    my $origin_z        = @data[2];
    my $phi             = @data[3];
    my $theta           = @data[4];
    my $coordinateType  = @data[5];
    my $description     = @data[6];
    
    my $origin_xml_information = "";
    
    my $originTag = generateOriginTag($origin_x, $origin_y, $origin_z, $phi, $theta);
    my $starttime = addUTCTimeZone(generateSOAPDateTimeObject($database->getStartOfExperimentWithID($exp_id)));
    my $endtime   = addUTCTimeZone(generateSOAPDateTimeObject($database->getEndOfExperimentWithID($exp_id)));
    my $duration_unit = "seconds";
 
    my $timeinfoTag = generateTimeInfoTag($starttime, $endtime, $duration_unit);
                            
    my $coordinateTypeTag = generateCoordinateTypeTag($coordinateType);         
    my $descriptionTag = generateDescriptionTag($description);
 
 
    $origin_xml_information.=$originTag;
    $origin_xml_information.=$timeinfoTag;
    $origin_xml_information.=$coordinateTypeTag;
    $origin_xml_information.=$descriptionTag;
    
     
    return $origin_xml_information;
}

sub generateStaticSetupAndNodeTags {
    
    my $exp_id=$_[0];
    my $database = new DatabaseInterface();
    #append everythin to the document
    my $preamble = "<setup>\n";    
    
    my $setupBlock = generateInitialSetupSection($exp_id);
    
    my $document = $preamble . $setupBlock;
    
    my $defaults_section = generateDefaultsSection($exp_id);
    
    $document .= $defaults_section;
    
    my @nodes = $database->getNodesOfExperiment($exp_id);
    my @urns = ();
    for(my $i=0; $i<@nodes; $i++){
       my @row = @{$nodes[$i]};
       my $node_urn = @row[1]; 
       push(@urns, $node_urn);
       my $node_tag = generateNodeTag($node_urn);
       $document = $document . $node_tag;
    }
    
     
    my $links_instances = generateLinksInstances($exp_id); 
     
    my $postamble = "</setup>\n";
    $document .= $postamble;
    
    return $document;    
}


sub generateLinksInstances {
    
    my $exp_id = $_[0];
    my $database = new DatabaseInterface();
    my @urns = $database->getNodesUrnsOfExperiment($exp_id);
    my ($datatype_default , $unit_default , $defaultvalue_default, $encrypted_default, $virtual_default) = $database->getLinkSettings("default", "default");
    
    my $links_instances = "";
    
    if(! $datatype_default eq "") {
	    # only generate link instances if the values differ from the default section
	    
	    for(my $i=0; $i<@urns; $i++){
	        for(my $j=0; $j<@urns; $j++){
	            if($i != $j) {
	                my ($datatype , $unit , $defaultvalue, $encrypted, $virtual) = $database->getLinkSettings(@urns[$i], @urns[$j]);
	                if( !$datatype eq $datatype_default || !$unit eq $unit_default || !$defaultvalue eq $defaultvalue_default || !$encrypted eq $encrypted_default || !$virtual eq $virtual_default ) {
	                    $links_instances.= generateLinkInstanceTag(@urns[$i], @urns[$j], $datatype, $unit, $defaultvalue, $encrypted, $virtual );
	                }
	            }
	        }
	    }
    }
    
    return $links_instances;
}

sub generateLinkInstanceTag {
    my $source    = $_[0];
    my $target    = $_[1];
    my $datatype  = $_[2];
    my $unit      = $_[3];
    my $default   = $_[4];
    my $encrypted = $_[5];
    my $virtual   = $_[6];
    
    my $linkTag  = "<link source=\"$source\" target=\"$target\"><virtual>$virtual</virtual><encrypted>$encrypted</encrypted><rssi datatype=\"$datatype\" unit=\"$unit\" default=\"$default\" /></link>\n";
    return $linkTag;
}


sub generateDefaultLinkTag {
    
    my $datatype  = $_[0];
    my $unit      = $_[1];
    my $default   = $_[2];
    my $encrypted = $_[3];
    my $virtual   = $_[4];
    
    my $linkTag  = "<link>\n";
    $linkTag    .=     "\t<encrypted>$encrypted</encrypted>\n";
    $linkTag    .=     "\t<virtual>$virtual</virtual>\n";
    $linkTag    .=     "\t<rssi datatype=\"$datatype\" unit=\"$unit\" default=\"$default\"/>\n";
    $linkTag    .= "</link>\n";
    return $linkTag;
}

sub getExperimentWiseMLFileName {
	my $exp_id = $_[0];
    my $run    = $_[1];
    my $database = new DatabaseInterface();
    
	my $start = &Utils::replaceSomeChars(&Utils::trimchomp($database->getStartOfExperimentWithID($exp_id)));
    my $end   = &Utils::replaceSomeChars(&Utils::trimchomp($database->getEndOfExperimentWithID($exp_id)));
    my $name  = &Utils::replaceSomeChars(&Utils::trimchomp($database->getNameOfExperimentWithID($exp_id)));
    $start =~ s/\ /\_/;
    $end =~ s/\ /\_/;
    my $expFilename = "experiment#("; 
    if(length($name<200)) {
    	$expFilename.="name=$name,";
    }	
    $expFilename= $expFilename . "id=".$exp_id.",run=".$run.",start=".$start.",end=".$end.").wiseml";
    my $unzipped_file_name = &Utils::correctPath($ENV{'TARWIS_DIRECTORY'} . "/" . $expFilename);
    return $unzipped_file_name;
}


sub generateWiseMLFileForExperiment {
 
    my $exp_id = $_[0];
    my $run    = $_[1];
    
    print "generateWiseMLFileForExperiment expid=$exp_id run=$run\n";

    my $unzipped_file = getExperimentWiseMLFileName($exp_id, $run);
    
    &Utils::flushFile($unzipped_file);
    
    my $preamble = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" . 
                   "<wiseml version=\"1.0\" xmlns=\"http://wisebed.eu/ns/wiseml/1.0\">\n";
    my $postamble = "</wiseml>";
    my $static_section = generateStaticSetupAndNodeTags($exp_id);
    
    # print static section to file!
    &Utils::appendStringToFile($unzipped_file, $preamble);
    &Utils::appendStringToFile($unzipped_file, $static_section);
    # export all output to the file specified as argument
    appendSensornodesOutputToXMLFile($unzipped_file, $exp_id);
    # append the postamble to file
    &Utils::appendStringToFile($unzipped_file, $postamble);
    
    return $unzipped_file;
}

sub generateDefaultsSection {
   my $exp_id = $_[0];
   my $database = new DatabaseInterface();
   my @urns = $database->getNodesUrnsOfExperiment($exp_id);

   my ($datatype , $unit , $defaultvalue, $encrypted, $virtual) = $database->getLinkSettings("default", "default");
    
    
   my $defaults = "<defaults>\n";
   my $link_tag = generateDefaultLinkTag($datatype , $unit , $defaultvalue, $encrypted, $virtual);
   $defaults   .= $link_tag;
   $defaults   .= "</defaults>\n";
}   

our %last_x = ();
our %last_y = ();
our %last_z = ();       

sub generateNodeOutputTag {
    my $urn = $_[0];
    my $data = $_[1];
    my $db_handle = $_[2];
    my $tag = "";
    my $nodeStartTag  = "<node id=\"$urn\">\n";
    $tag.=$nodeStartTag;
           
           
    if( ! $db_handle eq "" ){       
        # fetch the position data
        my @positionData = $db_handle->getSensorNodePositionData($urn);
        my $x = @positionData[0];
        my $y = @positionData[1];
        my $z = @positionData[2];
        
        if($last_x{$urn} != $x || $last_y{$urn} != $y || $last_z{$urn} != $z) {
             # save the position data in the last_x last_y last_z hash!
            $last_x{$urn} = $x;
            $last_y{$urn} = $y;
            $last_z{$urn} = $z;     
            my $positionTag  = "<position>\n\t<x>$x</x>\n\t<y>$y</y>\n\t<z>$z</z>\n</position>\n";
            $tag.=$positionTag;
        }
    }
    # add the tag!
    $tag.=$data;
           
    my $nodeEndTag = "</node>\n";
    $tag.=$nodeEndTag;
    return $tag;
}
    

sub appendSensornodesOutputToXMLFile {
    my $filepath = $_[0];
    my $exp_id   = $_[1];
    print "appendSensornodesOutputToXMLFile: saving experiment $exp_id to $filepath\n";
    
    my $database = new DatabaseInterface();
    my $experiment_start_time = $database->getStartOfExperimentWithID($exp_id);
    my $experiment_end_time   = $database->getEndOfExperimentWithID($exp_id);
    my $starttime_secs = &Utils::parseDate($experiment_start_time);
    my $endtime_secs = &Utils::parseDate($experiment_end_time);
       
    my @allRows = $database->getAllTextOutputOfNodes_orderedByTime($exp_id);
    my $numberOfRows = @allRows;
    
    &Utils::appendStringToFile($filepath, "<trace id=\"experiment_" . $exp_id . "_output\">\n");
   
    
   
    for(my $i=0; $i<$numberOfRows; $i++){
       my $urn        = @allRows[$i]->[0];
       my $textMessage  = @allRows[$i]->[1];
       my $urn_prefix = &Utils::getURNPrefix($urn);
       
       # calculate timestamp
       my $time = &Utils::removeBadEncoding(@allRows[$i]->[2]);
       my $time_secs = &Utils::parseDate($time);
       my $timestamp = $time_secs - $starttime_secs;
       my $usecs = &Utils::removeBadEncoding(@allRows[$i]->[3]);   
       my $fraction  = $usecs / 1000000.0; 
       $timestamp += $fraction; 
       
       # print "timestamp = $timestamp\n";
       # print "starttime_secs = $starttime_secs\n";
       # print "endtime_secs   = $endtime_secs\n";
       
       if($timestamp > 0.0 && $timestamp < ($endtime_secs-$starttime_secs)) { 
           
           my $timestampTag      = "<timestamp>$timestamp</timestamp>\n"; 
           my $row .= $timestampTag;
            
            ## CHECK IF THE TEXTMESSAGE IS A WISEML DATA TAG ##
           if($textMessage =~ m/\<data\ +key\ *=/ && $textMessage =~ m/\<\/data\>/){
                my $temp_text = $textMessage;
                $temp_text =~ s/\<data\ +key\ *=\ *//;
                $temp_text =~ s/\ *\<\/data\>//;
                $temp_text =~ s/\"//;
                $temp_text =~ s/\"//;
                $temp_text =~ s/\'//;
                $temp_text =~ s/\'//;
                $temp_text = &Utils::trimchomp($temp_text);
                my @tokens = split('>', $temp_text);
                if(@tokens==2){
                    my $key   = &Utils::trimchomp(@tokens[0]);
                    my $value = &Utils::trimchomp(@tokens[1]);
                    print "data tag found: key: $key value: $value\n";
                    my $dataEntryTag = "<data key=\"" . $key . "\">$value</data>\n";
                    $row.=generateNodeOutputTag($urn, $dataEntryTag, $database);
                }
            }
            elsif($textMessage =~ m/\<\!\>\<link\ +source\ *="|'.*\ +target\ *="|'.*\>/) {
                my $temp_text = $textMessage;
                # print "matches link tag!\n";
                $temp_text =~ s/\<\!\>\<link\ +source\ *=//;
                $temp_text =~ s/^"//;
                $temp_text =~ s/^'//;
                my $source = $temp_text;
                my $dest = $temp_text;
                $source =~ s/".*//;
                $dest =~ s/.*".*target\ *="//;
                $dest =~ s/".*//;
                my $rssi = "";
                if($textMessage =~ /\<rssi.*/) {
                   $rssi = $textMessage;
                   $rssi =~ s/.*<rssi\>//;
                   $rssi =~ s/\<\/rssi\>.*//;
                   $rssi = "<rssi>$rssi</rssi>";
                }
                # print "source=$source dest=$dest\n";
                # print "rssi: $rssi\n";
                
                if(!&Utils::hasURNPrefix($dest)){
                    $dest = $urn_prefix . $dest;
                }
                if(!&Utils::hasURNPrefix($source)){
                    $source = $urn_prefix . $source;
                }
                
                
                my $linkEntryTag = "<link source=\"" . $source . "\" target=\"" . $dest . "\">$rssi</link>\n";
                $row.=$linkEntryTag;
                
                # keep the original message here for debugging
                # $textMessage = &Utils::escapeXML($textMessage);
                # my $dataEntryTag = "<data key=\"debug\">$textMessage</data>\n";
                # $row.=generateNodeOutputTag($urn, $dataEntryTag, $database);   
            }
            else {
                $textMessage = &Utils::escapeXML($textMessage);
                my $dataEntryTag = "<data key=\"text\">$textMessage</data>\n";
                $row.=generateNodeOutputTag($urn, $dataEntryTag, $database);
                
            }
            &Utils::appendStringToFile($filepath, $row);
       } 
       else {
            print "NetworkResourcesReader::appendSensornodesOutputToXMLFile - ignoring message delivered at time=$time as time is out of defined range\n";
       } 
    }
    &Utils::appendStringToFile( $filepath , "</trace>\n");
}    
    

sub generateNodeTag {
    my $urn = $_[0];
    my $database = new DatabaseInterface();
    my @rows = $database->getPropertiesOfSensorNode_withTypeAndUnit($urn);
    
    
    my $gateway_content;
    my $gateway_type;
    my $gateway_unit;
    my $type_content;
    my $type_type;
    my $type_unit;
    my $description_content;
    my $description_type;   
    my $description_unit;   
    
    # iterate over the properties to get the mandatory properties
    for(my $i=0; $i<@rows; $i++){
      my @row = @{$rows[$i]};
      if ( @row[1] =~ "gateway") {
          $gateway_content = $database->getSensorNodePropertyValue(@row[0], $urn);
          $gateway_type = @row[2];
          $gateway_unit = @row[3];
      }
      if ( @row[1] =~ "nodeType") {
          $type_content = $database->getSensorNodePropertyValue(@row[0], $urn);
          $type_type = @row[2];
          $type_unit = @row[3];
      }
      if ( @row[1] =~ "description") {
          $description_content = $database->getSensorNodePropertyValue(@row[0], $urn);
          $description_type = @row[2];
          $description_unit = @row[3];
      }
    }
    
    # fetch the position data
    my @positionData = $database->getSensorNodePositionData($urn);
    my $x = @positionData[0];
    my $y = @positionData[1];
    my $z = @positionData[2];
    
    # save the position data in the last_x last_y last_z hash!
    $last_x{$urn} = $x;
    $last_y{$urn} = $y;
    $last_z{$urn} = $z;      
    
    my $nodeTag     = "<node id=\"$urn\">\n";
    
    my $positionTag = "<position>\n\t<x>$x</x>\n\t<y>$y</y>\n\t<z>$z</z>\n</position>\n";
    my $gatewayTag  = "<gateway>$gateway_content</gateway>\n";
    my $nodeTypeTag = "<nodeType>$type_content</nodeType>\n";
    my $description = "<description>$description_content</description>\n";

    $nodeTag.= $positionTag;
    $nodeTag.= $gatewayTag;
    $nodeTag.= $nodeTypeTag;
    $nodeTag.= $description;

    # getting all the capabilities
    for(my $i=0; $i<@rows; $i++){
       my @row = @{$rows[$i]};
       if ( @row[1] =~ "urn:wisebed:node:capability") {
           my $property_id = @row[0];
           my $capability_name = @row[1];
           my $capability_content = $database->getSensorNodePropertyValue($property_id, $urn);
           my $capability_type = @row[2];
           my $capability_unit = @row[3];
           my $capabilityTag = "<capability>\n\t<name>$capability_name</name>\n\t<datatype>$capability_type</datatype>\n\t<unit>$capability_unit</unit>\n\t<default>$capability_content</default>\n\t</capability>\n";
           # append to nodetag
           $nodeTag .= $capabilityTag;
       }
    }
    $nodeTag.= "<\/node>\n";  
}


     
1;