<?php
/*
 * SNA-Portal mediates between one or more Authentication and Authorisation
 * Infrastructures and non-AAI-enabled roles.
 * Copyright (C) 2003  Swiss Virtual Campus (http://www.swissvirtualcampus.ch)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * $Id: MySQLConnection.class.inc,v 1.22 2006/11/06 16:42:44 vlry Exp $
 */

/**
* The current MySQL connection.
*
* @var MySQLConnection 
* @access protected
*/
$_MYSQLCONNECTION_CURRENT_CONNECTION = NULL;

/**
 * The default host name for a MySQL server.
 */
if (! defined("DEFAULT_MYSQL_HOST")) {
  define("DEFAULT_MYSQL_HOST",      "localhost");
}

/**
 * The default user name for connecting to a MySQL server
 */
if (! defined("DEFAULT_MYSQL_USER")) {
  define("DEFAULT_MYSQL_USER",      "mysql");
}

/**
 * The default password for connecting to a MySQL server
 */
if (! defined("DEFAULT_MYSQL_PASS")) {
  define("DEFAULT_MYSQL_PASS",      "");
}

/**
 * The default database name to be used on a MySQL server.
 */
if (! defined("DEFAULT_MYSQL_BASE")) {
  define("DEFAULT_MYSQL_BASE",      "");
}

/**
 * MySQLConnection is a wrapper class for the MySQL connection object
 * provided by the mysql library. The class provides 
 * <ol>
 *   <li>static methods for managing a single, global instance of
 *      of MySQL database connection (see 
 *      {@link MySQLConnection::createConnection()},
 *      {@link MySQLConnection::getConnection()},
 *      {@link MySQLConnection::closeConnection()}
 *      )
 *  </li>
 *
 *   <li>instance methods similar to those found in the MySQL library.
 *      Instance methods usually wrap mysql library calls with some
 *      error handling code.
 *  </li>  
 *
 * </ol>
 *
 * <b>Typical Usage</b>
 * <pre>
 *    // create and open the connection
 *    //
 *    MySQLConnection::createConnection(
 *      array(
 *        "MYSQL_HOST"  => "mysql.my.domain",
 *        "MYSQL_USER"  => "username",
 *        "MYSQL_PASS"  => "mypassword"
 *      )
 *    );
 *
 *    // get the current connection
 *    //
 *    $conn = MySQLConnection::getConnection();
 *
 *    // query the database
 *    //
 *    $conn->query("SELECT * FROM SomeTable");
 *    while($row = $conn->fetch_array()) {
 *      print_r($row);
 *    } 
 *
 *    // closes the current connection  
 *    //
 *    MySQLConnection::closeConnection();
 * </pre>
 *
 *
 * @author Thomas Spreng <spreng@iam.unibe.ch>
 * @version $Revision: 1.22 $
 * @package Library
 *
 */
require_once('include/Logger.class.inc');

class MySQLConnection {

  /* ********************************************************************** */
  /*                     static methods                                     */
  /* ********************************************************************** */ 
  
  /**
  * Establishes a connection to the MySQL database.
  *
  * The connection parameters are either submitted in the parameter array
  * $param or defined in global constants.  
  *
  * <b>Typical Usage - Use function parameters</b>
  * <pre>
  *    // create and open the connection
  *    //
  *    MySQLConnection::createConnection(
  *      array(
  *        "MYSQL_HOST"  => "mysql.my.domain",
  *        "MYSQL_USER"  => "username",
  *        "MYSQL_PASS"  => "mypassword"
  *      )
  *    );
  * </pre>
  *
  * <b>Typical Usage - Use constants</b>
  * <pre>
  *    // create and open the connection
  *    // Initializes the database connection with the values of the global
  *    // constants
  *    //   MYSQL_HOST
  *    //   MYSQL_USER
  *    //   MYSQL_PASS
  *    MySQLConnection::createConnection();
  * </pre>
  *
  * The following parameters are important when the connection created and opened:
  * <ul>
  *  <li><b>Host</b>: the host name or IP address of the MySQL
  *     server. The following rules apply:
  *    <ol>
  *      <li>the host is set to $param["MYSQL_HOST"], if $param["MYSQL_HOST"] is
  *           set</li>
  *      <li>the host is set to {@link MYSQL_HOST}, if {@link MYSQL_HOST} is
  *           defined</li>
  *      <li>the host is set to {@link DEFAULT_MYSQL_HOST}, if  {@link DEFAULT_MYSQL_HOST} is
  *           defined</li>
  *    </ol>
  *  </li>
  *
  *  <li><b>User</b>: the user name for connecting to the database. 
  *    The following rules apply:
  *    <ol>
  *      <li>the user is set to $param["MYSQL_USER"], if $param["MYSQL_USER"] is
  *           set</li>
  *      <li>the user is set to {@link MYSQL_USER}, if {@link MYSQL_USER} is
  *           defined</li>
  *      <li>the user is set to {@link DEFAULT_MYSQL_USER}, if  {@link DEFAULT_MYSQL_USER} is
  *           defined</li>
  *    </ol>
  *  </li>
  *
  *  <li><b>Password</b>: the password for connecting to the database. 
  *    The following rules apply:
  *    <ol>
  *      <li>the password is set to $param["MYSQL_PASS"], if $param["MYSQL_PASS"] is
  *           set</li>
  *      <li>the password is set to {@link MYSQL_PASS}, if {@link MYSQL_PASS} is
  *           defined</li>
  *      <li>the password is set to {@link DEFAULT_MYSQL_PASS}, if  {@link DEFAULT_MYSQL_PASS} is
  *           defined</li>
  *    </ol>
  *  </li>
  *
  *  <li><b>Database</b>: the database
  *    The following rules apply:
  *    <ol>
  *      <li>the database is set to $param["MYSQL_BASE"], if $param["MYSQL_BASE"] is
  *           set</li>
  *      <li>the database is set to {@link MYSQL_BASE}, if {@link MYSQL_BASE} is
  *           defined</li>
  *      <li>the database is set to {@link DEFAULT_MYSQL_BASE}, if  {@link DEFAULT_MYSQL_BASE} is
  *           defined</li>
  *    </ol>
  *  </li>
  * </ul>
  *
  * @see MySQLConnection::getConnection()
  *
  * @static
  *
  */
  function & createConnection($param = NULL) {
    $opt = array();
    if (! is_null($param) && is_array($param)) {
       foreach($param as $name => $value) {
         $opt[strtoupper($name)] = $value; 
       } 
    }
    
    if (isset($opt["MYSQL_HOST"])) {
      $host =  $opt["MYSQL_HOST"];
    } elseif (defined("MYSQL_HOST")) {
      $host = MYSQL_HOST; 
    } else {
      $host = DEFAULT_MYSQL_HOST; 
    }
    
    if (isset($opt["MYSQL_USER"])) {
      $user =  $opt["MYSQL_USER"];
    } elseif (defined("MYSQL_USER")) {
      $user = MYSQL_USER; 
    } else {
      $user = DEFAULT_MYSQL_USER; 
    }    
    
    if (isset($opt["MYSQL_PASS"])) {
      $password =  $opt["MYSQL_PASS"];
    } elseif (defined("MYSQL_PASS")) {
      $password = MYSQL_PASS; 
    } else {
      $password = DEFAULT_MYSQL_PASS; 
    }    

    if (isset($opt["MYSQL_BASE"])) {
      $database =  $opt["MYSQL_BASE"];
    } elseif (defined("MYSQL_BASE")) {
      $database = MYSQL_BASE; 
    } else {
      $database = DEFAULT_MYSQL_BASE; 
    }    

    $conn = new MySQLConnection($host, $user, $password, $database);
    $conn->connect();
    $GLOBALS["_MYSQLCONNECTION_CURRENT_CONNECTION"] = $conn;    
    return $conn;
  }
  
  /* ********************************************************************** */ 
  /**
  * Closes the currently opened connection.
  *
  * @static    
  */
  function closeConnection() {
    $conn = $GLOBALS["_MYSQLCONNECTION_CURRENT_CONNECTION"];
    if (! is_null($conn)) {
       $conn->close();
    }
    $GLOBALS["_MYSQLCONNECTION_CURRENT_CONNECTION"] = NULL;
  }

  /* ********************************************************************** */ 
  /**
  * Replies the current database connection.  
  *
  * @return MySQLConnection  the current database connection; NULL, if no database
  *   connection is currently established 
  *
  * @see MySQLConnection::createConnection()
  *
  * @static
  *
  */
  function & getConnection() {   
    if (isset($GLOBALS["_MYSQLCONNECTION_CURRENT_CONNECTION"]) 
        and is_null($GLOBALS["_MYSQLCONNECTION_CURRENT_CONNECTION"])) {
        MySQLConnection::createConnection();
    }
    $conn= $GLOBALS["_MYSQLCONNECTION_CURRENT_CONNECTION"];
    return $conn;
  }

  /* ********************************************************************** */
  /*                     instance variables                                 */
  /* ********************************************************************** */ 

  /** 
   * Hostname or IP address of MySQL server
   *
   * @var string   
   * @access private 
   */
   var $server;  

  /**
   * Username for the MySQL connection
   * 
   * @var string
   * @access private
   */
   var $user;    

  /**
   * Password for the MySQL connection
   *
   * @var string
   * @access private
   */
   var $passwd;   

  /**
   * The name of the database to connect to.
   *
   * @var string
   * @access private
   */
   var $database;  

  /** 
   * The open MySQL connection.
   *
   * @var handle 
   * @access private
   */
   var $link;    


  /**
   * Temporarily holds query results
   *
   * @var  handle
   * @access private 
   */
   var $result;
   
   /**
    * Error string
    *
    * @var string
    * @access private
    */
   var $error;

  /* ********************************************************************** */
  /*                     constructor                                        */
  /* ********************************************************************** */ 
  
  /**
   * constructor
   *
   * Creates a new MySQL connection with the given parameters. The 
   * connection to the database it not opended yet. Use
   * {@link function connect() } to open the connection.
   *  
   * @param string $server  the hostname or IP address of the MySQL server.
   * @param string $user    the name of the database user
   * @param string $passwd  the password for connecting to the database
   * @param string $database the name of the database to connect to
   *
   * @see connect()
   * @see close()
   *
   * @access private
   *
   */
   function MySQLConnection($server=MYSQL_HOST, $user=MYSQL_USER, $passwd=MYSQL_PASS, $database=MYSQL_BASE) {
	$this->server = $server;
	$this->user = $user;
	$this->passwd = $passwd;
	$this->database = $database;
	$this->link = false;
	$this->result = false;
	$this->error= NULL;
  }

  /* ********************************************************************** */ 
  /**
   * Opens a connection to the MySQL database. 
   *
   */
  function connect() {

	if (!($this->link)) {
		$this->link = mysql_connect($this->server, $this->user, $this->passwd)
			or die("MySQL Error: ".mysql_error());
		$this->select_db();
	}

  }
  
  /* ********************************************************************** */ 
  /**
   * Selects the database in the current MySQL server instance.
   *
   * @access private 
   */
  function select_db() {

	if ($this->link)
		mysql_select_db($this->database, $this->link)
			or die("MySQL Error: ".mysql_error());

  }

  /* ********************************************************************** */ 
  /**
   * Closes the current connection
   */
   function close() {	
	if ($this->link)
		mysql_close($this->link)
			or die("MySQL Error: ".mysql_error());

   }
   
  /* ********************************************************************** */ 
  /**
   * Evaluates an SQL query on the database. The result of the
   * query is cached internally and can be retrieved with the
   * method {@link fetch_array() }.
   *
   * @param string $query  an SQL query 
   *
   * @see fetch_array()
   * @see num_rows()
   * @see insert_id() 
   * 
   */
   function query($query, $die= true) {

    if (!($this->link)) {
        $this->connect();
    }
    $logger=& Logger::getInstance();
    //XXX: $logger->debug("SQL query: $query");
    $this->result = mysql_query($query);
    if (! $this->result and $die ) {
        die("MySQL Error: ".mysql_error(). "(Query: ".$query.")");
    }
    else if (! $this->result) {
        $this->error= mysql_error();
        $logger->warn("SQL error: '". $query . "': " . $this->error);
        return false;
    }
    else {
        return true;
    }
   }
  
  /* ********************************************************************** */ 
  /**
   * Invokes mysql_fetch_array on the result of the last {@link  query() }.
   *
   * @return array  an array of data if there are still database records
   *    to fetch; false otherwise 
   *  
   */    
  function fetch_array() {
	
	return mysql_fetch_array($this->result);

  }

  /* ********************************************************************** */ 
  /**
   * Replies the number of rows from the last {@link  query() }.
   *
   * @return int  the number of rows from the last query 
   *  
   */    
   function num_rows() {
	
	return mysql_num_rows($this->result);

   }


  /* ********************************************************************** */ 
  /**
   * Replies the automatically inserted database ID of the last {@link  query() },
   * if the last query was an INSERT statement on a table with an autoincrement
   * column.
   *
   * @return int  the automatically inserted database ID of the last {@link  query() }
   *  
   */    

   function insert_id() {	
	return mysql_insert_id();
   }

  /* ********************************************************************** */ 
  /**
   * Frees the roles allocated to the last query (result objects, etc.)
   *
   *  
   */    

   function freeresult() {
	
	mysql_freeresult($this->result);
	
   }

  /* ********************************************************************** */ 
  /**
   * Replies the number of rows affected by the last query
   *
   * @return int  the number of rows affected by the last query 
   *
   */    
   function affected_rows() {
	    return mysql_affected_rows($this->link);
   }

   /**
    * Returns the underlying error message (MySQL).
    *
    * @return String The underlying error message or NULL.
    */
   function getError() {
       return $this->error;
   }
  
   /**
    * Returns the MySQL version.
    *
    * @return String The MySQL version.
    */
   function server_info() {
       return mysql_get_server_info($this->link);
   }
 
    /**
     * SQL escape the given string.
     * @return String the escaped string.
     */
    function escape_string($string) {
        // PHP >= 4.3.0
        if (function_exists('mysql_real_escape_string')) {
            if (!($this->link)) $this->connect();
            return mysql_real_escape_string($string,$this->link);
        }
        else {
            return mysql_escape_string($string);
        }
    }
     
 
} // end class 

?>
