<?php
/*
 * SNA-Portal mediates between one or more Authentication and Authorisation
 * Infrastructures and non-AAI-enabled roles.
 *
 * Copyright (C) 2003  Swiss Virtual Campus (http://www.swissvirtualcampus.ch) 
 * Copyright (C) 2004-2005 SWITCH (http://www.switch.ch)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * $Id: User.class.inc,v 1.104 2007/10/18 08:51:13 vlry Exp $
 */

require_once("include/constants.inc");
require_once("include/Subscription.class.inc");
require_once("include/UserPrivilege.class.inc");
require_once("include/Logger.class.inc");

/**
 * This class represents a principial who uses the SNA-Portal. 
 *
 * @author Thomas Spreng  <spreng@iam.unibe.ch>
 * @author Valery Tschopp  <tschopp@switch.ch>
 * @version $Revision: 1.104 $
 *
 * @see MySQLConnection
 *
 * @package Library 
 */ 
class User {


  /* ************************************************************* */
  /*               static functions                                */
  /* ************************************************************* */

  /**
   * Reconstitutes a user from the database. The user is specified 
   * by its internal database ID. 
   * 
   * <b>Typical usage</b>
   * <pre>
   *    $user = User::reconstituteById(5);
   *    if (is_null($user)) {
   *      echo "ERROR: failed to reconstitute user with database id '5'";
   *    }
   * </pre>
   *
   * @param int  $id       the internal database ID 
   * 
   * @return User an initialized instance of class User, if the respective user
   *   exists; NULL, otherwise
   *
   * @static 
   */
   function & reconstituteById($id) {
    
    $conn = MySQLConnection::getConnection();
    
    $query = " SELECT user_id, user_uid, user_privilege FROM User " 
           . " WHERE user_id = $id "; 

    $conn->query($query);
    if (! ( $row = $conn->fetch_array() ) ) {
      $conn->freeresult();
      $logger=& Logger::getInstance();
      $logger->error("reconstituteById($id): no such user in DB.");
      $user= null;
      return $user;
    }

    $user = new User();
    $user->uid= $row["user_uid"];
    $user->user_id= $row["user_id"];
    $user->user_privilege= $row["user_privilege"];

    return $user;
 } 


 /**
  * Reconstitutes a user from the database using the external user identity 
  * as key. Typical examples for external user IDs are "portaladmin"
  * or "demo@aaitest.switch.ch".
  *
  * <b>Typical usage</b>
  * <pre>
  *    
  *    $user = User::reconstituteByUid('portaladmin');
  *    if (is_null($user)) {
  *      echo "ERROR: failed to reconfstitute user with database id '5'";
  *    }
  * </pre>
  *
  * @param string $uid  the user id
  *
  * @return User  an initialized instance of class User, if the respective user
  *     exists; NULL, otherwise
  *
  * @static 
  */
 function & reconstituteByUid($uid) {
    
    if (empty($uid)) {
        $user= null;
        return $user;
    }
    
    $conn=& MySQLConnection::getConnection();
    $user_uid= $conn->escape_string($uid);
    $query = " SELECT user_id, user_uid, user_privilege FROM User " 
           . " WHERE user_uid = '$user_uid' ";

    $conn->query($query);
    if (! ( $row = $conn->fetch_array() ) ) {
      $conn->freeresult();
      $logger=& Logger::getInstance();
      $logger->error("reconstituteByUid($uid): no such user in DB.");
      $user= null;
      return $user;
    }

    $user= new User();
    $user->uid= $row["user_uid"];
    $user->user_id= $row["user_id"];
    $user->user_privilege= $row["user_privilege"];
    
    return $user;
 }


 /* ************************************************************* */
 /** 
  * Checks whether a specific external user ID already exists in the 
  * database.
  *
  * @param string  $uid   an external User ID
  *
  * @return boolean  true, if there is already a user with User ID $uid
  *   false, otherwise
  *
  * @static 
  *
  */  
 function existsUid($uid) {
  
    $conn=& MySQLConnection::getConnection();
    $user_uid= $conn->escape_string($uid); 
    $query = " SELECT user_uid "
           . " FROM User "
           . " WHERE user_uid ='$user_uid' ";

    $conn->query($query);
    $ret = $conn->num_rows() > 0;
    $conn->freeresult();
    return $ret;  
 }

 /* ************************************************************* */
 /** 
  * Creates a new user with user ID $uid 
  *
  * @param string  $uid  the user_uid (username)
  * @param int $admin_id create by admin ID (default: 1)
  *
  * @return User the new user object 
  *
  * @static 
  *
  */  
 function & create($uid, $admin_id= 1) {
  
    $conn=& MySQLConnection::getConnection();
    $user_uid= $conn->escape_string($uid); 
    $query  = " INSERT into User ";
    $query .= " (user_privilege, user_uid, user_valid_until, date_created, addr_created) ";
    $query .= " VALUES ( ";
    $query .= " 1, ";
    $query .= " '$user_uid', ";
    $query .= " NULL, ";
    $query .= " NOW(), ";
    $remote= '127.0.0.1';
    if (isset($_SERVER['REMOTE_ADDR'])) { $remote= $_SERVER['REMOTE_ADDR']; }
    $query .= " '$remote' ";
    $query .= ")";
    
    $conn->query($query);
    
    $user=& User::reconstituteByUid($uid);
    $user_id= $user->user_id;
    $user_uid= $user->uid;
    
    // create record in UserAttributes table
    $query=  " INSERT INTO UserAttributes ";
    $commonName= $conn->escape_string($user_uid);
    $query.= " (user_id,user_uid,commonName,privilegeName,date_created, addr_created) ";
    $query.= " VALUES ( $user_id, '$user_uid', '$commonName', 'User', NOW(), '$remote' ) ";
    $conn->query($query);
   
    Event::storeUserEvent("User created: $user_uid", $admin_id);
 
    return $user;
 }


 /**
  * Lists all user ID (internal DB id: User.user_id)
  *
  * @return array List of all user_id.
  * @static
  */
 function getUserIDs() {
     $conn = MySQLConnection::getConnection();
     $query= " SELECT user_id "
           . " FROM User ";
      $conn->query($query);
      $user_ids= array();
      while($row = $conn->fetch_array()) {
         $user_ids[]= $row['user_id'];
      }
      $conn->freeresult();
      return $user_ids;
 }

 /**
  * Lists all user UID (username, uniqueid)
  *
  * @return array List of all user_uid.
  * @static
  */
 function getUIDs() {
     $conn = MySQLConnection::getConnection();
     $query= " SELECT user_uid "
           . " FROM User ";
      $conn->query($query);
      $user_uids= array();
      while($row = $conn->fetch_array()) {
         $user_uids[]= $row['user_uid'];
      }
      $conn->freeresult();
      return $user_uids;
 }

 /**
  * Count of orphan users.
  *
  * @return int number of orphan users.
  * @static
  */
 function getOrphansCount() {
     $conn=& MySQLConnection::getConnection();
     $query= " SELECT user_id "
           . " FROM User LEFT JOIN Subscription "
           . " ON ( user_id = subscription_user ) "
           . " WHERE subscription_role IS NULL "
           . "   AND user_privilege = 1";
      $conn->query($query);
      $n_orphans= $conn->num_rows();
      $conn->freeresult();
      return $n_orphans;
 }

  /**
   * Count of users
   *
   * @return int Number of users.
   * @static
   */
  function getUsersCount($search_field= null, $search_value= null) {
     $conn=& MySQLConnection::getConnection();
     $query= " SELECT User.user_id "
           . " FROM User, UserAttributes "
           . " WHERE User.user_id = UserAttributes.user_id ";
     if (!is_null($search_field) and !is_null($search_value) ) {
         if ($search_field == 'user_uid') { $search_field= 'User.user_uid'; }
         $esc_search_value= $conn->escape_string($search_value);
         $query.= " AND $search_field LIKE '$esc_search_value' ";
     }
     $conn->query($query);
     $n_users= $conn->num_rows();
     $conn->freeresult();
     return $n_users;
  }

  /**
   * Attributes of users
   *
   * @return array Attributes of users.
   * @static
   */
  function getUsersAttributes( $search_field= null, $search_value= null,
                                  $orderby= null, $offset= null, $count= null ) {
     $conn=& MySQLConnection::getConnection();
     $query= " SELECT User.user_id, User.user_uid, privilegeName, surname, givenName, user_last_login "  
           . " FROM User, UserAttributes "
           . " WHERE User.user_id = UserAttributes.user_id ";
     if (!is_null($search_field) and !is_null($search_value) ) {
         if ($search_field == 'user_uid') { $search_field= 'User.user_uid'; }
         $esc_search_value= $conn->escape_string($search_value);
         $query.= " AND $search_field LIKE '$esc_search_value' ";
     }
     if (!is_null($orderby)) {
        if (strpos($orderby, 'surname') !== false) {
            $query.= " ORDER BY surname, givenName, privilegeName, user_uid ";
        }
        elseif (strpos($orderby, 'givenName') !== false) {
            $query.= " ORDER BY givenName, surname, privilegeName, user_uid ";
        }
        elseif (strpos($orderby, 'privilegeName') !== false) {
            $query.= " ORDER BY privilegeName, surname, givenName, user_uid ";
        }
        elseif (strpos($orderby, 'user_last_login') !== false) {
            $query.= " ORDER BY user_last_login DESC, surname, givenName, user_uid ";
        }
        //XXX: NO DEFAULT
        //else {
        //    $query.= " ORDER BY $orderby ";        
        //}
     }
     if (!is_null($offset) and !is_null($count)) {
        $query.= " LIMIT $offset,$count ";
     }
     $conn->query($query);
     $users_attrs= array();
     while($row = $conn->fetch_array()) {
        $users_attrs[]= $row;
     }
     $conn->freeresult();
     return $users_attrs;
  }

 /**
  * All orphan users attributes: 
  *  user_id, user_uid, privilegeName, surname, givenName
  *
  * @return array of user attributes.
  * @static
  */
  function getAllOrphansAttributes( $orderby= null, $offset= null, $count= null ) {
     $conn=& MySQLConnection::getConnection();
     $query= " SELECT User.user_id, User.user_uid, privilegeName, surname, givenName, user_last_login "  
           . " FROM ( User, UserAttributes ) "
           . " LEFT JOIN Subscription "
           . "      ON ( UserAttributes.user_id = subscription_user ) "
           . " WHERE subscription_role IS NULL "
           . "   AND UserAttributes.user_id = User.user_id "
           . "   AND user_privilege = " . USER_PRIVILEGE_ID_USER;
     if (!is_null($orderby)) {
        if (strpos($orderby, 'surname') !== false) {
            $query.= " ORDER BY surname, givenName, privilegeName, user_uid ";
        }
        elseif (strpos($orderby, 'givenName') !== false) {
            $query.= " ORDER BY givenName, surname, privilegeName, user_uid ";
        }
        elseif (strpos($orderby, 'privilegeName') !== false) {
            $query.= " ORDER BY privilegeName, surname, givenName, user_uid ";

        }
        elseif (strpos($orderby, 'user_last_login') !== false) {
            $query.= " ORDER BY user_last_login, surname, givenName, user_uid ";
        }
        else {
            $query.= " ORDER BY $orderby ";        
        }
     }
     if (!is_null($offset) and !is_null($count)) {
        $query.= " LIMIT $offset,$count ";
     }
     $conn->query($query);
     $users_attrs= array();
     while($row = $conn->fetch_array()) {
        $users_attrs[]= $row;
     }
     $conn->freeresult();
     return $users_attrs;
  }

  /* ************************************************************* */
  /*               member variables                                */
  /* ************************************************************* */

  /**
   * The unique external user ID (i.e. <code>portaladmin</code> or
   * <code>demo@aaitest.unibe.ch</code>
   * 
   * @var string
   * @access private
   */
  var $uid;   
  
  /**
  * The internal database ID of the User (primary key in table
  * <code>User</code>.
  *
  * @var int
  * @access private
  */ 
  var $user_id;    // user_id from User Table in portal database

  /**
  * The user privilege id, as define in table UserPrivilege 
  *
  * @var int
  * @access private
  */ 
  var $user_privilege;


  /* ************************************************************* */
  /*               constructor                                     */
  /* ************************************************************* */

  /**
   * constructor
   * 
   * @access private
   */     
  function User() {
  }

  /* ************************************************************* */
  /*                         methods                               */
  /* ************************************************************* */


  /**
  * Deletes this user from the database. 
  * user_id == 1 (portaladmin) can not be deleted.
  *
  * <ul>
  *  <li>deletes all subscriptions of this user
  *  <li>deletes all roles owned by this user
  *  <li>deletes all attributes associated with this user 
  *  <li>deletes the user itself
  * </ul>
  *
  */
  function delete($admin_id= 1) {

    $logger=& Logger::getInstance();
    // user portaladmin CAN NOT BE DELETED
    if ($this->user_id > 1) {

      $logger->info("delete user: [$this->user_id] $this->uid");

      $conn =& MySQLConnection::getConnection();

      $query = "DELETE FROM Subscription WHERE subscription_user = $this->user_id";
      $conn->query($query);
      
      // Don't delete user owned role, portaladmin (id:1) becomes owner
      $query= "UPDATE Role SET role_user = 1 WHERE role_user = $this->user_id";
      $conn->query($query);
      
      $query = "DELETE FROM RoleManager WHERE user_id = $this->user_id";
      $conn->query($query);
      $query = "DELETE FROM Attribute WHERE attribute_user = $this->user_id";
      $conn->query($query);
      $query = "DELETE FROM User WHERE user_id = $this->user_id";
      $conn->query($query);
      // Delete record in UserAttributes table
      $query = "DELETE FROM UserAttributes WHERE user_id = $this->user_id";
      $conn->query($query);
      $this->user_id = false;
    
      Event::storeUserEvent("User deleted: $this->uid", $admin_id, EVENT_LEVEL_WARNING);
    }
    else {
        Event::storeSystemEvent("User: $this->user_uid can not be deleted", EVENT_LEVEL_WARNING);
        $logger->warn("delete user: [$this->user_id] $this->uid CAN NOT BE DELETED");
    }

  }

  /**
  * Updates the user login date.
  */
  function updateLoginDate() {
    $query= "UPDATE User "
          . "SET user_last_login = NOW() "
          . "WHERE user_id = $this->user_id ";
    $conn=& MySQLConnection::getConnection();
    $conn->query($query);
  }
  
  /**
   * Returns the user last login date.
   * @return String SQL date or null.
   */
  function getLastLogin() {
      $query= "SELECT user_last_login FROM User "
            . "WHERE user_id = $this->user_id ";
      $conn=& MySQLConnection::getConnection();    
      $conn->query($query);
      $row= $conn->fetch_array();
      $conn->freeresult();

      $last_login= null;
      if ( ! empty($row['user_last_login']) ) {
        $last_login= $row['user_last_login'];
      }
      return $last_login;
    
  }

  /**
  * Retrieves a list of attribute IDs of all attributes associated
  * with this user.
  *
  * @return  array  an array of attribute type IDs 
  *
  */
  function getStoredAttributeTypeIds() {
  
    $conn = MySQLConnection::getConnection();
    
    $query  = "select attributetype_id ";
    $query .= "from AttributeType, Attribute, User ";
    $query .= "where attribute_type = attributetype_id ";
    $query .= "and attribute_user = user_id ";
    $query .= "and user_id = $this->user_id ";
  
    $conn->query($query);
    while ($row = $conn->fetch_array()) {
      $attributetype_id[] = $row["attributetype_id"];
    }
    $conn->freeresult();
    return $attributetype_id;

  }

  /* ************************************************************* */
  /**
  * Checks whether an attribute value of a specific attribute type 
  * is trusted for this user. 
  *
  * An attribute value is trusted if it is delivered by the underlying
  * AAI platform rather than entered by the user in the SNA-Portal web
  * interface.
  *
  * @param  mixed  the attribute type ID. Either a number or the
  *     unique LDAP name of the attribute type. Example: the numeric code
  *     "1" and the LDAP name "surname" both denote
  *     the attribute type "Family Name".
  *
  * @return boolean true 
  *
  */
  function isAttributeTrusted($attributeType) {

    $conn =& MySQLConnection::getConnection();
      
    if (is_int($attributeType)) {
  
      $attributeType_id = $attributeType;
      $query  = "select attribute_trusted from Attribute, AttributeType, User ";
      $query .= "where attribute_type = attributetype_id ";
      $query .= "and attribute_user = user_id ";
      $query .= "and user_uid = '$this->uid' ";
      $query .= "and attributetype_id = $attributeType_id";
  
    } else {
      
      $query  = "select attribute_trusted from Attribute, AttributeType, User ";
      $query .= "where attribute_type = attributetype_id ";
      $query .= "and attribute_user = user_id ";
      $query .= "and user_uid = '$this->uid' ";
      $query .= "and attributetype_ldapname = '$attributeType'";
    
    }
  
    $conn->query($query);
    $row = $conn->fetch_array();
    $conn->freeresult();
    return $row['attribute_trusted'];

  }

  /* ************************************************************* */
  /**
  * Replies the attribute value for a specific attribute type.
  * 
  * @param  mixed  the attribute type ID. Either a number or the
  *     unique LDAP name of the attribute type. Example: the numeric code
  *     "1" and the LDAP name "surname" both denote
  *     the attribute type "Family Name".
  *
  * @return string the attribute value or NULL
  *
  */
  function getAttribute($attributeType) {

    $conn =& MySQLConnection::getConnection();
      
    if (is_int($attributeType)) {
  
      $attributeType_id = $attributeType;
      $query  = "select attribute_value from Attribute, AttributeType ";
      $query .= "where attribute_type = attributetype_id ";
      $query .= "and attribute_user = $this->user_id ";
      $query .= "and attributetype_id = $attributeType_id";
  
    } else {
      
      $query  = "select attribute_value from Attribute, AttributeType ";
      $query .= "where attribute_type = attributetype_id ";
      $query .= "and attribute_user = $this->user_id ";
      $query .= "and attributetype_ldapname = '$attributeType'";
    
    }
  
    $conn->query($query);
    $row = $conn->fetch_array();
    $conn->freeresult();

    if ( empty($row['attribute_value']) ) {
        $attr= null;
        return $attr;
    }
    else {
        return $row['attribute_value'];
    }
 }

  /**
   * @param int The AttributeTypeID of the attribute, see ATID_*
   * @return string The attribute value iff attribute is trusted, NULL otherwise.
   */
  function getTrustedAttribute($atid) {
      $conn=& MySQLConnection::getConnection();
      $query= "SELECT attribute_value FROM Attribute "
            . "WHERE attribute_type = $atid "
            . "AND attribute_user = $this->user_id "
            . "AND attribute_trusted = 1 ";
    
      $conn->query($query);
      $row= $conn->fetch_array();
      $conn->freeresult();

      if ( empty($row['attribute_value']) ) {
          $attr= null;
          return $attr;
      }
      else {
          return $row['attribute_value'];
      }
  }

 
 /**
  * Returns a hash (atid => (attribute_value => value, attribute_trusted => trusted))
  * of attribute values. Used without parameters, returns all ATID_* without the last (local password).
  *
  * @params comma-separated list of attribute type id (ATID_*), no param returns all atid.
  * @return array Hashtable (atid => (attribute_value => value, attribute_trusted => trusted)), with empty array if not defined.  
  *
  * @see func_get_args()
  */
  function getAttributes() {
    $atids= func_get_args();
    if (empty($atids)) {
        // no params, all ATID_* without local password
        $atids= range(ATID_EDUPERSONUNIQUEID,ATID_EDUPERSONENTITLEMENT);
    }    

    $attributes= array();
    $or_clause= array();
    foreach($atids as $atid) {
        $or_clause[]= "attribute_type = $atid";
        // initialize hashtable value with empty array;
        $attributes[$atid]= array();
    }
    $or_query= implode(' OR ', $or_clause);
    
    $query= " SELECT attribute_type, attribute_value, attribute_trusted "
          . " FROM Attribute "
          . " WHERE attribute_user = $this->user_id "
          . "   AND ( $or_query ) ";
    
    $conn=& MySQLConnection::getConnection();
    $conn->query($query);
    while( $row= $conn->fetch_array() ) {
        $attributes[$row['attribute_type']]= array( 'attribute_value' => $row['attribute_value'],
                                                    'attribute_trusted' => $row['attribute_trusted'] );
    }
    $conn->freeresult();
    
    return $attributes;
      
  }

  /**
  * Returns the attribute value for the UserAttributes table.
  * 
  * @param  the attribute name: use constant ATTR_*
  * @return string the attribute value or NULL
  *
  */ 
  function getUserAttribute($attr_name) {
      $conn=& MySQLConnection::getConnection();
      
      $query= "SELECT $attr_name FROM UserAttributes "
            . "WHERE user_id = $this->user_id ";
            
      $conn->query($query);
      $row= $conn->fetch_array();
      $conn->freeresult();

      $attr_value= null;
      if ( ! empty($row[$attr_name]) ) {
        $attr_value= $row[$attr_name];
      }
      return $attr_value;
  }
 
   /**
  * Returns the user attributes name, value for the UserAttributes table.
  * 
  * @param  strings
  * @return array of attribute name, value
  *
  */ 
  function getUserAttributes() {
      $attr_names= func_get_args();
      if (empty($attr_names)) {
         // no params, select all
         $attr_names= array('*');
      }
      
      $attrs_clause= implode(',', $attr_names);
      $query= "SELECT $attrs_clause FROM UserAttributes "
            . "WHERE user_id = $this->user_id ";

      $conn=& MySQLConnection::getConnection();
      $conn->query($query);
      while( $row= $conn->fetch_array() ) {
          $attrs= $row;
      }
      $conn->freeresult();
      return $attrs;
  }
 

 /* ************************************************************* */
  /**
  * Marks all attributes of this user as beeing untrusted. 
  * 
  */
  function resetAttributeTrustedState() {
     
    $conn =& MySQLConnection::getConnection();
    
    $query  = "update Attribute ";
    $query .= "set attribute_trusted = 0 ";
    $query .= "where attribute_user =  $this->user_id ";
  
    $conn->query($query);

  }

  /* ************************************************************* */
  /**
  * Persists the attribute value for a specific attribute type.
  * 
  * @param  mixed  $attributeType the attribute type ID. Either a number or the
  *     unique LDAP name of the attribute type. Example: the numeric code
  *     "1" and the LDAP name "surname" both denote
  *     the attribute type "Family Name".
  *
  * @param string $attributeValue  the attribute value
  *
  * @param  boolean $trusted  indicates whether $attributeValue is trusted
  *    or not 
  *
  */
  function storeAttribute($attributeType, $attributeValue, $trusted=false) {
  
        $conn=& MySQLConnection::getConnection();
        
        if (empty($attributeValue)) {
            return false;
        }
        
        // get attributetype id and ldapname
        if (is_int($attributeType)) {
            $attributetype_id= $attributeType;
            $attributetype_ldapname= AttributeType::getAttributeTypeLdapName($attributetype_id);
        }
        else {
            $attributetype_ldapname= $attributeType;
            $attributetype_id= AttributeType::getAttributeTypeIdByLdapName($attributetype_ldapname);
        }

        if ( empty($attributetype_ldapname) or 
             $attributetype_id == -1) {
            trigger_error(
              "User::storeAttribute: failed to save attribute with type '$attributeType'. No such attribute type.",
              E_USER_ERROR
            );
            return false;     
        }   

        // check if we need an update or an insert
        $query  = "select attribute_id from Attribute ";
        $query .= "where attribute_type = $attributetype_id ";
        $query .= "and attribute_user = $this->user_id ";
        $conn->query($query);
        $row = $conn->fetch_array();

        $remote_addr= '127.0.0.1';
        if (isset($_SERVER['REMOTE_ADDR'])) {
            $remote_addr= $_SERVER['REMOTE_ADDR'];
        }
        
        $attr_value= $conn->escape_string($attributeValue);
        if ($conn->num_rows() > 0) {
            // we need an update
            $attribute_id = $row["attribute_id"];
            $query  = "UPDATE Attribute ";
            $query .= "SET attribute_value = '$attr_value', ";
            $query .= "    attribute_trusted = " . (int)$trusted . ", ";
            $query .= "    date_modified = NOW(), ";
            $query .= "    addr_modified = '$remote_addr' ";
            $query .= "WHERE attribute_id = $attribute_id ";
            $query .= "  AND attribute_user = $this->user_id ";
            
            $conn->query($query);
        } 
        else {
      
            // we need an insert
            $query  = "insert into Attribute ";
            $query .= "(attribute_user, attribute_type, attribute_trusted, ";
            $query .= " attribute_value, date_created, addr_created) ";
            $query .= "values ";
            $query .= "($this->user_id, $attributetype_id, " . (int)$trusted . ", ";
            $query .= " '$attr_value', NOW(), '$remote_addr')";
      
            $conn->query($query);
        }
        
        if ( ATID_SURNAME <= $attributetype_id and 
               $attributetype_id <= ATID_EDUPERSONENTITLEMENT ) {
              // set null value in UserAttributes
              $query = "UPDATE UserAttributes ";
              $query.= "SET $attributetype_ldapname = '$attr_value', ";
              // handle modification of commonName
              if ($attributetype_id == ATID_SURNAME or 
                  $attributetype_id == ATID_GIVENNAME) {
                  $commonName= $this->getNameOrUid();
                  $commonName= $conn->escape_string($commonName);
                  $query.= "    commonName = '$commonName', ";
              }
              $query.= "    date_modified = NOW(), ";
              $query.= "    addr_modified = '$remote_addr' ";
              $query.= "WHERE user_id = $this->user_id ";
              $conn->query($query);
        }
        return true;
  }

  /* ************************************************************* */
  /**
  * Deletes a specific attribute from this user. 
  * 
  * @param  mixed  $attributeType the attribute type ID. Either a number or the
  *     unique LDAP name of the attribute type. Example: the numeric code
  *     "1" and the LDAP name "surname" both denote
  *     the attribute type "Family Name".
  *
  */
  function deleteAttribute($attributeType) {
  
    $conn =& MySQLConnection::getConnection();
    
    if (is_int($attributeType)) {
      $atid = $attributeType;
      $query = "SELECT attribute_id, attributetype_id, attributetype_ldapname ";
      $query.=  "FROM Attribute, AttributeType, User ";
      $query.= "WHERE attribute_type = attributetype_id ";
      $query.= "  AND attribute_user = user_id ";
      $query.= "  AND user_id = $this->user_id ";
      $query.= "  AND attributetype_id = $atid ";
    } 
    else {  
      $query = "SELECT attribute_id, attributetype_id, attributetype_ldapname ";
      $query.=  "FROM Attribute, AttributeType, User ";
      $query.= "WHERE attribute_type = attributetype_id ";
      $query.= "  AND attribute_user = user_id ";
      $query.= "  AND user_id = $this->user_id ";
      $query.= "  AND attributetype_ldapname = '$attributeType'";
    }

    $conn->query($query);
    $row = $conn->fetch_array();

    if ($conn->num_rows() > 0) {

      $conn->freeresult();
      $attribute_id= $row['attribute_id'];
      $attributetype_id= $row['attributetype_id'];
      $attributetype_ldapname= $row['attributetype_ldapname'];
      
      // delete the corresponding entry from Attribute
      $query= "DELETE FROM Attribute WHERE attribute_id = $attribute_id";
      $conn->query($query); 
      
      // update modified attribute store in UserAttributes table.
      if ( ATID_SURNAME <= $attributetype_id and 
           $attributetype_id <= ATID_EDUPERSONENTITLEMENT ) {
          // set null value in UserAttributes
          $query = "UPDATE UserAttributes ";
          $query.= "SET $attributetype_ldapname = NULL, ";
          // handle modification of commonName
          if ($attributetype_id == ATID_SURNAME or 
              $attributetype_id == ATID_GIVENNAME) {
              $commonName= $this->getNameOrUid();
              $commonName= $conn->escape_string($commonName);
              $query.= "    commonName = '$commonName', ";
          }
          $query.= "    date_modified = NOW(), ";
          $remote_addr= '127.0.0.1';
          if (isset($_SERVER['REMOTE_ADDR'])) $remote_addr= $_SERVER['REMOTE_ADDR'];
          $query.= "    addr_modified = '$remote_addr' ";
          $query.= "WHERE user_id = $this->user_id ";
          $conn->query($query);
      }
    }
  }

  /**
  * Replies the display name for this user.
  *
  * If the attributes "surname" and "givenName" exist, the concatenation
  * of these two attribute values is replies.
  *
  * Otherwise, the external user ID is replied.
  *
  * @return  string  the  diplay name of the user 
  *  
  */
  function getNameOrUid() {
     // if givenName is stored return the full name else return userid
    $name= $this->uid;
    $surname= $this->getAttribute(ATID_SURNAME);
    if (! is_null($surname) ) { 
        $name= "$surname";
        $givenname= $this->getAttribute(ATID_GIVENNAME);
        if (! is_null($givenname) ) { 
            $name= "$name $givenname";
        }
    } 
    return $name;
  }

  function getCommonName() {
    $commonName= $this->getUserAttribute('commonName');
    return $commonName;
  }

  /* ************************************************************* */
  /**
  * Replies a list of role IDs for all roles this user
  * is subscribed to. 
  *
  * @return  array  an array of role IDs for all role this
  *   user is subscribed to 
  *  
  */

  function getSubscribedRoleIDs() {

    $conn =& MySQLConnection::getConnection();
    
    $query =  "SELECT role_id FROM Role, Subscription ";
    $query .= "WHERE subscription_role = role_id ";
    $query .= "  AND subscription_user = $this->user_id ";
    
    $conn->query($query);
    $res_ids= array();
    while ($row = $conn->fetch_array()) {
      $res_ids[]= $row["role_id"];
    }
    $conn->freeresult();
    return $res_ids;

  }

 
 /* ************************************************************* */
  /**
  * Replies the external user ID of the user (i.e. "portaladmin" or
  * "demo@aaitest.switch.ch".
  *
  * @return  string   the external user ID of the user 
  *  
  */
  function getUid() {
    return $this->uid;
  } 

 /* ************************************************************* */
  /**
  * Replies the internal (database) ID of the user 
  *
  * @return  int   the internal (database) ID of the user  
  *  
  */
  function getUserId() {
    return $this->user_id;
  } 
  
  function getID() {
    return $this->user_id;
  }

  /* ************************************************************* */
  /**
  * Sets the user privilege for this user. 
  *
  * The following user privileges can be set: 
  * <ul>
  *   <li>1: User
  *   <li>2: Role Administrator 
  *   <li>3: Portal Administrator 
  *   <li>4: Subscription Administrator 
  * </ul>  
  *
  * @param  int   $privilegeId  the ID of the privilege to set 
  *  
  */

  function setPrivilege($privilegeId) {
     
    $conn =& MySQLConnection::getConnection();
    $query=  "UPDATE User ";
    $query.= "SET user_privilege = $privilegeId ";
    $query.= "WHERE user_id = $this->user_id";
    $conn->query($query);

    $this->user_privilege= $privilegeId;

    $privilegeName= $this->getPrivilegeName();
    $query=  "UPDATE UserAttributes ";
    $privilegeName= $conn->escape_string($privilegeName);
    $query.= "SET privilegeName = '$privilegeName' ";
    $query.= "WHERE user_id = $this->user_id";
    $conn->query($query);

  }

  /* ************************************************************* */
  /**
  * Replies the user privilege for this user. 
  *
  * One of the following user privileges is replies: 
  * <ul>
  *   <li>1: User
  *   <li>2: Role Administrator 
  *   <li>3: Portal Administrator 
  * </ul>  
  *
  * @return  int  the ID of the current user privilege 
  *  
  */
 function getPrivilege() {
    return $this->user_privilege;

 }

  /* ************************************************************* */
  /**
  * Replies the display name of this user's user privilege.
  *
  * @return  string  the display name of this user's user privilege.
  *  
  */
  function getPrivilegeName() {
    
    $conn =& MySQLConnection::getConnection();
    $query  = "SELECT userprivilege_description ";
    $query .= "  FROM UserPrivilege ";
    $query .= " WHERE userprivilege_id = $this->user_privilege ";
    $conn->query($query);
    
    $privilegeName= null;
    if ($conn->num_rows() > 0) {
      $row = $conn->fetch_array(); 
      $conn->freeresult();
      $privilegeName= $row['userprivilege_description'];
    }
    return $privilegeName;
    
 }
    

  /* **************************************************************** */
  /**
  * Replies true if this user is subscribed to a specific role.
  * 
  * @param int  $rid   the role ID
  * 
  * @return boolean  true, if the user is currently subscribed to this role; 
  *    false otherwise 
  *
  */
  function isSubscribedTo($rid) 
  {
     $conn =& MySQLConnection::getConnection();
     
     $query =  " select subscription_user, subscription_role "
             . " from Subscription "
             . " where subscription_user     = $this->user_id " 
             . "   and subscription_role =  $rid";
    
     $conn->query($query);
     $ret = ($conn->num_rows() > 0);
     $conn->freeresult();
     return $ret;
  }         


  /* **************************************************************** */
  /** 
   * Unsubscribes this user from a specific role.
   * 
   * @param  int $rid   the role ID
   * 
   */
  function unsubscribeFromRole($rid) 
  {
     $role = Role::reconstituteById($rid);
     if (! $role) {
        trigger_error(
          "User::unsubscribeFromRole: failed to reconstitute role with id '$rid'.",
          E_USER_ERROR 
        );
        return;
     }     

     Subscription::delete(
        $this,
        $role
     );
  }         



  /* **************************************************************** */
  /** 
   * Replies a list of the roles this user is currently subscribed
   * to. 
   *
   * Replies an array of subscription records. Each subscription record
   * is itself an array, i.e. a collection of name/value pairs. The same
   * names are used as for the columns in the table Subcription and the
   * table Role. 
   * 
   * @return array  an array of subscription records 
   * 
   */
  function getSubscribedRoles() {

    $conn =& MySQLConnection::getConnection();
    
    
    $query =  " select * from Role, Subscription ";
    $query .= " where subscription_role = role_id ";
    $query .= "  and subscription_user = $this->user_id ";
    $query .= " order by role_title ";
    
    $conn->query($query);
    $ret = array();
    while ($row = $conn->fetch_array()) {         
      $ret[] = $row;
    }
    $conn->freeresult();
    return $ret;
  }
  
  /**
   * @return integer number of user's subscription.
   */
  function getSubscriptionsCount() {
      $conn =& MySQLConnection::getConnection();
      $query= " SELECT subscription_role "
            . " FROM Subscription "
            . " WHERE subscription_user = $this->user_id ";
      $conn->query($query);
      $n_subs= $conn->num_rows();
      $conn->freeresult();
      return $n_subs;
  
  }
  /**
   * Replies an array of subscription attributes:
   *   role_id, role_title
   *   subscription_state, subscription_last_login, date_created
   * 
   * @return array  an array of subscription attributes
   */
  function getSubscriptionsAttributes($orderby= null, $offset= null, $count= null) {

      $conn =& MySQLConnection::getConnection();
      $query= " SELECT role_id, role_title, "
            . "   subscription_state, subscription_last_login, Subscription.date_created "
            . " FROM Role, Subscription "
            . " WHERE subscription_role = role_id "
            . "   AND subscription_user = $this->user_id ";
     if (!is_null($orderby)) {
        if (strpos($orderby, 'subscription_state') !== false) {
            $query.= " ORDER BY subscription_state, role_title ";
        }
        elseif (strpos($orderby, 'date_created') !== false) {
            $query.= " ORDER BY Subscription.date_created DESC, role_title ";
        }
        elseif (strpos($orderby, 'subscription_last_login') !== false) {
            $query.= " ORDER BY subscription_last_login DESC, role_title ";
        }
        else {
            $query.= " ORDER BY $orderby ";        
        }
     }
     else {
        $query.= " ORDER BY role_title ";        
     }
     if (!is_null($offset) and !is_null($count)) {
        $query.= " LIMIT $offset,$count ";
     }
    
    $conn->query($query);
    $subs_attrs= array();
    while ($row = $conn->fetch_array()) {         
      $subs_attrs[]= $row;
    }
    $conn->freeresult();
    return $subs_attrs;
  }

  /* **************************************************************** */
  /** 
   * Replies a the list of roles this user can access. 
   *
   * The list of accessible roles consists of all roles
   * for which this user submitted a subscription request which
   * has been accepted by the role owner.
   * 
   * @return array  an array of subscription records 
   * 
   */
  function getAccessibleRoles() {
    
    $query =  " SELECT * FROM Role, Subscription ";
    $query .= " WHERE subscription_role = role_id ";
    $query .= "   AND subscription_user = ". $this->user_id;
    $query .= "   AND subscription_state = " . SUBSCRIPTION_STATE_ACCEPTED;
    $query .= "   AND ( role_state = " . RESOURCE_ACCESS_STATE_OPEN;
    $query .= "      OR role_state = " . RESOURCE_ACCESS_STATE_FREE;
    $query .= "      OR role_state = " . RESOURCE_ACCESS_STATE_PASSWORD . " ) ";
    $query .= " ORDER BY subscription_last_login DESC, role_title ";

    $conn =& MySQLConnection::getConnection();    
    $conn->query($query);
    $ret = array();
    while ($row = $conn->fetch_array()) {         
      $ret[] = $row;                
    }
    $conn->freeresult();
    return $ret;

}


  /* **************************************************************** */
  /** 
   * Replies the list of pending roles for this user. 
   *
   * The list of pending roles consists of all roles
   * for which this user submitted a subscription request which
   * has neither been accepted nor rejected so far by the
   * role owner. 
   *
   * Returns a two-dimensional array $info[i][field] where
   * <ul>
   *    <li> i is the i-th record found
   *    <li> field  is on the database column names in the table
   *             Subscription or Role 
   * </ul>
   *
   * @return array  an array of subscription records 
   * 
   */
  function getPendingSubscriptions() {

    $conn =& MySQLConnection::getConnection();
        
    $query =  " select * from Role, Subscription ";
    $query .= " where subscription_role = role_id ";
    $query .= "  and subscription_user = ". $this->user_id. " ";
    $query .= "  and subscription_state = " . SUBSCRIPTION_STATE_PENDING . " ";
    $query .= " order by role_title ";
    
    $conn->query($query);
    $i=0;
    $ret = array();
    while ($row = $conn->fetch_array()) {
       $ret[$i++] = $row;
    }
    $conn->freeresult();
    return $ret;
    
   } 

   /* ************************************************************* */
  /**
   * Checks whether the supplied password is correct for this
   * administrator.
   *
   * @param string  $pass   the password
   * @return boolean  true, if the supplied password is correct;
   *    false, otherwise
   *
   */
  function checkPassword($pass) {
    
    // empty passwords are not allowed
    //
    if (empty($pass)) {
      return false;
    }
    $storedPass = $this->getAttribute(ATID_LOCALPASSWORD);
    if (empty($storedPass)) {
      return false;  
    }
    return ($storedPass == $pass);
  }


  /* ************************************************************* */
  /**
   * Replies true if this user is an administrator, i.e. if he
   * is assigend the privilege {@link USER_PRIVILEGE_ID_RESOURCE_ADMINISTRATOR}
   * or {@link USER_PRIVILEGE_ID_PORTAL_ADMINISTRATOR}.
   *
   * @return boolean  true, if this user is an administrator ;
   *    false, otherwise
   *
   */  
  function isAdministrator() {
     switch($this->getPrivilege()) {
       case  USER_PRIVILEGE_ID_RESOURCE_ADMINISTRATOR:
       case  USER_PRIVILEGE_ID_SUBSCRIPTION_ADMINISTRATOR:
       case  USER_PRIVILEGE_ID_PORTAL_ADMINISTRATOR:
         return true;
          
       default:
         return false; 
     }    
  }
}  // end class

?>
