<?PHP
/*
* SNAportal mediates between one or more Authentication and Authorisation
* Infrastructures and non-AAI-enabled roles.
*
* Copyright (C) 2003 Swiss Virtual Campus (http://www.swissvirtualcampus.ch)
* Copyright (C) 2004-2006 SWITCH (http://www.switch.ch)
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* $Id: WebCTCE_StandardAPI_RegisterAdapter.inc,v 1.2 2006/12/01 13:13:17 vlry Exp $
*/

require_once("include/Logger.class.inc");
require_once("include/webct/WebCTAPI.inc");
require_once("include/webct/WebCTCERegisterAdapter.inc");

/**
* The WebCTCE_StandardAPI_RegisterAdapter process the enrollement + autosignon request
* for the WebCT CE 4 (limited Focus license) using only Standard API.
*
* @version $Revision: 1.2 $
* @author Valery Tschopp - SWITCH
*
* @package WebCTAPI
*/
class WebCTCE_StandardAPI_RegisterAdapter extends WebCTCERegisterAdapter {

    
    /**
    * Constructor:
    * Creates a new WebCT CE Register Adapter. 
    *
    * This adapter can create/update the user, enroll him to the given course 
    * and proceed to an autosignon.
    *
    * @param string $server_url The WebCT CE server basename URL.
    * @param string $ims_secret The Standard API shared secret.
    * @param string $as_secret The AutoSignon shared secret.
    */
    function WebCTCE_StandardAPI_RegisterAdapter($server_url, $api_secret, $as_secret) {
        parent::WebCTCERegisterAdapter($server_url, $api_secret, $as_secret);
    }
    
    
    /**
    * Creates/Updates a user in WebCT CE and enrolls him to course. 
    * User Shib-WISEBEDPerson-personUniqueID will be converted to a valid WebCT CE ID.
    *
    * @param string $uniqueid  WISEBEDaai Shib-WISEBEDPerson-personUniqueID.
    * @param string $courseid  WebCT CE course ID
    * @param string $givenname WISEBEDaai user given name (firstname).
    * @param string $surname   WISEBEDaai user name.
    * @param array $fields     Additional attribute to be updated.
    *
    * @return boolean   false if an error occurs, true otherwise.
    */
    function register( $uniqueid, $courseid, $givenname, $surname, $fields= array() ) {  
        $logger= Logger::getInstance();
     
        // only get first value in wrongly multi-valued fields.
        if ($pos= strpos($givenname,';')) {
            $givenname= substr($givenname,0,$pos);
            $givenname= trim($givenname);
        }
        if ($pos= strpos($surname,';')) {
            $surname= substr($surname,0,$pos);
            $surname= trim($surname);
        }
 
        // create the WebCT ID
        // default: 7853629@example.ch -> 7853629_example.ch
        $factory= WebCTIDFactory::getInstance($this->webctid_factory_);
        $webctid= $factory->createWebCTID($uniqueid, $givenname, $surname);
        if ($logger->isDebugEnable()) {
            $logger->debug("register: createWebCTID($uniqueid,$givenname,$surname): $webctid");
        }
        
        // check if a user already exists with this WebCTID

        // 1. get course/role list (<COURSE>;<ROLE>:...)
        $courses= $this->get_courses_list($webctid);
        if ($courses == null) {
            $is_new_user= true;
        }
        else {
            $is_new_user= false;
        }
        // 2. check if course/role already is in the list
        if (strpos($courses, "$courseid;") === false) {
            // not in the list, then added as student
            if ($courses==null) {
                // no other course, define
                $courses= "$courseid;S";
            }
            else {
                // other courses, append at the end
                $courses= "$courses:$courseid;S";
            }
        } 


        // add/create the user in the global database
        $error= $this->update_global_user($webctid, $courses, $surname, $givenname, $is_new_user);
        if ($error) {
            $logger->error("register: " . $this->getErrorMessage());
            return false;
        }

        // update additional global info and course info
        if ( !empty($fields) ) {
            $this->update_global_fields( $webctid, $fields );
            $this->update_course_fields( $webctid, $courseid, $fields );
        }

        return true;
    }


    /**
    * Uses the WebCT CE API to find the course/role list of the user.
    *
    * @param string $webctid    The user WebCT ID 
    * @return string            The courses list or null is the user doesn't exist. 
    *                           ie: latinumtest;D:urz001;TA:urz002;S
    * @private
    */
    function get_courses_list($webctid) {
        $logger= Logger::getInstance();
        if ($logger->isDebugEnable()) {
            $logger->debug("get_courses_list($webctid)");
        }

        // create API request processor
        $api= new WebCTIMS_API();
        $api_req_url= $this->getAPIRequestURL();
        $api->setRequestURL($api_req_url);
        $api->setSharedAPISecret($this->api_secret_);
    
        // create find request
        $find_request= WebCTAPI_Request::createFindUserGlobalDatabaseRequest($webctid);

        // process req
        $api_response= $api->invoke( $find_request );
        $content= $api_response->getServerResponse();

        $courses= null;
        if (strpos($content,"Success: WebCT ID=$webctid") !== false) {
            $elements= explode(",", $content);
            foreach($elements as $element) {
                $is_courses= strpos($element, "Courses=");
                if ($is_courses !== false) {
                    $courses= substr($element, strlen("Courses="));
                    $courses= trim($courses);
                    break;
                }
            }
        }
        if ($logger->isDebugEnable()) {
            $logger->debug("get_courses_list: courses='$courses'");
        }

        return $courses;
    }

    /**
     *
     */
    function update_global_user( $webctid, $courses, $surname, $givenname, $new_user= false ) {
        $logger= Logger::getInstance();
        if ($logger->isDebugEnable()) {
            $logger->debug("update_global_user($webctid,$courses,$surname,$givenname,$new_user)");
        }

        // create API request processor
        $api= new WebCTIMS_API();
        $api_req_url= $this->getAPIRequestURL();
        $api->setRequestURL($api_req_url);
        $api->setSharedAPISecret($this->api_secret_);

        $fields= array();
        $fields[WEBCTCE_API_FIELD_LASTNAME]= $surname;
        $fields[WEBCTCE_API_FIELD_FIRSTNAME]= $givenname;
        $fields[WEBCTCE_API_FIELD_COURSES]= $courses;

        if ($new_user) {
            // add new user
            $user_request= WebCTAPI_Request::createAddUserGlobalDatabaseRequest(
                                $webctid, 
                                $fields);
        }
        else {
            // or update existing user
            $user_request= WebCTAPI_Request::createUpdateUserGlobalDatabaseRequest(
                                $webctid,
                                $fields);
        }

        // process req
        $api_response= $api->invoke( $user_request );
        $content= $api_response->getServerResponse();
        if (strpos($content, "Error:") !== false) {
            $this->is_error_= true;
            $this->server_message_= $content;
        }

        return $this->is_error_;
    }

}

?>
