<?PHP
/*
 * Copyright (C) 2005 SWITCH (http://www.switch.ch)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: WebCTIDFactory.inc,v 1.6 2006/11/30 16:53:13 vlry Exp $
 */

/**
 * Factory to generate/create a webctid based on various parameters.
 * The returned factory implements the method createWebCTID(...) and
 * return the WebCTID as String.
 *
 * <code>
 * $impl= 'webctce'; // NET-ETHZ or WebCTVista or WebCTCE
 * $factory= WebCTIDFactory::getInstance($impl);
 * $webctid= $factory->createWebCTID($param1, $param2, ...)
 * </code>
 *
 * @version $Revision: 1.6 $
 * @author Valery Tschopp - SWITCH
 */
class WebCTIDFactory {

    /**
     * Returns the factory instance.
     * @static
     */
    function & getInstance($impl) {
        $factory= NULL;
        $impl= strtolower($impl);
        if ($impl == 'net-ethz') {
            $factory=& new WebCTIDFactory_NETETHZImpl();
        }
        else if ($impl == 'webctvista' or $impl == 'webctce6' ) {
            // uniqueid (IMS ID) == webctid
            $factory=& new WebCTIDFactory_VistaImpl();
        }
        else if ($impl == 'webctce') {
            // replace @ by _
            $factory=& new WebCTIDFactory_CEImpl();
        }
        else {
            // default WebCT CE, doesn't support @
            $factory=& new WebCTIDFactory_CEImpl();
        }
        return $factory;
    }
}

/**
 * WebCT CE factory implementation.
 */
class WebCTIDFactory_CEImpl {

    /**
     * Creates a WeCTID based on AAI Unique ID or portal UID:
     *
     * 23423@example.ch -> 23423_example.ch
     * testadmin -> snaportal_testadmin
     *
     * @param $uniqueid AAI UniqueID.
     * @return String The WebCTID derived from UniqueID.
     */
    function createWebCTID() {
        $argc= func_num_args();
        if ($argc < 1) {
            $webctid= null;
            return $webctid;
        }
        $uniqueid= func_get_arg(0);
        // check if Shib-WISEBEDPerson-personUniqueID contains an '@'
        if (strpos($uniqueid,'@')) {
            $webctid= strtr($uniqueid, '@', '_');
        } 
        else {
            $webctid= $uniqueid;
        }
        return $webctid;
    }
}


/**
 * WebCT Vista Factory implementation for AAI: webctid( uniqueid ) == uniqueid.
 */
class WebCTIDFactory_VistaImpl {
    /**
     * Creates a WeCTID same as the AAI Unique ID:
     *
     * 23423@example.ch -> 23423@example.ch
     *
     * @param $uniqueid AAI UniqueID.
     * @return string The WebCTID (same as UniqueID).
     */
    function createWebCTID() {
        $argc= func_num_args();
        if ($argc < 1) {
            $webctid= null;
            return $webctid;
        }
        return func_get_arg(0);
    }
}

/**
 * Factory implementation for NET ETHZ.
 */
class WebCTIDFactory_NETETHZImpl {

    /**
     * ISO 8859-1 table to convert accentued characters
     */
    var $ISO_8859_1_ACCENT_FILTERS= array(
        192 => 'A', 193 => 'A', 194 => 'A', 195 => 'A', 196 => 'Ae', 197 => 'A', 
        198 => 'AE', 199 => 'C', 
        200 => 'E', 201 => 'E', 202 => 'E', 203 => 'E',
        204 => 'I', 205 => 'I', 206 => 'I', 207 => 'I',
        209 => 'N', 
        210 => 'O', 211 => 'O', 212 => 'O', 213 => 'O', 214 => 'Oe', 216 => 'O',
        217 => 'U', 218 => 'U', 219 => 'U', 220 => 'Ue', 
        221 => 'Y', 223 => 'ss',
        224 => 'a', 225 => 'a', 226 => 'a', 227 => 'a', 228 => 'ae', 229 => 'a',
        230 => 'ae', 231 => 'c',
        232 => 'e', 233 => 'e', 234 => 'e', 235 => 'e',
        236 => 'i', 237 => 'i', 238 => 'i', 239 => 'i',
        241 => 'n',
        242 => 'o', 243 => 'o', 244 => 'o', 245 => 'o', 246 => 'oe', 248 => 'o', 
        249 => 'u', 250 => 'u', 251 => 'u', 252 => 'ue',
        253 => 'y', 255 => 'y'
    );

    /**
     * Creates a WeCTID based on AAI Unique ID, user 
     * first and lastname.
     *
     * @param $uniqueid AAI UniqueID.
     * @param $uniqueid First name.
     * @param $uniqueid Last name.
     * @return String The WebCTID derived from UniqueID
     */
    function createWebCTID() {
        $argc= func_num_args();
        if ($argc < 3) {
            $webctid= null;
            return $webctid;
        }
        $uniqueid= func_get_arg(0);
        $firstname= func_get_arg(1);
        $lastname= func_get_arg(2);
        if (strpos($uniqueid,'@')) {
            ereg("^([^\@]+)\@(.+)", $uniqueid, $regs);
            $webctid= $regs[2]; // domain
            $webctid.= "_" . $lastname . "_" . $firstname . "_";
            $webctid.= $regs[1]; // id
        } 
        else {
            $webctid= "snaportal_" . $lastname . "_" . $firstname . "_" . $uniqueid;
        }
       
        // filter special iso-8859-1 and other special characters
        $webctid= strtolower($webctid);
        $webctid= $this->filter_iso_8859_1($webctid);
        $webctid= $this->filter_special_chars($webctid);
        return $webctid;
    }
    
    /**
     * Removes or converts special characters from a given string.
     *
     * @param string $in  The string to filter
     * @return string     The filtered string (ISO-8859-1)
     */
    function filter_special_chars( $in ) {
        $out= $in;
        // PHP magic_quotes_gpc is default ON, quote and single quote are escaped!
        if (get_magic_quotes_gpc()) {
            $out= stripslashes( $out );
        }
        // remove: single quote are removed
        $out= str_replace("'", "", $out);
        // convert:  Jean-Pierre | Jean Pierre => Jean_Pierre
        $out= strtr($out, '- ', '__');
        return $out;
    }

    /**
     * Filters the accentued characters from a given string.
     *
     * @param string $str The string to filter
     * @return string     The filtered string (ISO-8859-1)
     */
    function filter_iso_8859_1( $in ) {
       $unknown= '#';
       $str_l= strlen( $in );
       $out= "";
       for( $i= 0; $i<$str_l; $i++ ) {
          $ord= ord( $in[$i] );
          if (array_key_exists($ord, $this->ISO_8859_1_ACCENT_FILTERS)) {
             // in conversion table, replace
             $filtered= $this->ISO_8859_1_ACCENT_FILTERS[$ord];
             $out .= $filtered;
          }
          elseif ($ord >= 160) {
             // not in conversion table, replace with ?
             $out .= $unknown;
          }
          else {
             $out .= chr( $ord );
          }
       }
       return $out;
    }

}


?>
