<?php
/* 
 * SNA-Portal mediates between one or more Authentication and Authorisation 
 * Infrastructures and non-AAI-enabled roles. 
 *
 * Copyright (C) 2003 Swiss Virtual Campus (http://www.swissvirtualcampus.ch)
 * Copyright (C) 2004-2006 SWITCH (http://www.switch.ch)
 * 
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 * 
 * $Id: WebCTIMS_IMSDataObject.class.inc,v 1.9 2006/04/06 11:35:55 vlry Exp $ 
 */

require_once("include/Logger.class.inc");

/**
 * The class WebCTIMS_IMSDataObject represents an IMS data object as it is sent
 * to WebCT in the WebCT IMS API.
 *
 * @package WebCTAPI
 * @version $Revision: 1.9 $
 *
 */
class WebCTIMS_IMSDataObject {

    /* ***************************************************************** */
    /*                    static methods                                 */
    /* ***************************************************************** */

    /**
    * Creates a new IMS data object by reading the data from a file.
    *
    * @param string $file_name  the file name
    *
    * @return IMSDataObject  an instance of an IMSDataObject initialized with
    *   the content in file $file_name
    *
    * @static 
    */      
    function createFromFile($file_name) {
        $handle = fopen($file_name, "r");
        if (! $handle) {
            trigger_error("WebCTIMS_IMSDataObject::createFromFile($file_name): could not open file",
                          E_USER_WARNING);
            return false;
        }
        $data= "";
        while(!feof($handle)) {
            $data .= fgets($handle);
        }
        fclose($handle);
        return new WebCTIMS_IMSDataObject($data);    }
  
    /* ***************************************************************** */
    /*                    private fields                                 */
    /* ***************************************************************** */
    /**
    * the XML formatted data 
    *
    * @var string
    * @access private
    */
    var $_xmlData= "";

    /**
    * the XML data encoding. Default: ISO-8859-1
    * Used by xml_parser_create(encoding) in isWellFormed() function.
    * @var string
    * @access private
    */
    var $_xmlEncoding= "ISO-8859-1";
    
    /* ***************************************************************** */
    /*                    constructor                                    */
    /* ***************************************************************** */
    
    /**
    * constructor
    *
    * @param string $data  the XML data
    */
    function WebCTIMS_IMSDataObject($data = "") {
       if (! empty($data)) {
         $this->setData($data);
       }
    }
  
    /* ***************************************************************** */
    /*                    public methods                                 */
    /* ***************************************************************** */
    
    /**
    * sets the data for the IMS data object. $data is an XML formatted
    * IMS object according to the IMS specification. 
    *
    * @param string $data   the data 
    */ 
    function setData($data) {
      $this->_xmlData = $data;
    }

    /* ***************************************************************** */
    /**
    * Replies the data hold by  this IMS data object
    *
    * @return string the data in XML format 
    */    
    function getData() {
       return $this->_xmlData;
    }

    /**
     * sets the XML encoding for the XML parser isWellFormed() check.
     */
    function setEncoding($encoding) {
        $this->_xmlEncoding= $encoding;
    }

    /**
     * returns the XML encoding.
     */
    function getEncoding() {
        return $this->_xmlEncoding;
    }

    /**
     * returns true iff the string is UTF-8 encoded.
     * @param string The string to check.
     */
    function is_utf8( $string ) {
        return utf8_encode( utf8_decode( $string )) == $string;
    }
    
    /**
    * Replies the checksum of the data hold by this IMS data object.
    * The return value is an int. 
    *
    * The checksum MUST be compute on ISO-8859-1 encoded string,
    * because WebCT is stupid!
    *
    * @return int  the checksum 
    */        
    function getChecksumAsInt() {
       if ($this->isEmpty()) {
         return 0;  
       }
       $iso_data= $this->_xmlData;
       // convert data to ISO-8859-1 if data is UTF-8 encoded.
       if ( $this->is_utf8($iso_data) ) {
            $iso_data= utf8_decode($iso_data);
       }
       $length= strlen($iso_data);
       $ret= 0;
       for($i= 0; $i<$length; $i++) {
            $ret+= ord($iso_data[$i]);
       }
       //XXX print "XXX: getChecksumAsInt(): $ret\n";
       return $ret;
    }

    /**
    * Replies the checksum of the data hold by this IMS data object.
    * The return value is a string, if checksum == 0, then the empty
    * string is returned. 
    *
    * @return string  the checksum 
    * @see getChecksumAsInt()
    */        
    function getChecksumAsString() {
       $ret = $this->getChecksumAsInt();
       return ($ret == 0) ? "" : "$ret";
    }
    
    /* ***************************************************************** */
    /**
    * XXX: XML validation disabled!
    * Replies always true. 
    * But a WARNING is logged if the XML parser find an error.
    *
    * @return boolean true. 
    */                
    function isWellFormed() {
       if ($this->isEmpty()) {
         return true;
       }
       $encoding= $this->getEncoding();
       // TODO: handle PHP4 and PHP5 difference with XML validation (encoding)
       $xml_parser= xml_parser_create($encoding);
       $ok= xml_parse($xml_parser, $this->getData(), true);
       if (! $ok) {
            $err_code= xml_get_error_code($xml_parser);
            $xml_error= xml_error_string($err_code);
            $err_line= xml_get_current_line_number($xml_parser);
            $err_column= xml_get_current_column_number($xml_parser);
            $logger= Logger::getInstance();
            $logger->warn("WebCTIMS_IMSDataObject:isWellFormed:xml_parse: [$err_line:$err_column] $xml_error");
            //XXX print "XXX: WebCTIMS_IMSDataObject:isWellFormed: [$err_line:$err_column] $xml_error\n";
       }
       xml_parser_free($xml_parser);
       // always true. PHP 4 and PHP 5 xml_parser doesn't validate the same.
       return true;
    }


    /* ***************************************************************** */
    /**
    * Replies true if the data hold by this IMS data object is empty
    *
    * @return boolean true, if the data hold by this IMS data object is empty
    */                    
    function isEmpty() {
       return empty($this->_xmlData);
    }
    
    /* ***************************************************************** */
    /**
    * Clears the data hold by this IMS data object 
    *
    */                          
    function clear() {
      $this->_xmlData = "";
    }
    
}

?>
