<?php
/* ****************************************************************** */
/*
 * SNA-Portal mediates between one or more Authentication and Authorisation
 * Infrastructures and non-AAI-enabled roles.
 * Copyright (C) 2003  Swiss Virtual Campus (http://www.swissvirtualcampus.ch)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 * $Id: WebCTIMS_Request.class.inc,v 1.13 2006/11/21 11:07:06 vlry Exp $
 */
/* ****************************************************************** */

require_once("include/Logger.class.inc");
require_once("include/webct/WebCTIMS_Constants.inc");
require_once("include/webct/WebCTIMS_IMSDataObject.class.inc");

require_once("include/pre-php-4.2.inc");

/**
 * The list of WebCT IMS paramters taken into consideration for
 * the MAC computation. 
 *
 * @var array 
 * @access private
 */
$_WEBCT_IMS_API_MAC_PARAM_NAMES = array(
     WEBCT_IMS_PARAM_ACTION,
     WEBCT_IMS_PARAM_OPTION,
     WEBCT_IMS_PARAM_TIMESTAMP,
     WEBCT_IMS_PARAM_TYPE,
     WEBCT_IMS_PARAM_SCTMODE,
     WEBCT_IMS_PARAM_DATASOURCE,
     WEBCT_IMS_PARAM_TARGET,
     WEBCT_IMS_PARAM_ID,
     WEBCT_IMS_PARAM_SOURCE,
     WEBCT_IMS_PARAM_COURSEID,
     WEBCT_IMS_PARAM_GLOBALID,
     WEBCT_IMS_PARAM_INSTRUCTOR,
     WEBCT_IMS_PARAM_INLINEMODE,
     WEBCT_IMS_PARAM_CHECKSUM
);

/**
* WebCTIMS_Request represents a request to the WebCT IMS API.
*
* @version $Revision: 1.13 $
* @package WebCTAPI
*
*/
class WebCTIMS_Request {

   /* ***************************************************** */
   /*                  static factory methods               */
   /* ***************************************************** */

   /* ***************************************************** */
   /**
   * Creates a WebCT IMS request of type "export".
   *
   * The request is initialized as follows:
   * <ul>
   *  <li>{@link WEBCT_IMS_PARAM_ACTION} is set to {@link WEBCT_IMS_ACTION_EXPORT}</li>
   *  <li>{@link WEBCT_IMS_PARAM_OPTION} is set to {@link WEBCT_IMS_OPTION_SNAPSHOT}</li>
   *  <li>{@link WEBCT_IMS_PARAM_TIMESTAMP} is set to the current time</li>
   * </ul>
   *
   * @return WebCTIMSAPIRequest the initiaized "export" request.
   *
   * @static
   */
   function createExportRequest($option= WEBCT_IMS_OPTION_SNAPSHOT) {

     // create the request 
     //
     $request = new WebCTIMS_Request();

     // init the request parameters with default values
     //     
     $request->setParameter(WEBCT_IMS_PARAM_ACTION, WEBCT_IMS_ACTION_EXPORT);
     $request->setParameter(WEBCT_IMS_PARAM_OPTION, $option);
     $request->setParameter(WEBCT_IMS_PARAM_TIMESTAMP, mktime());
     
     return $request;     
   }

   /* ***************************************************** */
   /**
   * Creates a WebCT IMS request of type "import".
   *
   * The request is initialized as follows:
   * <ul>
   *  <li>{@link WEBCT_IMS_PARAM_ACTION} is set to {@link WEBCT_IMS_ACTION_IMPORT}</li>
   *  <li>{@link WEBCT_IMS_PARAM_OPTION} is set to {@link WEBCT_IMS_OPTION_UNRESTRICT}</li>
   *  <li>{@link WEBCT_IMS_PARAM_TIMESTAMP} is set to the current time</li>
   *  <li>{@link WEBCT_IMS_PARAM_ADAPTOR} is set to {@link WEBCT_IMS_ADAPTOR_IMS}</li>   
   * </ul>
   *
   * @return WebCTIMSAPIRequest the initiaized "import" request.
   *
   * @static
   */
   function createImportRequest() {
     
     // create the request 
     //
     $request = new WebCTIMS_Request();
     
     // init the request parameters with default values
     //
     $request->setParameter(WEBCT_IMS_PARAM_ACTION, WEBCT_IMS_ACTION_IMPORT);
     $request->setParameter(WEBCT_IMS_PARAM_OPTION, WEBCT_IMS_OPTION_UNRESTRICT);
     $request->setParameter(WEBCT_IMS_PARAM_TIMESTAMP, mktime());
     $request->setParameter(WEBCT_IMS_PARAM_ADAPTOR, WEBCT_IMS_ADAPTOR_IMS);
     
     return $request;     
   }


   /* ***************************************************** */
   /**
   * Creates a WebCT IMS request of type "configure".
   *
   * The request is initialized as follows:
   * <ul>
   *  <li>{@link WEBCT_IMS_PARAM_ACTION} is set to {@link WEBCT_IMS_ACTION_CONFIGURE}</li>
   *  <li>{@link WEBCT_IMS_PARAM_TIMESTAMP} is set to the current time</li>
   * </ul>
   *
   * @return WebCTIMSAPIRequest the initiaized "configure" request.
   *
   * @static
   */
   function createConfigureRequest() {
     
     // create the request 
     //
     $request = new WebCTIMS_Request();
     
     // init the request parameters with default values
     //
     $request->setParameter(WEBCT_IMS_PARAM_ACTION, WEBCT_IMS_ACTION_CONFIGURE);
     $request->setParameter(WEBCT_IMS_PARAM_TIMESTAMP, mktime());
     
     return $request;     
   }


   /* ***************************************************** */
   /*                  private members                      */
   /* ***************************************************** */
   
   /**
   * the request parameters 
   *
   * @var array 
   * @access private 
   */ 
   var $_parameters = array();
   
   /**
   * the IMS data object
   *
   * @var IMSDataObject
   * @access private 
   */    
   var $_ims_data_object = NULL;

   /**
   * The IMS API shared secret 
   *
   * @var string
   * @access private 
   */    
   var $_shared_secret = false;
   
   /* ***************************************************** */
   /*                  constructor                          */
   /* ***************************************************** */
   
   /**
   * constructor  
   *
   * @param  array $parameters  an array of request parameters 
   */
   function WebCTIMS_Request($parameters = NULL) {
     if (! is_null($parameters)) {
        $this->setParameters($parameters);    
     }
   }
    
   /* ***************************************************** */
   /*                  setters and getters                  */
   /* ***************************************************** */
    
   /**
   * sets the value of a  request parameter
   *
   * @param string $name  the name of a request parameter
   *
   * @param  string  $value the value of the request parameter
   */
   function setParameter($name, $value) {
      $name=strtoupper($name);
      $this->_parameters[$name] = $value; 
   }
   
   /* ***************************************************** */
   /**
   * Replies the value of a request parameter.
   *
   * @param string $name  the name of a request parameter
   *
   * @return  string the value of the request parameter
   */
   function getParameter($name) {
      $name=strtoupper($name);
      if (! isset($this->_parameters, $name)) {
        $value= null;
        return $value; // null
      } 
      return $this->_parameters[$name];
   }
   
   /* ***************************************************** */
   /**
   * Sets the request parameters.
   *
   * @param  array $params the request parameters
   */   
   function setParameters($params) {
      if (! $params || ! is_array($params)) {
        trigger_error(
          "WebCTIMSAPIRequest::setParameters: expected non-null array in parameter \$param, got '$params'",
          E_USER_WARNING
        );
        return; 
      }
      $this->_parameters = array();
      foreach($params as $name => $value) {
        $name = strtoupper($name);
        $this->_parameters[$name] = $value; 
      }     
   }

   /* ***************************************************** */
   /**
   * Sets the IMS data object.
   *
   * @param  WebCTIMS_IMSDataObject  $data_object   the IMS data object; can be NULL.
   *
   */      
   function setIMSDataObject($data_object) {
     if (is_null($data_object)) {
       $this->_ims_data_object  = NULL;
     }
     if (! is_object($data_object) || ! is_a($data_object, "WebCTIMS_IMSDataObject")) {
        trigger_error(
          "WebCTIMSRequest::setIMSDataObject: expected parameter \$data_object of type 'WebCTIMS_IMSDataObject', got '$data_object'",
          E_USER_ERROR
        );  
     }
     $this->_ims_data_object = $data_object;  
   }

   /* ***************************************************** */
   /**
   * Replies the IMS data object
   *
   * @return  WebCTIMS_IMSDataObject the IMS data object; NULL, if no IMS data object
   *   is set 
   *
   */         
   function getIMSDataObject() {
     return $this->_ims_data_object;  
   }

   /* ***************************************************** */
   /**
   * Sets the shared secret.
   *
   * @param string $sharedSecret the IMS API shared secret 
   *
   */         
   function setSharedSecret($sharedSecret) {
     return $this->_shared_secret = $sharedSecret;  
   }

   /* ***************************************************** */
   /**
   * Replies the shared secret 
   *
   * @return string the IMS API shared secret 
   *
   */         
   function getSharedSecret() {
     return $this->_shared_secret;  
   }

   /* ***************************************************** */
   /*                 checksum and mac calculation          */
   /* ***************************************************** */
   
   /**
   * Computes the IMS data object checksum and the Message
   * Authentication Code (MAC) for this request and sets
   * the respective request parameters.
   *
   * The IMS data object checksum is saved in the parameter
   * {@link WEBCT_IMS_PARAM_CHECKSUM}. It can be retrieved
   * with the method {@link getParameter() }.
   *
   * The MAC is saved in the parameter
   * {@link WEBCT_IMS_PARAM_AUTH}. It can be retrieved
   * with the method {@link getParameter() }. 
   *
   * @see getParameter()
   * @see WebCTIMS_IMSDataObject::getChecksumAsString()
   *
   * @global $_WEBCT_IMS_API_MAC_PARAM_NAMES
   *
   */   
   function refreshChecksumAndMAC() {
      global $_WEBCT_IMS_API_MAC_PARAM_NAMES;
      
      $mac_data = ""; 
     
      for($i=0; $i < count($_WEBCT_IMS_API_MAC_PARAM_NAMES ); $i++) {       
        if (isset($this->_parameters[$_WEBCT_IMS_API_MAC_PARAM_NAMES[$i]])) {
           $mac_data .= $this->_parameters[$_WEBCT_IMS_API_MAC_PARAM_NAMES[$i]];
        }
      } 
      if (! is_null($this->getIMSDataObject())) {
        $ims = $this->getIMSDataObject();
        $checksum = $ims->getChecksumAsString();
        $this->setParameter(WEBCT_IMS_PARAM_CHECKSUM, $checksum);
      } else {
        $checksum = "";
        unset($this->_parameters[WEBCT_IMS_PARAM_CHECKSUM]);
      }
      
      $mac_data .= $checksum; 
      $cs = 0;
      for ($i = 0; $i < strlen($mac_data); $i++) {
         $cs += ord($mac_data[$i]);
      }
      $secret = $this->getSharedSecret();
      if ($secret == false) {
         trigger_error(
           "WebCTIMS_Request::refreshChecksumAndMAC: potential configuration error - shared secret is not set",
           E_USER_ERROR
         );
      }
      $mac = strtoupper(md5("$cs" . "$secret"));
      
      // escape quotes for building up command line
      //
      //$mac_data = str_replace("\"", "\\\"", $mac_data);
      //$cmd = WEBCT_IMS_GET_MD5_EXECUTABLE . " " . WEBCT_IMS_SHARED_SECRET_FILE . " \"$mac_data\"";
      //$mac = `$cmd`;
      $this->setParameter(WEBCT_IMS_PARAM_AUTH, $mac);
   }   
   
    
   /**
   * Replies a suitable MIME mulipart boundary string 
   *
   * @return string a suitable MIME multipart boundary string 
   *
   * @access private   
   */            

   function _createMimeMultipartBoundary() {
     return "WebCT_Enterprise_API_Boundary";    
   }
   

   /* ***************************************************** */
   /*               MIME multipart output                   */
   /* ***************************************************** */
   
   /**
   * Writes the request as MIME mulitpart message.
   *
   * @param handle  $handle   the IO handle (file, socket) the
   *    output is written to
   *
   * @param string $mime_boundary  the MIME boundary string to
   *    be used  
   */
   function writeAsMimeMultipart($handle) { 
    
      $mime_multipart_boundary = $this->_createMimeMultipartBoundary() ;
      $body = "";
      
      // refresh checksum and MAC
      //
      $this->refreshChecksumAndMAC();

      // write the request parameters 
      //
          
	  //REFACTORING: subclass could implement customized getParametersAsMimeMultipart 
	  // get parameters as mime multipart
	  $body.= $this->getParametersAsMimeMultipart( $mime_multipart_boundary );
      
	  //REFACTORING: subclass could implement customized  getImsXmlAsMimeMultipart 
	  // get XML file as mime multipart
	  $body.= $this->getImsXmlAsMimeMultipart( $mime_multipart_boundary );
      
      // close the MIME multipart message
      // 
      $body.= "--$mime_multipart_boundary--\r\n";
      $body_length= strlen( $body );
			
      
      // write the leading MIME mulitpart header
      //
      $header=  "Content-Length: $body_length\r\n";
      $header.= "Content-Type: multipart/form-data; boundary=$mime_multipart_boundary\r\n";
      $header.= "\r\n";

      //XXX
      $logger=& Logger::getInstance();
      if ($logger->isDebugEnable()) {
         $logger->debug($header);
         $logger->debug($body);
      }
      //print "---IMS_Request---HEADER BEGIN---\n";
      //print $header;
      //print "---IMS_Request---HEADER END---\n";
      //print "---IMS_Request---BODY BEGIN---\n";
      //print $body;
      //print "---IMS_Request---BODY END---\n";


      // write header and body to handle     
      fwrite($handle, $header);
      fwrite($handle, $body);
      
   }

   /**
    * REFACTORING: Returns the IMS XML data object as a file MIME multipart.
    *
    * @param string The MIME boundary tag.
    * @return string The MIME multipart content.
    */  
   function getImsXmlAsMimeMultipart( $mime_boundary ) {
	  $mime= "";
   	  if (! is_null($this->getIMSDataObject())) {
        $ims_data_obj= $this->getIMSDataObject();
        // write the IMS data object 
        $mime.= "--$mime_boundary\r\n";
        $mime.= "Content-Disposition: form-data; name=\"FILENAME\"; filename=\"sna-portal-request.xml\"\r\n";
        $mime.= "Content-Type: text/xml\r\n";
        $mime.= "\r\n";
        $mime.= $ims_data_obj->getData() . "\r\n";            
      }
   	  return $mime;
   }


   /**
    * REFACTORING: Returns the request parameters as MIME multipart.
    *
    * @param string The MIME boundary tag.
    * @return string The MIME multipart content.
    */
   function getParametersAsMimeMultipart( $mime_boundary ) {
      $mime= "";		
      foreach($this->_parameters as $name => $value) {
         $mime .= "--$mime_boundary\r\n";  
         $mime .= "Content-Disposition: form-data; name=\"$name\"";
         $mime .= "\r\n\r\n";
         $mime .= "$value\r\n";                  
      }      
      return $mime;
   	
   }
}


?>
