<?PHP
/*
* SNAportal mediates between one or more Authentication and Authorisation
* Infrastructures and non-AAI-enabled roles.
*
* Copyright (C) 2003 Swiss Virtual Campus (http://www.swissvirtualcampus.ch)
* Copyright (C) 2004-2006 SWITCH (http://www.switch.ch)
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* $Id: WebCTVistaRegisterAdapter.inc,v 1.29 2007/03/06 12:13:40 vlry Exp $
*/

require_once("include/webct/WebCTVistaLoginAdapter.inc");
require_once("include/webct/WebCTVistaIMS.inc");

/**
* The WebCTVistaRegisterAdapter process the enrollement + autosignon request.
*
* @version $Revision: 1.29 $
* @author Valery Tschopp - SWITCH
*
* @package WebCTAPI
*/
class WebCTVistaRegisterAdapter extends WebCTVistaLoginAdapter {

    /** WebCTID factory to use */
    var $webctid_factory_name_= 'WebCTVista'; // default uses WebCTVista factory (AAI UniqueID -> WebCTID)
    
    /** IMS person.extension.customcolumns.column */
    var $person_extension_columns_= array();
    
    /**
    * Constructor:
    * Creates a new WebCT Vista Register Adapter. This adapter will create/update the user,
    * enroll him to the given course and proceed to an autosignon.
    *
    * @param string $server_url The WebCT Vista server basename URL.
    * @param string $glcid The target institution glcId.
    * @param string $ims_secret The IMS API shared secret.
    * @param string $standard_secret The Standard API shared secret.
    */
    function WebCTVistaRegisterAdapter($server_url, $glcid, $ims_secret, $standard_secret) {
        parent::WebCTVistaLoginAdapter($server_url, $glcid, $standard_secret, $ims_secret);
    }
    
    /**
     * Sets the WebCTID factory name to use.
     * @param string WebCTIDFactory name.
     */
    function setWebCTIDFactoryName($impl_name) {
        $this->webctid_factory_name_= $impl_name;
    }

    /**
     * Returns the WebCTID factory name to use.
     * @return string WebCTIDFactory name.
     */
    function getWebCTIDFactoryName() {
        return $this->webctid_factory_name_;
    }

    /**
     * Add a person custom column (in IMS person.extension.customcolumns)
     */
    function addIMSPersonExtensionColumn( $name, $value ) {
        // the column name and value must be UTF-8 encoded.
        if ($this->is_utf8($name)) { $utf8_name= $name; }
        else { $utf8_name= utf8_encode($name); }
        if ($this->is_utf8($value)) { $utf8_value= $value; }
        else { $utf8_value= utf8_encode($value); }

        $this->person_extension_columns_[$utf8_name]= $utf8_value;
    }

    /**
     * Sets IMS person.sourcedid.source and membership.member.sourcedid.source
     */
    function setIMSPersonSource($ims_source) {
        $this->setIMSParameter('WEBCTVISTA_IMS_PERSON_SOURCE', $ims_source );
        $this->setIMSParameter('WEBCTVISTA_IMS_MEMBER_SOURCE', $ims_source );
    }
        
    /**
    * Creates/Updates a user in WebCT Vista and enrolls him to course. 
    * User webctid will be used as WebCT id.
    *
    * If the courseid is null, then only a person record will be generated,
    * without the enrollement.
    *
    * @param string $uniqueid  WISEBEDaai Shib-WISEBEDPerson-personUniqueID.
    * @param string $courseid  WebCT Vista course ID (long).
    * @param string $givenname User given name (firstname).
    * @param string $surname   User family name.
    * @param string $home_org  WISEBEDaai home organization.
    * @param string $email     User email (optional).
    * @return boolean   false if an error occurs, true otherwise.
    */
    function register( $uniqueid, $courseid, $givenname, $surname, $home_org, $email= null ) {        

        // which XML encoding?
        $encoding= $this->get_xml_encoding();

        // get IMS XML request (with header)
        $ims_xml_req= $this->get_ims_xml_req($uniqueid, $courseid, $givenname, $surname, $home_org, $email);

        // create IMS data object
        $ims_data_obj= new WebCTIMS_IMSDataObject( $ims_xml_req );
        // set encoding for XML validation
        $ims_data_obj->setEncoding( $encoding );
        //XXX: isWellFormed always return true! (PHP4 and PHP5 difference problem)
        if (! $ims_data_obj->isWellFormed()) {
            $this->is_error_= true;
            $this->server_message_= "WebCTIMS_IMSDataObject not well formed: $ims_xml_req";
            $logger= Logger::getInstance();
            $logger->error($this->server_message_);
            return false;   
        }
        
        // create import req
        $this->is_error_= false;
        $glcid= $this->getGlcId();
        $ims_import_req= WebCTVistaIMS_Request::createImportRequest($glcid, $ims_data_obj);
        $ims= new WebCTIMS_API();
        $ims_req_url= $this->getIMSRequestURL();       
        $ims->setRequestUrl($ims_req_url);
        $ims_secret= $this->getIMSSecret();
        $ims->setSharedAPISecret($ims_secret);
        
        // process import req
        $response= $ims->invoke( $ims_import_req );

        $this->server_message_= $response->getServerResponse();
        if (! $response->isOk() ) {
            $this->is_error_= true;
            //TODO: parse server response for better answer
            $logger= Logger::getInstance();
            $logger->error("WebCTVistaRegisterAdapter:register:\n" . $this->server_message_);

            return false;   
        }

        return true;
    }


    /**
     * Returns the IMS XML request (Person and Membership) as IMS Enterprise.
     *
     * If the $courseid is null, then only a person without membership XML request 
     * will be generated.
     *
     * @return string IMS XML request.
     * @protected
     */
    function get_ims_xml_req($uniqueid, $courseid, $givenname, $surname, $home_org, $email ) {

        // only get first value in wrongly multi-valued fields.
        if ($pos= strpos($givenname,';')) {
            $givenname= substr($givenname,0,$pos);
            $givenname= trim($givenname);
        }
        if ($pos= strpos($surname,';')) {
            $surname= substr($surname,0,$pos);
            $surname= trim($surname);
        }

        // IMS user parameters
        $ims_person_params= array();
        $ims_person_params['Shib-WISEBEDPerson-personUniqueID']= $uniqueid;
        // givenname and surname must be sent UTF-8 encoded.
        if ( $this->is_utf8($givenname)) {
            // already UTF-8
            $ims_person_params['givenName']= $givenname;
        } 
        else {
            // not UTF-8, convert
            $ims_person_params['givenName']= utf8_encode( $givenname );        
        }
        if ( $this->is_utf8($surname)) {
            // already UTF-8
            $ims_person_params['surname']= $surname;
        } 
        else {
            // not UTF-8, convert
            $ims_person_params['surname']= utf8_encode( $surname );        
        }
        // add email if available
        if (! empty($email)) {
            $ims_person_params['email']= $email;
        }
        // add homeOrg if available
        if (! empty($home_org)) {
            $ims_person_params['Shib-WISEBEDPerson-homeOrganization']= $home_org;
        }
        
        // set extension custom columns if any
        if (!empty($this->person_extension_columns_)) {
            $this->setIMSParameter('WEBCTVISTA_IMS_PERSON_EXTENSION_COLUMNS',
                                   $this->person_extension_columns_);
        }
        
        
        // WebDAV problem: check if new user, if not don't set password.
        $webctid= $this->get_webctid( $uniqueid );
        if (is_null($webctid)) {
            // new user -> create a new WebCTID
            $factory_name= $this->getWebCTIDFactoryName();
            // WebCTIDFactory:
            //  - NET-ETHZ: create a webctid for this user based on givenname, surname and uniqueid
            //              112358@switch.ch, Valry, Tschopp -> switch.ch_tschopp_valery_112358
            //  - WebCTVista: 7853629@example.ch -> 7853629@example.ch
            //  - WebCTCE: 7853629@example.ch -> 7853629_example.ch
            $factory= WebCTIDFactory::getInstance($factory_name);
            $webctid= $factory->createWebCTID($uniqueid, $givenname, $surname);
            $password= WebCTVistaIMS_SwitchAAIAdapting::createPassword();
            $this->setIMSParameter('WEBCTVISTA_IMS_PERSON_EXTENSION_PASSWORD',
                                   $password);
        } 
        else {
            // existing user
            // WebDAV problem: don't set password.
            $this->setIMSParameter('WEBCTVISTA_IMS_PERSON_EXTENSION_PASSWORD', '');
        }
        $ims_person_params['webctid']= $webctid;

        // prepare create/update user request
        $ims_person= WebCTVistaIMS_SwitchAAIAdapting::prepareCreateOrUpdateUserRequest(
            $ims_person_params, 
            $this->ims_params_ );
       
        // don't generate the membership if $coursid is null
        $ims_membership= null;
        if (!empty($courseid)) { 
            // IMS membership parameters
            $ims_person_params['courseid']= $courseid;
            // prepare create/update membership request
            $ims_membership= WebCTVistaIMS_SwitchAAIAdapting::prepareCreateOrUpdateMembershipRequest(
                $ims_person_params, 
                $this->ims_params_ );
        }
    
        // IMS enterprise envelop
        $ims_enterprise= WebCTVistaIMS_SwitchAAIAdapting::prepareEnterpriseDataObject($this->ims_params_);
        $ims_enterprise->addChild($ims_person);
        if (!empty($ims_membership)) {
            $ims_enterprise->addChild($ims_membership);
        }

        // convert to IMS XML request (with header)
        $xml_encoding= $this->get_xml_encoding();
        $enterprise_xml= $ims_enterprise->toXML(true,$xml_encoding);
        $logger=& Logger::getInstance();
        $logger->debug("WebCTVistaRegisterAdapter:get_ims_xml_req:\n$enterprise_xml");
        return $enterprise_xml;
    
    }

    /**
     * Returns the XML encoding used in IMS data
     * @return string The XML encoding: UTF-8
     * @protected
     */
    function get_xml_encoding() {
        return "UTF-8";
    }
   
   
    /**
     * returns true iff the string is UTF-8 encoded.
     * @param string The string to check.
     */
    function is_utf8( $string ) {
        return utf8_encode( utf8_decode( $string )) == $string;
    }
}

?>
