<?php
/*
 * SNA-Portal mediates between one or more Authentication and Authorisation
 * Infrastructures and non-AAI-enabled roles.
 *
 * Copyright (C) 2003  Swiss Virtual Campus (http://www.swissvirtualcampus.ch)
 * Copyright (C) 2004-2005 SWITCH (http://www.switch.ch)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Id: subscribe.php,v 1.36 2007/03/06 17:22:02 vlry Exp $
 */

/**
* This page creates a subscription request for the current user and
* a given role.
*
* <b>Request Parameters</b>
* <ul>
*  <li>rid=<rid>: MANDATORY. An internal role ID. <rid> must be a valid
*     integer > 0. The role ID <rid> must exist, must be visible, and
*     must be open.
*   </li>
*  <li>action=<action>: OPTIONAL. If missing, just displays a confirmation
*     page with details about the subscription. If <action>=do_subscribe,
*     persists the subscription and redirects to a follow-up page.
*     If <action>=do_cancel, does not save the subscription to the database
*     and redirects to the calling page.
*   </li>
*   <li>caller=<url>: OPTIONAL. If present, this is the URL the user is
*      redirected when he or she cancels the subscription. If absent, the
*      user is redirected to {@link SUBSCRIBE_POST_CANCEL_URL}.
*   </li>
*   <li>redirect=<url>: OPTIONAL. If present, this is the URL the user is
*      redirected when he sucessfully subscribe to the role. 
*      If absent, the user is redirected to 
*      {@link SUBSCRIBE_POST_SUBSCRIPTION_URL}.
*   </li>
**
* </ul>
*
* <b>Configuration Parameters</b>
* The following configuration paramters can be used in the global configuration file
* {@link constants.inc}:
* <ul>
*   <li>SUBSCRIBE_POST_CANCEL_URL: the URL a user is redirected to after having cancelled
*      a subscription on this page
*   </li>
*
*   <li>SUBSCRIBE_POST_SUBSCRIPTION_URL: the URL a user is redirected to after having submitted
*       a subscription request from this page 
*   </li>
* </ul>
*
* @version $Revision: 1.36 $
* @package UserPortal
*/


// bootstrap application
// 
require_once("include/application.inc");

/**
 * The default caller URL. Used if request parameter "caller" is missing.
 */
if (!defined("SUBSCRIBE_POST_CANCEL_URL")) {
  define('SUBSCRIBE_POST_CANCEL_URL',
          PORTAL_BASE_URL . "/user/all-roles.php");
}

/**
 * The post-subscription URL, i.e. the URL the user is redirected to
 * after having subscribed.
 *
 */
if (! defined("SUBSCRIBE_POST_SUBSCRIPTION_URL")) {
   define('SUBSCRIBE_POST_SUBSCRIPTION_URL',
           PORTAL_BASE_URL . "/user/my-roles.php");
}


// start session and ensure access restrictions
Session::ensureUserSessionValid();
I18N::initLocale();
I18N::loadMessageBundle("user/subscribe");


/* ***************************************************************** */
/**
 * Persists the supplied custom attribute values to the database.
 *
 * @param Subscription $subscription  the subsription obect  
 *
 * @access private
 *
 */
function persistRequiredAttributes(&$subscription){
 
    $user =& $subscription->getUser(); 
    $role =& $subscription->getRole();

    $attributeIDs = $role->getRequiredAttributeIDs();    
      
    foreach($attributeIDs as $atid) {
        $atid = (int)$atid;

        // only untrusted attributes are send in REQUEST
        if (isset($_REQUEST["attribute$atid"])) { 
            $formValue = $_REQUEST["attribute$atid"];     
            $isTrusted = $user->isAttributeTrusted((int)$atid);  
              
            if (! $isTrusted) {
                $subscription->persistRequiredAttribute($atid, $formValue); 
            }
        }
    }
}


/* ***************************************************************** */
/**
 * Checks whether all required attribute values are filled in. 
 *
 * @param array &$errorMessages  an OUT-parameter with the error messages 
 * 
 * @access private 
 */

function validate_inputs(&$errorMessages,&$role) {
  $valid= true;
  foreach($_REQUEST as $name => $value) {
    if (ereg("^attribute([0-9]+)$", $name,$regs)) {
       $value= trim($value);
       if (empty($value)) {
         $errorMessages[$name] = I18N::getMessage("subscribe.err.attr-value-empty");
         $valid= false;
       }  
    }
  }
  if ($role->isPasswordProtected()) {
    if (isset($_REQUEST['subscribe_password'])) {
        $subs_password= trim($_REQUEST['subscribe_password']);
        $res_password= $role->getPassword();
        if ($subs_password !== $res_password) {
            $errorMessages['subscribe_password']= I18N::getMessage('subscribe.err.wrong-password');
            $valid= false;
        }
    }
    else {
        $errorMessages['subscribe_password']= I18N::getMessage('subscribe.err.missing-password');
        $valid= false;
    }
  }   
  return $valid; 
}

//
// make sure the current user exists
//
$uid= Session::getUID();
$user=& User::reconstituteByUid($uid);
if (is_null($user)) {
    ErrorHandling::fatalErrorAndExit("No user with UID($uid).");
}

// check request parameters
//
if (!isset($_REQUEST["rid"])) 
{
   ErrorHandling::fatalErrorAndExit("Request parameter 'rid=' is missing");
}
$rid= $_REQUEST["rid"];    // the role ID 
if ( ! is_numeric($rid) || $rid < 0) {
   ErrorHandling::fatalErrorAndExit("Request parameter has an illegal value: 'rid=$rid'");
}

$role =& Role::reconstituteById($rid);
if (is_null($role)) {
  ErrorHandling::fatalErrorAndExit("No role with ID($rid).");
}

// check whether the role can be subscribed
//
if (! $role->canBeSubscribed()) {
   ErrorHandling::fatalErrorAndExit("Role ID($rid) can't be subscribed.");
}

// action dispatching 
//
$errorMessages = array();

if (isset($_REQUEST["do_cancel"])){
  if (! empty($_REQUEST["caller"])){
      $caller = $_REQUEST["caller"];
   } else {
      $caller = SUBSCRIBE_POST_CANCEL_URL;
   }    
   header("Location: $caller");
   exit;  
} 
elseif (isset($_REQUEST["do_subscribe"])){
  if (validate_inputs(&$errorMessages,&$role)) {

    $subscription =& Subscription::create(&$user, &$role);
    persistRequiredAttributes(&$subscription);

    // free subscription role
    // or password protected role, password already checked in validate_inputs...
    if ($role->isFree() or $role->isPasswordProtected()) {
        $subscription->accept();
    }
    else {
        // send out notification message to role provider and administrators.
        $message= new Messaging();
        if ($message->is_enabled()) {
            /* get email of role and user provider */
            $user_attrs= $user->getUserAttributes(ATTR_EMAIL, ATTR_COMMONNAME);
            $owner=& $role->getRoleProvider();
            $owner_attrs= $owner->getUserAttributes(ATTR_EMAIL, ATTR_COMMONNAME);
            $admins_attrs= $role->getRoleAdministratorsAttributes();
            $to_emails= array(); // ( email => name, ... )
            if (isset($owner_attrs[ATTR_EMAIL])) {
                $to_emails[$owner_attrs[ATTR_EMAIL]]= $owner_attrs[ATTR_COMMONNAME];
            }
            foreach($admins_attrs as $admin_attrs) {
                if (isset($admin_attrs[ATTR_EMAIL])) {
                    $to_emails[$admin_attrs[ATTR_EMAIL]]= $admin_attrs[ATTR_COMMONNAME];
                }
            }
            if (isset($user_attrs[ATTR_EMAIL]) and !empty($to_emails)) {
                $from= $user_attrs[ATTR_EMAIL];
                // to: CSV of emails
                $to= implode(array_keys($to_emails),",");
                // parameters for notification
                $data= array();
                $data['from_name']= $user_attrs[ATTR_COMMONNAME];
                $data['from_uid']= $uid;
                $data['role_name']= $role->getTitle();
                $data['role_key']= $role->getRoleKey();
                $data['to_name']= implode(array_values($to_emails),", ");
                $message->stateChanged($from, $to, NEW_SUBSCRIPTION, false, $data);
            }
        }
    }
    $redirect= SUBSCRIBE_POST_SUBSCRIPTION_URL;
//    if (isset($_REQUEST['redirect']) and !empty($_REQUEST['redirect'])) {
//        $redirect= trim($_REQUEST['redirect']);
//    }
    header("Location: " . $redirect);
    exit; 
  }  
} 

if (! empty($_REQUEST["caller"])) {
  $caller = $_REQUEST["caller"];
} else {
  $caller = SUBSCRIBE_POST_CANCEL_URL;
}    

/* ***************************************************************** */
/*                       HTML output                                 */
/* ***************************************************************** */

// common page header
//
include("include/header.inc");
?>
  <h1><?= I18N::getMessage("subscribe.pagetitle") ?></h1>

  <?= I18N::getMessage("subscribe.text1") ?>

  <p/>

  <table width="90%" border="0" cellspacing="0" class="form-table">

  <tr>
    <td width="30%" valign="top" class="form-label form-first"><?= I18N::getMessage("subscribe.form1.label1") ?></td>
    <td valign="top" class="form-entry form-first">
      <?= Formatting::formatTableCellContent($role->getTitle()) ?>
    </td>
  </tr>
  
  <tr>
    <td valign="top" class="form-label"><?= I18N::getMessage("subscribe.form1.label2") ?></td>
    <td valign="top" class="form-entry">
    <?=  Formatting::formatTableCellContent($role->getDescription())  ?>
    </td>
  </tr>
  <tr>    
  </tr>    
  </table>  

  <form name="subscribeForm" action="subscribe.php" method="post">
   <?= I18N::getMessage("subscribe.text2") ?>
    <p/>

    <?php
    $attributeIDs = $role->getRequiredAttributeIDs();    
    ?>
 
    <table width="90%" border="0" cellspacing="0" class="form-table">
    <?php 
    $row= 0;
    foreach($attributeIDs as $attrID): 
       $value = $user->getAttribute((int)$attrID);
       $name =   AttributeType::getAttributeTypeName((int)$attrID);
       $first = ($row++ == 0) ? "form-first" : ""; 
    ?>
        <tr>
          <td width="30%" valign="top" class="form-label <?= $first ?>"><?= $name ?></td>
          <td width="70%" valign="top" class="form-entry <?= $first ?>"> 
             <?php if ($user->isAttributeTrusted((int)$attrID)): ?>
               <?= $value ?>              
             <?php else: ?>
               <input type="text" name="attribute<?= $attrID ?>" value="<?= $value ?>" >
               <?php if(isset($errorMessages["attribute" . $attrID])) :?>
                <br/><span class="error-message"><?= $errorMessages["attribute" . $attrID] ?></span>
               <?php endif; ?>
             <?php endif; ?>  
           </td>
        </tr>
    <?php endforeach; ?>
    </table>

    <?php if ($role->isPasswordProtected()): ?>
    <p/>
    <?= I18N::getMessage("subscribe.password.text") ?>
    <p/>
    <table width="90%" border="0" cellspacing="0" class="form-table">
    <tr>
        <td width="30%" valign="top" class="form-label form-first"><?= I18N::getMessage("subscribe.password.label") ?></td>
        <td valign="top" class="form-entry form-first">
        <input type="password" name="subscribe_password" size="40" maxlength="40" />
        <?php if(isset($errorMessages['subscribe_password'])) :?>
           <br/><span class="error-message"><?= $errorMessages['subscribe_password'] ?></span>
        <?php endif; ?>
        </td>
    </tr>
    </table>
    <?php endif; ?>  

  <?php /* ******************************************************************* */ ?>

  <p></p>
  <table width="90%" >
  <tr>
    <td width="100%"  valign="top" align="center" class="form-control" style="padding: 5pt;">
      <input type="hidden" name="rid" value="<?= $rid ?>">
      <input type="hidden" name="caller" value="<?= $caller ?>">        
      <input type="submit" name="do_cancel" value="<?= I18N::getMessage("subscribe.act.cancel") ?>" />
      <input type="submit" name="do_subscribe" value="<?= I18N::getMessage("subscribe.act.subscribe") ?>" />
    </td>    
  </tr>  
  </table>
  </form>
  
    
  </p>
  <?= I18N::getMessage("subscribe.text3") ?>
  <p></p>
  
<?php
// common page footer
//
include("include/footer.inc");
?>
